/*
Copyright (c) 2017-2020,
Battelle Memorial Institute; Lawrence Livermore National Security, LLC; Alliance for Sustainable
Energy, LLC.  See the top-level NOTICE for additional details. All rights reserved.
SPDX-License-Identifier: BSD-3-Clause
*/
#pragma once
#ifdef _MSC_VER
#    pragma warning(push, 0)
#    include <fmt/format.h>
#    pragma warning(pop)
#else
#    ifdef __GNUC__
#        pragma GCC system_header
#    endif
#    include <fmt/format.h>
#endif
