/*
Copyright (c) 2017-2018,
Battelle Memorial Institute; Lawrence Livermore National Security, LLC; Alliance for Sustainable
Energy, LLC All rights reserved. See LICENSE file and DISCLAIMER for more details.
*/
#include "helics/helics-config.h"

#include "gtest/gtest.h"

TEST(MpiCore_tests, init_test) {}
