/*
Copyright (c) 2017-2020,
Battelle Memorial Institute; Lawrence Livermore National Security, LLC; Alliance for Sustainable
Energy, LLC.  See the top-level NOTICE for additional details. All rights reserved.
SPDX-License-Identifier: BSD-3-Clause
*/

/** @file
Header file for inclusion of the entire CPP shared library
*/
#pragma once

#include "application_api.hpp"
#include "core/helicsVersion.hpp"
#include "core/helics_definitions.hpp"
#include "helics_enums.h"
