/*
Copyright (c) 2017-2020,
Battelle Memorial Institute; Lawrence Livermore National Security, LLC; Alliance for Sustainable
Energy, LLC.  See the top-level NOTICE for additional details. All rights reserved.
SPDX-License-Identifier: BSD-3-Clause
*/
#pragma once

#include "../application_api/BrokerApp.hpp"

// for maintaining the symbol in a public header
namespace helics {
namespace apps {
    using BrokerApp = BrokerApp;
}  // namespace apps

}  // namespace helics
