/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (http://www.swig.org).
 * Version 4.0.1
 *
 * This file is not intended to be easily readable and contains a number of
 * coding conventions designed to improve portability and efficiency. Do not make
 * changes to this file unless you know what you are doing--modify the SWIG
 * interface file instead.
 * ----------------------------------------------------------------------------- */


#ifndef SWIGPYTHON
#define SWIGPYTHON
#endif

#define SWIG_PYTHON_DIRECTOR_NO_VTABLE

/* -----------------------------------------------------------------------------
 *  This section contains generic SWIG labels for method/variable
 *  declarations/attributes, and other compiler dependent labels.
 * ----------------------------------------------------------------------------- */

/* template workaround for compilers that cannot correctly implement the C++ standard */
#ifndef SWIGTEMPLATEDISAMBIGUATOR
# if defined(__SUNPRO_CC) && (__SUNPRO_CC <= 0x560)
#  define SWIGTEMPLATEDISAMBIGUATOR template
# elif defined(__HP_aCC)
/* Needed even with `aCC -AA' when `aCC -V' reports HP ANSI C++ B3910B A.03.55 */
/* If we find a maximum version that requires this, the test would be __HP_aCC <= 35500 for A.03.55 */
#  define SWIGTEMPLATEDISAMBIGUATOR template
# else
#  define SWIGTEMPLATEDISAMBIGUATOR
# endif
#endif

/* inline attribute */
#ifndef SWIGINLINE
# if defined(__cplusplus) || (defined(__GNUC__) && !defined(__STRICT_ANSI__))
#   define SWIGINLINE inline
# else
#   define SWIGINLINE
# endif
#endif

/* attribute recognised by some compilers to avoid 'unused' warnings */
#ifndef SWIGUNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define SWIGUNUSED __attribute__ ((__unused__))
#   else
#     define SWIGUNUSED
#   endif
# elif defined(__ICC)
#   define SWIGUNUSED __attribute__ ((__unused__))
# else
#   define SWIGUNUSED
# endif
#endif

#ifndef SWIG_MSC_UNSUPPRESS_4505
# if defined(_MSC_VER)
#   pragma warning(disable : 4505) /* unreferenced local function has been removed */
# endif
#endif

#ifndef SWIGUNUSEDPARM
# ifdef __cplusplus
#   define SWIGUNUSEDPARM(p)
# else
#   define SWIGUNUSEDPARM(p) p SWIGUNUSED
# endif
#endif

/* internal SWIG method */
#ifndef SWIGINTERN
# define SWIGINTERN static SWIGUNUSED
#endif

/* internal inline SWIG method */
#ifndef SWIGINTERNINLINE
# define SWIGINTERNINLINE SWIGINTERN SWIGINLINE
#endif

/* exporting methods */
#if defined(__GNUC__)
#  if (__GNUC__ >= 4) || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4)
#    ifndef GCC_HASCLASSVISIBILITY
#      define GCC_HASCLASSVISIBILITY
#    endif
#  endif
#endif

#ifndef SWIGEXPORT
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   if defined(STATIC_LINKED)
#     define SWIGEXPORT
#   else
#     define SWIGEXPORT __declspec(dllexport)
#   endif
# else
#   if defined(__GNUC__) && defined(GCC_HASCLASSVISIBILITY)
#     define SWIGEXPORT __attribute__ ((visibility("default")))
#   else
#     define SWIGEXPORT
#   endif
# endif
#endif

/* calling conventions for Windows */
#ifndef SWIGSTDCALL
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   define SWIGSTDCALL __stdcall
# else
#   define SWIGSTDCALL
# endif
#endif

/* Deal with Microsoft's attempt at deprecating C standard runtime functions */
#if !defined(SWIG_NO_CRT_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_CRT_SECURE_NO_DEPRECATE)
# define _CRT_SECURE_NO_DEPRECATE
#endif

/* Deal with Microsoft's attempt at deprecating methods in the standard C++ library */
#if !defined(SWIG_NO_SCL_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_SCL_SECURE_NO_DEPRECATE)
# define _SCL_SECURE_NO_DEPRECATE
#endif

/* Deal with Apple's deprecated 'AssertMacros.h' from Carbon-framework */
#if defined(__APPLE__) && !defined(__ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES)
# define __ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES 0
#endif

/* Intel's compiler complains if a variable which was never initialised is
 * cast to void, which is a common idiom which we use to indicate that we
 * are aware a variable isn't used.  So we just silence that warning.
 * See: https://github.com/swig/swig/issues/192 for more discussion.
 */
#ifdef __INTEL_COMPILER
# pragma warning disable 592
#endif


#if defined(__GNUC__) && defined(_WIN32) && !defined(SWIG_PYTHON_NO_HYPOT_WORKAROUND)
/* Workaround for '::hypot' has not been declared', see https://bugs.python.org/issue11566 */
# include <math.h>
#endif

#if defined(_DEBUG) && defined(SWIG_PYTHON_INTERPRETER_NO_DEBUG)
/* Use debug wrappers with the Python release dll */
# undef _DEBUG
# include <Python.h>
# define _DEBUG 1
#else
# include <Python.h>
#endif

/* -----------------------------------------------------------------------------
 * swigrun.swg
 *
 * This file contains generic C API SWIG runtime support for pointer
 * type checking.
 * ----------------------------------------------------------------------------- */

/* This should only be incremented when either the layout of swig_type_info changes,
   or for whatever reason, the runtime changes incompatibly */
#define SWIG_RUNTIME_VERSION "4"

/* define SWIG_TYPE_TABLE_NAME as "SWIG_TYPE_TABLE" */
#ifdef SWIG_TYPE_TABLE
# define SWIG_QUOTE_STRING(x) #x
# define SWIG_EXPAND_AND_QUOTE_STRING(x) SWIG_QUOTE_STRING(x)
# define SWIG_TYPE_TABLE_NAME SWIG_EXPAND_AND_QUOTE_STRING(SWIG_TYPE_TABLE)
#else
# define SWIG_TYPE_TABLE_NAME
#endif

/*
  You can use the SWIGRUNTIME and SWIGRUNTIMEINLINE macros for
  creating a static or dynamic library from the SWIG runtime code.
  In 99.9% of the cases, SWIG just needs to declare them as 'static'.

  But only do this if strictly necessary, ie, if you have problems
  with your compiler or suchlike.
*/

#ifndef SWIGRUNTIME
# define SWIGRUNTIME SWIGINTERN
#endif

#ifndef SWIGRUNTIMEINLINE
# define SWIGRUNTIMEINLINE SWIGRUNTIME SWIGINLINE
#endif

/*  Generic buffer size */
#ifndef SWIG_BUFFER_SIZE
# define SWIG_BUFFER_SIZE 1024
#endif

/* Flags for pointer conversions */
#define SWIG_POINTER_DISOWN        0x1
#define SWIG_CAST_NEW_MEMORY       0x2
#define SWIG_POINTER_NO_NULL       0x4

/* Flags for new pointer objects */
#define SWIG_POINTER_OWN           0x1


/*
   Flags/methods for returning states.

   The SWIG conversion methods, as ConvertPtr, return an integer
   that tells if the conversion was successful or not. And if not,
   an error code can be returned (see swigerrors.swg for the codes).

   Use the following macros/flags to set or process the returning
   states.

   In old versions of SWIG, code such as the following was usually written:

     if (SWIG_ConvertPtr(obj,vptr,ty.flags) != -1) {
       // success code
     } else {
       //fail code
     }

   Now you can be more explicit:

    int res = SWIG_ConvertPtr(obj,vptr,ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
    } else {
      // fail code
    }

   which is the same really, but now you can also do

    Type *ptr;
    int res = SWIG_ConvertPtr(obj,(void **)(&ptr),ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
      if (SWIG_IsNewObj(res) {
        ...
	delete *ptr;
      } else {
        ...
      }
    } else {
      // fail code
    }

   I.e., now SWIG_ConvertPtr can return new objects and you can
   identify the case and take care of the deallocation. Of course that
   also requires SWIG_ConvertPtr to return new result values, such as

      int SWIG_ConvertPtr(obj, ptr,...) {
        if (<obj is ok>) {
          if (<need new object>) {
            *ptr = <ptr to new allocated object>;
            return SWIG_NEWOBJ;
          } else {
            *ptr = <ptr to old object>;
            return SWIG_OLDOBJ;
          }
        } else {
          return SWIG_BADOBJ;
        }
      }

   Of course, returning the plain '0(success)/-1(fail)' still works, but you can be
   more explicit by returning SWIG_BADOBJ, SWIG_ERROR or any of the
   SWIG errors code.

   Finally, if the SWIG_CASTRANK_MODE is enabled, the result code
   allows to return the 'cast rank', for example, if you have this

       int food(double)
       int fooi(int);

   and you call

      food(1)   // cast rank '1'  (1 -> 1.0)
      fooi(1)   // cast rank '0'

   just use the SWIG_AddCast()/SWIG_CheckState()
*/

#define SWIG_OK                    (0)
#define SWIG_ERROR                 (-1)
#define SWIG_IsOK(r)               (r >= 0)
#define SWIG_ArgError(r)           ((r != SWIG_ERROR) ? r : SWIG_TypeError)

/* The CastRankLimit says how many bits are used for the cast rank */
#define SWIG_CASTRANKLIMIT         (1 << 8)
/* The NewMask denotes the object was created (using new/malloc) */
#define SWIG_NEWOBJMASK            (SWIG_CASTRANKLIMIT  << 1)
/* The TmpMask is for in/out typemaps that use temporal objects */
#define SWIG_TMPOBJMASK            (SWIG_NEWOBJMASK << 1)
/* Simple returning values */
#define SWIG_BADOBJ                (SWIG_ERROR)
#define SWIG_OLDOBJ                (SWIG_OK)
#define SWIG_NEWOBJ                (SWIG_OK | SWIG_NEWOBJMASK)
#define SWIG_TMPOBJ                (SWIG_OK | SWIG_TMPOBJMASK)
/* Check, add and del mask methods */
#define SWIG_AddNewMask(r)         (SWIG_IsOK(r) ? (r | SWIG_NEWOBJMASK) : r)
#define SWIG_DelNewMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_NEWOBJMASK) : r)
#define SWIG_IsNewObj(r)           (SWIG_IsOK(r) && (r & SWIG_NEWOBJMASK))
#define SWIG_AddTmpMask(r)         (SWIG_IsOK(r) ? (r | SWIG_TMPOBJMASK) : r)
#define SWIG_DelTmpMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_TMPOBJMASK) : r)
#define SWIG_IsTmpObj(r)           (SWIG_IsOK(r) && (r & SWIG_TMPOBJMASK))

/* Cast-Rank Mode */
#if defined(SWIG_CASTRANK_MODE)
#  ifndef SWIG_TypeRank
#    define SWIG_TypeRank             unsigned long
#  endif
#  ifndef SWIG_MAXCASTRANK            /* Default cast allowed */
#    define SWIG_MAXCASTRANK          (2)
#  endif
#  define SWIG_CASTRANKMASK          ((SWIG_CASTRANKLIMIT) -1)
#  define SWIG_CastRank(r)           (r & SWIG_CASTRANKMASK)
SWIGINTERNINLINE int SWIG_AddCast(int r) {
  return SWIG_IsOK(r) ? ((SWIG_CastRank(r) < SWIG_MAXCASTRANK) ? (r + 1) : SWIG_ERROR) : r;
}
SWIGINTERNINLINE int SWIG_CheckState(int r) {
  return SWIG_IsOK(r) ? SWIG_CastRank(r) + 1 : 0;
}
#else /* no cast-rank mode */
#  define SWIG_AddCast(r) (r)
#  define SWIG_CheckState(r) (SWIG_IsOK(r) ? 1 : 0)
#endif


#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef void *(*swig_converter_func)(void *, int *);
typedef struct swig_type_info *(*swig_dycast_func)(void **);

/* Structure to store information on one type */
typedef struct swig_type_info {
  const char             *name;			/* mangled name of this type */
  const char             *str;			/* human readable name of this type */
  swig_dycast_func        dcast;		/* dynamic cast function down a hierarchy */
  struct swig_cast_info  *cast;			/* linked list of types that can cast into this type */
  void                   *clientdata;		/* language specific type data */
  int                    owndata;		/* flag if the structure owns the clientdata */
} swig_type_info;

/* Structure to store a type and conversion function used for casting */
typedef struct swig_cast_info {
  swig_type_info         *type;			/* pointer to type that is equivalent to this type */
  swig_converter_func     converter;		/* function to cast the void pointers */
  struct swig_cast_info  *next;			/* pointer to next cast in linked list */
  struct swig_cast_info  *prev;			/* pointer to the previous cast */
} swig_cast_info;

/* Structure used to store module information
 * Each module generates one structure like this, and the runtime collects
 * all of these structures and stores them in a circularly linked list.*/
typedef struct swig_module_info {
  swig_type_info         **types;		/* Array of pointers to swig_type_info structures that are in this module */
  size_t                 size;		        /* Number of types in this module */
  struct swig_module_info *next;		/* Pointer to next element in circularly linked list */
  swig_type_info         **type_initial;	/* Array of initially generated type structures */
  swig_cast_info         **cast_initial;	/* Array of initially generated casting structures */
  void                    *clientdata;		/* Language specific module data */
} swig_module_info;

/*
  Compare two type names skipping the space characters, therefore
  "char*" == "char *" and "Class<int>" == "Class<int >", etc.

  Return 0 when the two name types are equivalent, as in
  strncmp, but skipping ' '.
*/
SWIGRUNTIME int
SWIG_TypeNameComp(const char *f1, const char *l1,
		  const char *f2, const char *l2) {
  for (;(f1 != l1) && (f2 != l2); ++f1, ++f2) {
    while ((*f1 == ' ') && (f1 != l1)) ++f1;
    while ((*f2 == ' ') && (f2 != l2)) ++f2;
    if (*f1 != *f2) return (*f1 > *f2) ? 1 : -1;
  }
  return (int)((l1 - f1) - (l2 - f2));
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if equal, -1 if nb < tb, 1 if nb > tb
*/
SWIGRUNTIME int
SWIG_TypeCmp(const char *nb, const char *tb) {
  int equiv = 1;
  const char* te = tb + strlen(tb);
  const char* ne = nb;
  while (equiv != 0 && *ne) {
    for (nb = ne; *ne; ++ne) {
      if (*ne == '|') break;
    }
    equiv = SWIG_TypeNameComp(nb, ne, tb, te);
    if (*ne) ++ne;
  }
  return equiv;
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if not equal, 1 if equal
*/
SWIGRUNTIME int
SWIG_TypeEquiv(const char *nb, const char *tb) {
  return SWIG_TypeCmp(nb, tb) == 0 ? 1 : 0;
}

/*
  Check the typename
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheck(const char *c, swig_type_info *ty) {
  if (ty) {
    swig_cast_info *iter = ty->cast;
    while (iter) {
      if (strcmp(iter->type->name, c) == 0) {
        if (iter == ty->cast)
          return iter;
        /* Move iter to the top of the linked list */
        iter->prev->next = iter->next;
        if (iter->next)
          iter->next->prev = iter->prev;
        iter->next = ty->cast;
        iter->prev = 0;
        if (ty->cast) ty->cast->prev = iter;
        ty->cast = iter;
        return iter;
      }
      iter = iter->next;
    }
  }
  return 0;
}

/*
  Identical to SWIG_TypeCheck, except strcmp is replaced with a pointer comparison
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheckStruct(swig_type_info *from, swig_type_info *ty) {
  if (ty) {
    swig_cast_info *iter = ty->cast;
    while (iter) {
      if (iter->type == from) {
        if (iter == ty->cast)
          return iter;
        /* Move iter to the top of the linked list */
        iter->prev->next = iter->next;
        if (iter->next)
          iter->next->prev = iter->prev;
        iter->next = ty->cast;
        iter->prev = 0;
        if (ty->cast) ty->cast->prev = iter;
        ty->cast = iter;
        return iter;
      }
      iter = iter->next;
    }
  }
  return 0;
}

/*
  Cast a pointer up an inheritance hierarchy
*/
SWIGRUNTIMEINLINE void *
SWIG_TypeCast(swig_cast_info *ty, void *ptr, int *newmemory) {
  return ((!ty) || (!ty->converter)) ? ptr : (*ty->converter)(ptr, newmemory);
}

/*
   Dynamic pointer casting. Down an inheritance hierarchy
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeDynamicCast(swig_type_info *ty, void **ptr) {
  swig_type_info *lastty = ty;
  if (!ty || !ty->dcast) return ty;
  while (ty && (ty->dcast)) {
    ty = (*ty->dcast)(ptr);
    if (ty) lastty = ty;
  }
  return lastty;
}

/*
  Return the name associated with this type
*/
SWIGRUNTIMEINLINE const char *
SWIG_TypeName(const swig_type_info *ty) {
  return ty->name;
}

/*
  Return the pretty name associated with this type,
  that is an unmangled type name in a form presentable to the user.
*/
SWIGRUNTIME const char *
SWIG_TypePrettyName(const swig_type_info *type) {
  /* The "str" field contains the equivalent pretty names of the
     type, separated by vertical-bar characters.  We choose
     to print the last name, as it is often (?) the most
     specific. */
  if (!type) return NULL;
  if (type->str != NULL) {
    const char *last_name = type->str;
    const char *s;
    for (s = type->str; *s; s++)
      if (*s == '|') last_name = s+1;
    return last_name;
  }
  else
    return type->name;
}

/*
   Set the clientdata field for a type
*/
SWIGRUNTIME void
SWIG_TypeClientData(swig_type_info *ti, void *clientdata) {
  swig_cast_info *cast = ti->cast;
  /* if (ti->clientdata == clientdata) return; */
  ti->clientdata = clientdata;

  while (cast) {
    if (!cast->converter) {
      swig_type_info *tc = cast->type;
      if (!tc->clientdata) {
	SWIG_TypeClientData(tc, clientdata);
      }
    }
    cast = cast->next;
  }
}
SWIGRUNTIME void
SWIG_TypeNewClientData(swig_type_info *ti, void *clientdata) {
  SWIG_TypeClientData(ti, clientdata);
  ti->owndata = 1;
}

/*
  Search for a swig_type_info structure only by mangled name
  Search is a O(log #types)

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_MangledTypeQueryModule(swig_module_info *start,
                            swig_module_info *end,
		            const char *name) {
  swig_module_info *iter = start;
  do {
    if (iter->size) {
      size_t l = 0;
      size_t r = iter->size - 1;
      do {
	/* since l+r >= 0, we can (>> 1) instead (/ 2) */
	size_t i = (l + r) >> 1;
	const char *iname = iter->types[i]->name;
	if (iname) {
	  int compare = strcmp(name, iname);
	  if (compare == 0) {
	    return iter->types[i];
	  } else if (compare < 0) {
	    if (i) {
	      r = i - 1;
	    } else {
	      break;
	    }
	  } else if (compare > 0) {
	    l = i + 1;
	  }
	} else {
	  break; /* should never happen */
	}
      } while (l <= r);
    }
    iter = iter->next;
  } while (iter != end);
  return 0;
}

/*
  Search for a swig_type_info structure for either a mangled name or a human readable name.
  It first searches the mangled names of the types, which is a O(log #types)
  If a type is not found it then searches the human readable names, which is O(#types).

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeQueryModule(swig_module_info *start,
                     swig_module_info *end,
		     const char *name) {
  /* STEP 1: Search the name field using binary search */
  swig_type_info *ret = SWIG_MangledTypeQueryModule(start, end, name);
  if (ret) {
    return ret;
  } else {
    /* STEP 2: If the type hasn't been found, do a complete search
       of the str field (the human readable name) */
    swig_module_info *iter = start;
    do {
      size_t i = 0;
      for (; i < iter->size; ++i) {
	if (iter->types[i]->str && (SWIG_TypeEquiv(iter->types[i]->str, name)))
	  return iter->types[i];
      }
      iter = iter->next;
    } while (iter != end);
  }

  /* neither found a match */
  return 0;
}

/*
   Pack binary data into a string
*/
SWIGRUNTIME char *
SWIG_PackData(char *c, void *ptr, size_t sz) {
  static const char hex[17] = "0123456789abcdef";
  const unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu =  u + sz;
  for (; u != eu; ++u) {
    unsigned char uu = *u;
    *(c++) = hex[(uu & 0xf0) >> 4];
    *(c++) = hex[uu & 0xf];
  }
  return c;
}

/*
   Unpack binary data from a string
*/
SWIGRUNTIME const char *
SWIG_UnpackData(const char *c, void *ptr, size_t sz) {
  unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu = u + sz;
  for (; u != eu; ++u) {
    char d = *(c++);
    unsigned char uu;
    if ((d >= '0') && (d <= '9'))
      uu = (unsigned char)((d - '0') << 4);
    else if ((d >= 'a') && (d <= 'f'))
      uu = (unsigned char)((d - ('a'-10)) << 4);
    else
      return (char *) 0;
    d = *(c++);
    if ((d >= '0') && (d <= '9'))
      uu |= (unsigned char)(d - '0');
    else if ((d >= 'a') && (d <= 'f'))
      uu |= (unsigned char)(d - ('a'-10));
    else
      return (char *) 0;
    *u = uu;
  }
  return c;
}

/*
   Pack 'void *' into a string buffer.
*/
SWIGRUNTIME char *
SWIG_PackVoidPtr(char *buff, void *ptr, const char *name, size_t bsz) {
  char *r = buff;
  if ((2*sizeof(void *) + 2) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,&ptr,sizeof(void *));
  if (strlen(name) + 1 > (bsz - (r - buff))) return 0;
  strcpy(r,name);
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackVoidPtr(const char *c, void **ptr, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      *ptr = (void *) 0;
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sizeof(void *));
}

SWIGRUNTIME char *
SWIG_PackDataName(char *buff, void *ptr, size_t sz, const char *name, size_t bsz) {
  char *r = buff;
  size_t lname = (name ? strlen(name) : 0);
  if ((2*sz + 2 + lname) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,ptr,sz);
  if (lname) {
    strncpy(r,name,lname+1);
  } else {
    *r = 0;
  }
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackDataName(const char *c, void *ptr, size_t sz, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      memset(ptr,0,sz);
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sz);
}

#ifdef __cplusplus
}
#endif

/*  Errors in SWIG */
#define  SWIG_UnknownError    	   -1
#define  SWIG_IOError        	   -2
#define  SWIG_RuntimeError   	   -3
#define  SWIG_IndexError     	   -4
#define  SWIG_TypeError      	   -5
#define  SWIG_DivisionByZero 	   -6
#define  SWIG_OverflowError  	   -7
#define  SWIG_SyntaxError    	   -8
#define  SWIG_ValueError     	   -9
#define  SWIG_SystemError    	   -10
#define  SWIG_AttributeError 	   -11
#define  SWIG_MemoryError    	   -12
#define  SWIG_NullReferenceError   -13



/* Compatibility macros for Python 3 */
#if PY_VERSION_HEX >= 0x03000000

#define PyClass_Check(obj) PyObject_IsInstance(obj, (PyObject *)&PyType_Type)
#define PyInt_Check(x) PyLong_Check(x)
#define PyInt_AsLong(x) PyLong_AsLong(x)
#define PyInt_FromLong(x) PyLong_FromLong(x)
#define PyInt_FromSize_t(x) PyLong_FromSize_t(x)
#define PyString_Check(name) PyBytes_Check(name)
#define PyString_FromString(x) PyUnicode_FromString(x)
#define PyString_Format(fmt, args)  PyUnicode_Format(fmt, args)
#define PyString_AsString(str) PyBytes_AsString(str)
#define PyString_Size(str) PyBytes_Size(str)	
#define PyString_InternFromString(key) PyUnicode_InternFromString(key)
#define Py_TPFLAGS_HAVE_CLASS Py_TPFLAGS_BASETYPE
#define PyString_AS_STRING(x) PyUnicode_AS_STRING(x)
#define _PyLong_FromSsize_t(x) PyLong_FromSsize_t(x)

#endif

#ifndef Py_TYPE
#  define Py_TYPE(op) ((op)->ob_type)
#endif

/* SWIG APIs for compatibility of both Python 2 & 3 */

#if PY_VERSION_HEX >= 0x03000000
#  define SWIG_Python_str_FromFormat PyUnicode_FromFormat
#else
#  define SWIG_Python_str_FromFormat PyString_FromFormat
#endif


/* Warning: This function will allocate a new string in Python 3,
 * so please call SWIG_Python_str_DelForPy3(x) to free the space.
 */
SWIGINTERN char*
SWIG_Python_str_AsChar(PyObject *str)
{
#if PY_VERSION_HEX >= 0x03000000
  char *newstr = 0;
  str = PyUnicode_AsUTF8String(str);
  if (str) {
    char *cstr;
    Py_ssize_t len;
    PyBytes_AsStringAndSize(str, &cstr, &len);
    newstr = (char *) malloc(len+1);
    memcpy(newstr, cstr, len+1);
    Py_XDECREF(str);
  }
  return newstr;
#else
  return PyString_AsString(str);
#endif
}

#if PY_VERSION_HEX >= 0x03000000
#  define SWIG_Python_str_DelForPy3(x) free( (void*) (x) )
#else
#  define SWIG_Python_str_DelForPy3(x) 
#endif


SWIGINTERN PyObject*
SWIG_Python_str_FromChar(const char *c)
{
#if PY_VERSION_HEX >= 0x03000000
  return PyUnicode_FromString(c); 
#else
  return PyString_FromString(c);
#endif
}

#ifndef PyObject_DEL
# define PyObject_DEL PyObject_Del
#endif

// SWIGPY_USE_CAPSULE is no longer used within SWIG itself, but some user
// interface files check for it.
# define SWIGPY_USE_CAPSULE
# define SWIGPY_CAPSULE_NAME ("swig_runtime_data" SWIG_RUNTIME_VERSION ".type_pointer_capsule" SWIG_TYPE_TABLE_NAME)

#if PY_VERSION_HEX < 0x03020000
#define PyDescr_TYPE(x) (((PyDescrObject *)(x))->d_type)
#define PyDescr_NAME(x) (((PyDescrObject *)(x))->d_name)
#define Py_hash_t long
#endif

/* -----------------------------------------------------------------------------
 * error manipulation
 * ----------------------------------------------------------------------------- */

SWIGRUNTIME PyObject*
SWIG_Python_ErrorType(int code) {
  PyObject* type = 0;
  switch(code) {
  case SWIG_MemoryError:
    type = PyExc_MemoryError;
    break;
  case SWIG_IOError:
    type = PyExc_IOError;
    break;
  case SWIG_RuntimeError:
    type = PyExc_RuntimeError;
    break;
  case SWIG_IndexError:
    type = PyExc_IndexError;
    break;
  case SWIG_TypeError:
    type = PyExc_TypeError;
    break;
  case SWIG_DivisionByZero:
    type = PyExc_ZeroDivisionError;
    break;
  case SWIG_OverflowError:
    type = PyExc_OverflowError;
    break;
  case SWIG_SyntaxError:
    type = PyExc_SyntaxError;
    break;
  case SWIG_ValueError:
    type = PyExc_ValueError;
    break;
  case SWIG_SystemError:
    type = PyExc_SystemError;
    break;
  case SWIG_AttributeError:
    type = PyExc_AttributeError;
    break;
  default:
    type = PyExc_RuntimeError;
  }
  return type;
}


SWIGRUNTIME void
SWIG_Python_AddErrorMsg(const char* mesg)
{
  PyObject *type = 0;
  PyObject *value = 0;
  PyObject *traceback = 0;

  if (PyErr_Occurred())
    PyErr_Fetch(&type, &value, &traceback);
  if (value) {
    PyObject *old_str = PyObject_Str(value);
    const char *tmp = SWIG_Python_str_AsChar(old_str);
    PyErr_Clear();
    Py_XINCREF(type);
    if (tmp)
      PyErr_Format(type, "%s %s", tmp, mesg);
    else
      PyErr_Format(type, "%s", mesg);
    SWIG_Python_str_DelForPy3(tmp);
    Py_DECREF(old_str);
    Py_DECREF(value);
  } else {
    PyErr_SetString(PyExc_RuntimeError, mesg);
  }
}

SWIGRUNTIME int
SWIG_Python_TypeErrorOccurred(PyObject *obj)
{
  PyObject *error;
  if (obj)
    return 0;
  error = PyErr_Occurred();
  return error && PyErr_GivenExceptionMatches(error, PyExc_TypeError);
}

SWIGRUNTIME void
SWIG_Python_RaiseOrModifyTypeError(const char *message)
{
  if (SWIG_Python_TypeErrorOccurred(NULL)) {
    /* Use existing TypeError to preserve stacktrace and enhance with given message */
    PyObject *newvalue;
    PyObject *type = NULL, *value = NULL, *traceback = NULL;
    PyErr_Fetch(&type, &value, &traceback);
#if PY_VERSION_HEX >= 0x03000000
    newvalue = PyUnicode_FromFormat("%S\nAdditional information:\n%s", value, message);
#else
    newvalue = PyString_FromFormat("%s\nAdditional information:\n%s", PyString_AsString(value), message);
#endif
    Py_XDECREF(value);
    PyErr_Restore(type, newvalue, traceback);
  } else {
    /* Raise TypeError using given message */
    PyErr_SetString(PyExc_TypeError, message);
  }
}

#if defined(SWIG_PYTHON_NO_THREADS)
#  if defined(SWIG_PYTHON_THREADS)
#    undef SWIG_PYTHON_THREADS
#  endif
#endif
#if defined(SWIG_PYTHON_THREADS) /* Threading support is enabled */
#  if !defined(SWIG_PYTHON_USE_GIL) && !defined(SWIG_PYTHON_NO_USE_GIL)
#    define SWIG_PYTHON_USE_GIL
#  endif
#  if defined(SWIG_PYTHON_USE_GIL) /* Use PyGILState threads calls */
#    ifndef SWIG_PYTHON_INITIALIZE_THREADS
#     define SWIG_PYTHON_INITIALIZE_THREADS  PyEval_InitThreads() 
#    endif
#    ifdef __cplusplus /* C++ code */
       class SWIG_Python_Thread_Block {
         bool status;
         PyGILState_STATE state;
       public:
         void end() { if (status) { PyGILState_Release(state); status = false;} }
         SWIG_Python_Thread_Block() : status(true), state(PyGILState_Ensure()) {}
         ~SWIG_Python_Thread_Block() { end(); }
       };
       class SWIG_Python_Thread_Allow {
         bool status;
         PyThreadState *save;
       public:
         void end() { if (status) { PyEval_RestoreThread(save); status = false; }}
         SWIG_Python_Thread_Allow() : status(true), save(PyEval_SaveThread()) {}
         ~SWIG_Python_Thread_Allow() { end(); }
       };
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK   SWIG_Python_Thread_Block _swig_thread_block
#      define SWIG_PYTHON_THREAD_END_BLOCK     _swig_thread_block.end()
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW   SWIG_Python_Thread_Allow _swig_thread_allow
#      define SWIG_PYTHON_THREAD_END_ALLOW     _swig_thread_allow.end()
#    else /* C code */
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK   PyGILState_STATE _swig_thread_block = PyGILState_Ensure()
#      define SWIG_PYTHON_THREAD_END_BLOCK     PyGILState_Release(_swig_thread_block)
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW   PyThreadState *_swig_thread_allow = PyEval_SaveThread()
#      define SWIG_PYTHON_THREAD_END_ALLOW     PyEval_RestoreThread(_swig_thread_allow)
#    endif
#  else /* Old thread way, not implemented, user must provide it */
#    if !defined(SWIG_PYTHON_INITIALIZE_THREADS)
#      define SWIG_PYTHON_INITIALIZE_THREADS
#    endif
#    if !defined(SWIG_PYTHON_THREAD_BEGIN_BLOCK)
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK
#    endif
#    if !defined(SWIG_PYTHON_THREAD_END_BLOCK)
#      define SWIG_PYTHON_THREAD_END_BLOCK
#    endif
#    if !defined(SWIG_PYTHON_THREAD_BEGIN_ALLOW)
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW
#    endif
#    if !defined(SWIG_PYTHON_THREAD_END_ALLOW)
#      define SWIG_PYTHON_THREAD_END_ALLOW
#    endif
#  endif
#else /* No thread support */
#  define SWIG_PYTHON_INITIALIZE_THREADS
#  define SWIG_PYTHON_THREAD_BEGIN_BLOCK
#  define SWIG_PYTHON_THREAD_END_BLOCK
#  define SWIG_PYTHON_THREAD_BEGIN_ALLOW
#  define SWIG_PYTHON_THREAD_END_ALLOW
#endif

/* -----------------------------------------------------------------------------
 * Python API portion that goes into the runtime
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#endif

/* -----------------------------------------------------------------------------
 * Constant declarations
 * ----------------------------------------------------------------------------- */

/* Constant Types */
#define SWIG_PY_POINTER 4
#define SWIG_PY_BINARY  5

/* Constant information structure */
typedef struct swig_const_info {
  int type;
  const char *name;
  long lvalue;
  double dvalue;
  void   *pvalue;
  swig_type_info **ptype;
} swig_const_info;

#ifdef __cplusplus
}
#endif


/* -----------------------------------------------------------------------------
 * pyrun.swg
 *
 * This file contains the runtime support for Python modules
 * and includes code for managing global variables and pointer
 * type checking.
 *
 * ----------------------------------------------------------------------------- */

#if PY_VERSION_HEX < 0x02070000 /* 2.7.0 */
# error "This version of SWIG only supports Python >= 2.7"
#endif

#if PY_VERSION_HEX >= 0x03000000 && PY_VERSION_HEX < 0x03020000
# error "This version of SWIG only supports Python 3 >= 3.2"
#endif

/* Common SWIG API */

/* for raw pointers */
#define SWIG_Python_ConvertPtr(obj, pptr, type, flags)  SWIG_Python_ConvertPtrAndOwn(obj, pptr, type, flags, 0)
#define SWIG_ConvertPtr(obj, pptr, type, flags)         SWIG_Python_ConvertPtr(obj, pptr, type, flags)
#define SWIG_ConvertPtrAndOwn(obj,pptr,type,flags,own)  SWIG_Python_ConvertPtrAndOwn(obj, pptr, type, flags, own)

#ifdef SWIGPYTHON_BUILTIN
#define SWIG_NewPointerObj(ptr, type, flags)            SWIG_Python_NewPointerObj(self, ptr, type, flags)
#else
#define SWIG_NewPointerObj(ptr, type, flags)            SWIG_Python_NewPointerObj(NULL, ptr, type, flags)
#endif

#define SWIG_InternalNewPointerObj(ptr, type, flags)	SWIG_Python_NewPointerObj(NULL, ptr, type, flags)

#define SWIG_CheckImplicit(ty)                          SWIG_Python_CheckImplicit(ty) 
#define SWIG_AcquirePtr(ptr, src)                       SWIG_Python_AcquirePtr(ptr, src)
#define swig_owntype                                    int

/* for raw packed data */
#define SWIG_ConvertPacked(obj, ptr, sz, ty)            SWIG_Python_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewPackedObj(ptr, sz, type)                SWIG_Python_NewPackedObj(ptr, sz, type)

/* for class or struct pointers */
#define SWIG_ConvertInstance(obj, pptr, type, flags)    SWIG_ConvertPtr(obj, pptr, type, flags)
#define SWIG_NewInstanceObj(ptr, type, flags)           SWIG_NewPointerObj(ptr, type, flags)

/* for C or C++ function pointers */
#define SWIG_ConvertFunctionPtr(obj, pptr, type)        SWIG_Python_ConvertFunctionPtr(obj, pptr, type)
#define SWIG_NewFunctionPtrObj(ptr, type)               SWIG_Python_NewPointerObj(NULL, ptr, type, 0)

/* for C++ member pointers, ie, member methods */
#define SWIG_ConvertMember(obj, ptr, sz, ty)            SWIG_Python_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewMemberObj(ptr, sz, type)                SWIG_Python_NewPackedObj(ptr, sz, type)


/* Runtime API */

#define SWIG_GetModule(clientdata)                      SWIG_Python_GetModule(clientdata)
#define SWIG_SetModule(clientdata, pointer)             SWIG_Python_SetModule(pointer)
#define SWIG_NewClientData(obj)                         SwigPyClientData_New(obj)

#define SWIG_SetErrorObj                                SWIG_Python_SetErrorObj                            
#define SWIG_SetErrorMsg                        	SWIG_Python_SetErrorMsg				   
#define SWIG_ErrorType(code)                    	SWIG_Python_ErrorType(code)                        
#define SWIG_Error(code, msg)            		SWIG_Python_SetErrorMsg(SWIG_ErrorType(code), msg) 
#define SWIG_fail                        		goto fail					   


/* Runtime API implementation */

/* Error manipulation */

SWIGINTERN void 
SWIG_Python_SetErrorObj(PyObject *errtype, PyObject *obj) {
  SWIG_PYTHON_THREAD_BEGIN_BLOCK; 
  PyErr_SetObject(errtype, obj);
  Py_DECREF(obj);
  SWIG_PYTHON_THREAD_END_BLOCK;
}

SWIGINTERN void 
SWIG_Python_SetErrorMsg(PyObject *errtype, const char *msg) {
  SWIG_PYTHON_THREAD_BEGIN_BLOCK;
  PyErr_SetString(errtype, msg);
  SWIG_PYTHON_THREAD_END_BLOCK;
}

#define SWIG_Python_Raise(obj, type, desc)  SWIG_Python_SetErrorObj(SWIG_Python_ExceptionType(desc), obj)

/* Set a constant value */

#if defined(SWIGPYTHON_BUILTIN)

SWIGINTERN void
SwigPyBuiltin_AddPublicSymbol(PyObject *seq, const char *key) {
  PyObject *s = PyString_InternFromString(key);
  PyList_Append(seq, s);
  Py_DECREF(s);
}

SWIGINTERN void
SWIG_Python_SetConstant(PyObject *d, PyObject *public_interface, const char *name, PyObject *obj) {   
  PyDict_SetItemString(d, name, obj);
  Py_DECREF(obj);
  if (public_interface)
    SwigPyBuiltin_AddPublicSymbol(public_interface, name);
}

#else

SWIGINTERN void
SWIG_Python_SetConstant(PyObject *d, const char *name, PyObject *obj) {   
  PyDict_SetItemString(d, name, obj);
  Py_DECREF(obj);                            
}

#endif

/* Append a value to the result obj */

SWIGINTERN PyObject*
SWIG_Python_AppendOutput(PyObject* result, PyObject* obj) {
  if (!result) {
    result = obj;
  } else if (result == Py_None) {
    Py_DECREF(result);
    result = obj;
  } else {
    if (!PyList_Check(result)) {
      PyObject *o2 = result;
      result = PyList_New(1);
      PyList_SetItem(result, 0, o2);
    }
    PyList_Append(result,obj);
    Py_DECREF(obj);
  }
  return result;
}

/* Unpack the argument tuple */

SWIGINTERN Py_ssize_t
SWIG_Python_UnpackTuple(PyObject *args, const char *name, Py_ssize_t min, Py_ssize_t max, PyObject **objs)
{
  if (!args) {
    if (!min && !max) {
      return 1;
    } else {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got none", 
		   name, (min == max ? "" : "at least "), (int)min);
      return 0;
    }
  }  
  if (!PyTuple_Check(args)) {
    if (min <= 1 && max >= 1) {
      Py_ssize_t i;
      objs[0] = args;
      for (i = 1; i < max; ++i) {
	objs[i] = 0;
      }
      return 2;
    }
    PyErr_SetString(PyExc_SystemError, "UnpackTuple() argument list is not a tuple");
    return 0;
  } else {
    Py_ssize_t l = PyTuple_GET_SIZE(args);
    if (l < min) {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got %d", 
		   name, (min == max ? "" : "at least "), (int)min, (int)l);
      return 0;
    } else if (l > max) {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got %d", 
		   name, (min == max ? "" : "at most "), (int)max, (int)l);
      return 0;
    } else {
      Py_ssize_t i;
      for (i = 0; i < l; ++i) {
	objs[i] = PyTuple_GET_ITEM(args, i);
      }
      for (; l < max; ++l) {
	objs[l] = 0;
      }
      return i + 1;
    }    
  }
}

/* A functor is a function object with one single object argument */
#define SWIG_Python_CallFunctor(functor, obj)	        PyObject_CallFunctionObjArgs(functor, obj, NULL);

/*
  Helper for static pointer initialization for both C and C++ code, for example
  static PyObject *SWIG_STATIC_POINTER(MyVar) = NewSomething(...);
*/
#ifdef __cplusplus
#define SWIG_STATIC_POINTER(var)  var
#else
#define SWIG_STATIC_POINTER(var)  var = 0; if (!var) var
#endif

/* -----------------------------------------------------------------------------
 * Pointer declarations
 * ----------------------------------------------------------------------------- */

/* Flags for new pointer objects */
#define SWIG_POINTER_NOSHADOW       (SWIG_POINTER_OWN      << 1)
#define SWIG_POINTER_NEW            (SWIG_POINTER_NOSHADOW | SWIG_POINTER_OWN)

#define SWIG_POINTER_IMPLICIT_CONV  (SWIG_POINTER_DISOWN   << 1)

#define SWIG_BUILTIN_TP_INIT	    (SWIG_POINTER_OWN << 2)
#define SWIG_BUILTIN_INIT	    (SWIG_BUILTIN_TP_INIT | SWIG_POINTER_OWN)

#ifdef __cplusplus
extern "C" {
#endif

/* The python void return value */

SWIGRUNTIMEINLINE PyObject * 
SWIG_Py_Void(void)
{
  PyObject *none = Py_None;
  Py_INCREF(none);
  return none;
}

/* SwigPyClientData */

typedef struct {
  PyObject *klass;
  PyObject *newraw;
  PyObject *newargs;
  PyObject *destroy;
  int delargs;
  int implicitconv;
  PyTypeObject *pytype;
} SwigPyClientData;

SWIGRUNTIMEINLINE int 
SWIG_Python_CheckImplicit(swig_type_info *ty)
{
  SwigPyClientData *data = (SwigPyClientData *)ty->clientdata;
  int fail = data ? data->implicitconv : 0;
  if (fail)
    PyErr_SetString(PyExc_TypeError, "Implicit conversion is prohibited for explicit constructors.");
  return fail;
}

SWIGRUNTIMEINLINE PyObject *
SWIG_Python_ExceptionType(swig_type_info *desc) {
  SwigPyClientData *data = desc ? (SwigPyClientData *) desc->clientdata : 0;
  PyObject *klass = data ? data->klass : 0;
  return (klass ? klass : PyExc_RuntimeError);
}


SWIGRUNTIME SwigPyClientData * 
SwigPyClientData_New(PyObject* obj)
{
  if (!obj) {
    return 0;
  } else {
    SwigPyClientData *data = (SwigPyClientData *)malloc(sizeof(SwigPyClientData));
    /* the klass element */
    data->klass = obj;
    Py_INCREF(data->klass);
    /* the newraw method and newargs arguments used to create a new raw instance */
    if (PyClass_Check(obj)) {
      data->newraw = 0;
      data->newargs = obj;
      Py_INCREF(obj);
    } else {
      data->newraw = PyObject_GetAttrString(data->klass, "__new__");
      if (data->newraw) {
	Py_INCREF(data->newraw);
	data->newargs = PyTuple_New(1);
	PyTuple_SetItem(data->newargs, 0, obj);
      } else {
	data->newargs = obj;
      }
      Py_INCREF(data->newargs);
    }
    /* the destroy method, aka as the C++ delete method */
    data->destroy = PyObject_GetAttrString(data->klass, "__swig_destroy__");
    if (PyErr_Occurred()) {
      PyErr_Clear();
      data->destroy = 0;
    }
    if (data->destroy) {
      int flags;
      Py_INCREF(data->destroy);
      flags = PyCFunction_GET_FLAGS(data->destroy);
      data->delargs = !(flags & (METH_O));
    } else {
      data->delargs = 0;
    }
    data->implicitconv = 0;
    data->pytype = 0;
    return data;
  }
}

SWIGRUNTIME void 
SwigPyClientData_Del(SwigPyClientData *data) {
  Py_XDECREF(data->newraw);
  Py_XDECREF(data->newargs);
  Py_XDECREF(data->destroy);
}

/* =============== SwigPyObject =====================*/

typedef struct {
  PyObject_HEAD
  void *ptr;
  swig_type_info *ty;
  int own;
  PyObject *next;
#ifdef SWIGPYTHON_BUILTIN
  PyObject *dict;
#endif
} SwigPyObject;


#ifdef SWIGPYTHON_BUILTIN

SWIGRUNTIME PyObject *
SwigPyObject_get___dict__(PyObject *v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *)v;

  if (!sobj->dict)
    sobj->dict = PyDict_New();

  Py_INCREF(sobj->dict);
  return sobj->dict;
}

#endif

SWIGRUNTIME PyObject *
SwigPyObject_long(SwigPyObject *v)
{
  return PyLong_FromVoidPtr(v->ptr);
}

SWIGRUNTIME PyObject *
SwigPyObject_format(const char* fmt, SwigPyObject *v)
{
  PyObject *res = NULL;
  PyObject *args = PyTuple_New(1);
  if (args) {
    if (PyTuple_SetItem(args, 0, SwigPyObject_long(v)) == 0) {
      PyObject *ofmt = SWIG_Python_str_FromChar(fmt);
      if (ofmt) {
#if PY_VERSION_HEX >= 0x03000000
	res = PyUnicode_Format(ofmt,args);
#else
	res = PyString_Format(ofmt,args);
#endif
	Py_DECREF(ofmt);
      }
      Py_DECREF(args);
    }
  }
  return res;
}

SWIGRUNTIME PyObject *
SwigPyObject_oct(SwigPyObject *v)
{
  return SwigPyObject_format("%o",v);
}

SWIGRUNTIME PyObject *
SwigPyObject_hex(SwigPyObject *v)
{
  return SwigPyObject_format("%x",v);
}

SWIGRUNTIME PyObject *
SwigPyObject_repr(SwigPyObject *v)
{
  const char *name = SWIG_TypePrettyName(v->ty);
  PyObject *repr = SWIG_Python_str_FromFormat("<Swig Object of type '%s' at %p>", (name ? name : "unknown"), (void *)v);
  if (v->next) {
    PyObject *nrep = SwigPyObject_repr((SwigPyObject *)v->next);
# if PY_VERSION_HEX >= 0x03000000
    PyObject *joined = PyUnicode_Concat(repr, nrep);
    Py_DecRef(repr);
    Py_DecRef(nrep);
    repr = joined;
# else
    PyString_ConcatAndDel(&repr,nrep);
# endif
  }
  return repr;  
}

/* We need a version taking two PyObject* parameters so it's a valid
 * PyCFunction to use in swigobject_methods[]. */
SWIGRUNTIME PyObject *
SwigPyObject_repr2(PyObject *v, PyObject *SWIGUNUSEDPARM(args))
{
  return SwigPyObject_repr((SwigPyObject*)v);
}

SWIGRUNTIME int
SwigPyObject_compare(SwigPyObject *v, SwigPyObject *w)
{
  void *i = v->ptr;
  void *j = w->ptr;
  return (i < j) ? -1 : ((i > j) ? 1 : 0);
}

/* Added for Python 3.x, would it also be useful for Python 2.x? */
SWIGRUNTIME PyObject*
SwigPyObject_richcompare(SwigPyObject *v, SwigPyObject *w, int op)
{
  PyObject* res;
  if( op != Py_EQ && op != Py_NE ) {
    Py_INCREF(Py_NotImplemented);
    return Py_NotImplemented;
  }
  res = PyBool_FromLong( (SwigPyObject_compare(v, w)==0) == (op == Py_EQ) ? 1 : 0);
  return res;  
}


SWIGRUNTIME PyTypeObject* SwigPyObject_TypeOnce(void);

#ifdef SWIGPYTHON_BUILTIN
static swig_type_info *SwigPyObject_stype = 0;
SWIGRUNTIME PyTypeObject*
SwigPyObject_type(void) {
    SwigPyClientData *cd;
    assert(SwigPyObject_stype);
    cd = (SwigPyClientData*) SwigPyObject_stype->clientdata;
    assert(cd);
    assert(cd->pytype);
    return cd->pytype;
}
#else
SWIGRUNTIME PyTypeObject*
SwigPyObject_type(void) {
  static PyTypeObject *SWIG_STATIC_POINTER(type) = SwigPyObject_TypeOnce();
  return type;
}
#endif

SWIGRUNTIMEINLINE int
SwigPyObject_Check(PyObject *op) {
#ifdef SWIGPYTHON_BUILTIN
  PyTypeObject *target_tp = SwigPyObject_type();
  if (PyType_IsSubtype(op->ob_type, target_tp))
    return 1;
  return (strcmp(op->ob_type->tp_name, "SwigPyObject") == 0);
#else
  return (Py_TYPE(op) == SwigPyObject_type())
    || (strcmp(Py_TYPE(op)->tp_name,"SwigPyObject") == 0);
#endif
}

SWIGRUNTIME PyObject *
SwigPyObject_New(void *ptr, swig_type_info *ty, int own);

SWIGRUNTIME void
SwigPyObject_dealloc(PyObject *v)
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  PyObject *next = sobj->next;
  if (sobj->own == SWIG_POINTER_OWN) {
    swig_type_info *ty = sobj->ty;
    SwigPyClientData *data = ty ? (SwigPyClientData *) ty->clientdata : 0;
    PyObject *destroy = data ? data->destroy : 0;
    if (destroy) {
      /* destroy is always a VARARGS method */
      PyObject *res;

      /* PyObject_CallFunction() has the potential to silently drop
         the active exception.  In cases of unnamed temporary
         variable or where we just finished iterating over a generator
         StopIteration will be active right now, and this needs to
         remain true upon return from SwigPyObject_dealloc.  So save
         and restore. */
      
      PyObject *type = NULL, *value = NULL, *traceback = NULL;
      PyErr_Fetch(&type, &value, &traceback);

      if (data->delargs) {
        /* we need to create a temporary object to carry the destroy operation */
        PyObject *tmp = SwigPyObject_New(sobj->ptr, ty, 0);
        res = SWIG_Python_CallFunctor(destroy, tmp);
        Py_DECREF(tmp);
      } else {
        PyCFunction meth = PyCFunction_GET_FUNCTION(destroy);
        PyObject *mself = PyCFunction_GET_SELF(destroy);
        res = ((*meth)(mself, v));
      }
      if (!res)
        PyErr_WriteUnraisable(destroy);

      PyErr_Restore(type, value, traceback);

      Py_XDECREF(res);
    } 
#if !defined(SWIG_PYTHON_SILENT_MEMLEAK)
    else {
      const char *name = SWIG_TypePrettyName(ty);
      printf("swig/python detected a memory leak of type '%s', no destructor found.\n", (name ? name : "unknown"));
    }
#endif
  } 
  Py_XDECREF(next);
  PyObject_DEL(v);
}

SWIGRUNTIME PyObject* 
SwigPyObject_append(PyObject* v, PyObject* next)
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  if (!SwigPyObject_Check(next)) {
    PyErr_SetString(PyExc_TypeError, "Attempt to append a non SwigPyObject");
    return NULL;
  }
  sobj->next = next;
  Py_INCREF(next);
  return SWIG_Py_Void();
}

SWIGRUNTIME PyObject* 
SwigPyObject_next(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  if (sobj->next) {    
    Py_INCREF(sobj->next);
    return sobj->next;
  } else {
    return SWIG_Py_Void();
  }
}

SWIGINTERN PyObject*
SwigPyObject_disown(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *)v;
  sobj->own = 0;
  return SWIG_Py_Void();
}

SWIGINTERN PyObject*
SwigPyObject_acquire(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *)v;
  sobj->own = SWIG_POINTER_OWN;
  return SWIG_Py_Void();
}

SWIGINTERN PyObject*
SwigPyObject_own(PyObject *v, PyObject *args)
{
  PyObject *val = 0;
  if (!PyArg_UnpackTuple(args, "own", 0, 1, &val)) {
    return NULL;
  } else {
    SwigPyObject *sobj = (SwigPyObject *)v;
    PyObject *obj = PyBool_FromLong(sobj->own);
    if (val) {
      if (PyObject_IsTrue(val)) {
        SwigPyObject_acquire(v,args);
      } else {
        SwigPyObject_disown(v,args);
      }
    } 
    return obj;
  }
}

static PyMethodDef
swigobject_methods[] = {
  {"disown",  SwigPyObject_disown,  METH_NOARGS,  "releases ownership of the pointer"},
  {"acquire", SwigPyObject_acquire, METH_NOARGS,  "acquires ownership of the pointer"},
  {"own",     SwigPyObject_own,     METH_VARARGS, "returns/sets ownership of the pointer"},
  {"append",  SwigPyObject_append,  METH_O,       "appends another 'this' object"},
  {"next",    SwigPyObject_next,    METH_NOARGS,  "returns the next 'this' object"},
  {"__repr__",SwigPyObject_repr2,   METH_NOARGS,  "returns object representation"},
  {0, 0, 0, 0}  
};

SWIGRUNTIME PyTypeObject*
SwigPyObject_TypeOnce(void) {
  static char swigobject_doc[] = "Swig object carries a C/C++ instance pointer";

  static PyNumberMethods SwigPyObject_as_number = {
    (binaryfunc)0, /*nb_add*/
    (binaryfunc)0, /*nb_subtract*/
    (binaryfunc)0, /*nb_multiply*/
    /* nb_divide removed in Python 3 */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc)0, /*nb_divide*/
#endif
    (binaryfunc)0, /*nb_remainder*/
    (binaryfunc)0, /*nb_divmod*/
    (ternaryfunc)0,/*nb_power*/
    (unaryfunc)0,  /*nb_negative*/
    (unaryfunc)0,  /*nb_positive*/
    (unaryfunc)0,  /*nb_absolute*/
    (inquiry)0,    /*nb_nonzero*/
    0,		   /*nb_invert*/
    0,		   /*nb_lshift*/
    0,		   /*nb_rshift*/
    0,		   /*nb_and*/
    0,		   /*nb_xor*/
    0,		   /*nb_or*/
#if PY_VERSION_HEX < 0x03000000
    0,   /*nb_coerce*/
#endif
    (unaryfunc)SwigPyObject_long, /*nb_int*/
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc)SwigPyObject_long, /*nb_long*/
#else
    0, /*nb_reserved*/
#endif
    (unaryfunc)0,                 /*nb_float*/
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc)SwigPyObject_oct,  /*nb_oct*/
    (unaryfunc)SwigPyObject_hex,  /*nb_hex*/
#endif
#if PY_VERSION_HEX >= 0x03050000 /* 3.5 */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_inplace_matrix_multiply */
#elif PY_VERSION_HEX >= 0x03000000 /* 3.0 */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_index, nb_inplace_divide removed */
#else
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_index */
#endif
  };

  static PyTypeObject swigpyobject_type;
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp = {
#if PY_VERSION_HEX >= 0x03000000
      PyVarObject_HEAD_INIT(NULL, 0)
#else
      PyObject_HEAD_INIT(NULL)
      0,                                    /* ob_size */
#endif
      "SwigPyObject",                       /* tp_name */
      sizeof(SwigPyObject),                 /* tp_basicsize */
      0,                                    /* tp_itemsize */
      (destructor)SwigPyObject_dealloc,     /* tp_dealloc */
      0,                                    /* tp_print */
      (getattrfunc)0,                       /* tp_getattr */
      (setattrfunc)0,                       /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
      0, /* tp_reserved in 3.0.1, tp_compare in 3.0.0 but not used */
#else
      (cmpfunc)SwigPyObject_compare,        /* tp_compare */
#endif
      (reprfunc)SwigPyObject_repr,          /* tp_repr */
      &SwigPyObject_as_number,              /* tp_as_number */
      0,                                    /* tp_as_sequence */
      0,                                    /* tp_as_mapping */
      (hashfunc)0,                          /* tp_hash */
      (ternaryfunc)0,                       /* tp_call */
      0,                                    /* tp_str */
      PyObject_GenericGetAttr,              /* tp_getattro */
      0,                                    /* tp_setattro */
      0,                                    /* tp_as_buffer */
      Py_TPFLAGS_DEFAULT,                   /* tp_flags */
      swigobject_doc,                       /* tp_doc */
      0,                                    /* tp_traverse */
      0,                                    /* tp_clear */
      (richcmpfunc)SwigPyObject_richcompare,/* tp_richcompare */
      0,                                    /* tp_weaklistoffset */
      0,                                    /* tp_iter */
      0,                                    /* tp_iternext */
      swigobject_methods,                   /* tp_methods */
      0,                                    /* tp_members */
      0,                                    /* tp_getset */
      0,                                    /* tp_base */
      0,                                    /* tp_dict */
      0,                                    /* tp_descr_get */
      0,                                    /* tp_descr_set */
      0,                                    /* tp_dictoffset */
      0,                                    /* tp_init */
      0,                                    /* tp_alloc */
      0,                                    /* tp_new */
      0,                                    /* tp_free */
      0,                                    /* tp_is_gc */
      0,                                    /* tp_bases */
      0,                                    /* tp_mro */
      0,                                    /* tp_cache */
      0,                                    /* tp_subclasses */
      0,                                    /* tp_weaklist */
      0,                                    /* tp_del */
      0,                                    /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
      0,                                    /* tp_finalize */
#endif
#ifdef COUNT_ALLOCS
      0,                                    /* tp_allocs */
      0,                                    /* tp_frees */
      0,                                    /* tp_maxalloc */
      0,                                    /* tp_prev */
      0                                     /* tp_next */
#endif
    };
    swigpyobject_type = tmp;
    type_init = 1;
    if (PyType_Ready(&swigpyobject_type) < 0)
      return NULL;
  }
  return &swigpyobject_type;
}

SWIGRUNTIME PyObject *
SwigPyObject_New(void *ptr, swig_type_info *ty, int own)
{
  SwigPyObject *sobj = PyObject_NEW(SwigPyObject, SwigPyObject_type());
  if (sobj) {
    sobj->ptr  = ptr;
    sobj->ty   = ty;
    sobj->own  = own;
    sobj->next = 0;
  }
  return (PyObject *)sobj;
}

/* -----------------------------------------------------------------------------
 * Implements a simple Swig Packed type, and use it instead of string
 * ----------------------------------------------------------------------------- */

typedef struct {
  PyObject_HEAD
  void *pack;
  swig_type_info *ty;
  size_t size;
} SwigPyPacked;

SWIGRUNTIME PyObject *
SwigPyPacked_repr(SwigPyPacked *v)
{
  char result[SWIG_BUFFER_SIZE];
  if (SWIG_PackDataName(result, v->pack, v->size, 0, sizeof(result))) {
    return SWIG_Python_str_FromFormat("<Swig Packed at %s%s>", result, v->ty->name);
  } else {
    return SWIG_Python_str_FromFormat("<Swig Packed %s>", v->ty->name);
  }  
}

SWIGRUNTIME PyObject *
SwigPyPacked_str(SwigPyPacked *v)
{
  char result[SWIG_BUFFER_SIZE];
  if (SWIG_PackDataName(result, v->pack, v->size, 0, sizeof(result))){
    return SWIG_Python_str_FromFormat("%s%s", result, v->ty->name);
  } else {
    return SWIG_Python_str_FromChar(v->ty->name);
  }  
}

SWIGRUNTIME int
SwigPyPacked_compare(SwigPyPacked *v, SwigPyPacked *w)
{
  size_t i = v->size;
  size_t j = w->size;
  int s = (i < j) ? -1 : ((i > j) ? 1 : 0);
  return s ? s : strncmp((const char *)v->pack, (const char *)w->pack, 2*v->size);
}

SWIGRUNTIME PyTypeObject* SwigPyPacked_TypeOnce(void);

SWIGRUNTIME PyTypeObject*
SwigPyPacked_type(void) {
  static PyTypeObject *SWIG_STATIC_POINTER(type) = SwigPyPacked_TypeOnce();
  return type;
}

SWIGRUNTIMEINLINE int
SwigPyPacked_Check(PyObject *op) {
  return ((op)->ob_type == SwigPyPacked_TypeOnce()) 
    || (strcmp((op)->ob_type->tp_name,"SwigPyPacked") == 0);
}

SWIGRUNTIME void
SwigPyPacked_dealloc(PyObject *v)
{
  if (SwigPyPacked_Check(v)) {
    SwigPyPacked *sobj = (SwigPyPacked *) v;
    free(sobj->pack);
  }
  PyObject_DEL(v);
}

SWIGRUNTIME PyTypeObject*
SwigPyPacked_TypeOnce(void) {
  static char swigpacked_doc[] = "Swig object carries a C/C++ instance pointer";
  static PyTypeObject swigpypacked_type;
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp = {
#if PY_VERSION_HEX>=0x03000000
      PyVarObject_HEAD_INIT(NULL, 0)
#else
      PyObject_HEAD_INIT(NULL)
      0,                                    /* ob_size */
#endif
      "SwigPyPacked",                       /* tp_name */
      sizeof(SwigPyPacked),                 /* tp_basicsize */
      0,                                    /* tp_itemsize */
      (destructor)SwigPyPacked_dealloc,     /* tp_dealloc */
      0,                                    /* tp_print */
      (getattrfunc)0,                       /* tp_getattr */
      (setattrfunc)0,                       /* tp_setattr */
#if PY_VERSION_HEX>=0x03000000
      0, /* tp_reserved in 3.0.1 */
#else
      (cmpfunc)SwigPyPacked_compare,        /* tp_compare */
#endif
      (reprfunc)SwigPyPacked_repr,          /* tp_repr */
      0,                                    /* tp_as_number */
      0,                                    /* tp_as_sequence */
      0,                                    /* tp_as_mapping */
      (hashfunc)0,                          /* tp_hash */
      (ternaryfunc)0,                       /* tp_call */
      (reprfunc)SwigPyPacked_str,           /* tp_str */
      PyObject_GenericGetAttr,              /* tp_getattro */
      0,                                    /* tp_setattro */
      0,                                    /* tp_as_buffer */
      Py_TPFLAGS_DEFAULT,                   /* tp_flags */
      swigpacked_doc,                       /* tp_doc */
      0,                                    /* tp_traverse */
      0,                                    /* tp_clear */
      0,                                    /* tp_richcompare */
      0,                                    /* tp_weaklistoffset */
      0,                                    /* tp_iter */
      0,                                    /* tp_iternext */
      0,                                    /* tp_methods */
      0,                                    /* tp_members */
      0,                                    /* tp_getset */
      0,                                    /* tp_base */
      0,                                    /* tp_dict */
      0,                                    /* tp_descr_get */
      0,                                    /* tp_descr_set */
      0,                                    /* tp_dictoffset */
      0,                                    /* tp_init */
      0,                                    /* tp_alloc */
      0,                                    /* tp_new */
      0,                                    /* tp_free */
      0,                                    /* tp_is_gc */
      0,                                    /* tp_bases */
      0,                                    /* tp_mro */
      0,                                    /* tp_cache */
      0,                                    /* tp_subclasses */
      0,                                    /* tp_weaklist */
      0,                                    /* tp_del */
      0,                                    /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
      0,                                    /* tp_finalize */
#endif
#ifdef COUNT_ALLOCS
      0,                                    /* tp_allocs */
      0,                                    /* tp_frees */
      0,                                    /* tp_maxalloc */
      0,                                    /* tp_prev */
      0                                     /* tp_next */
#endif
    };
    swigpypacked_type = tmp;
    type_init = 1;
    if (PyType_Ready(&swigpypacked_type) < 0)
      return NULL;
  }
  return &swigpypacked_type;
}

SWIGRUNTIME PyObject *
SwigPyPacked_New(void *ptr, size_t size, swig_type_info *ty)
{
  SwigPyPacked *sobj = PyObject_NEW(SwigPyPacked, SwigPyPacked_type());
  if (sobj) {
    void *pack = malloc(size);
    if (pack) {
      memcpy(pack, ptr, size);
      sobj->pack = pack;
      sobj->ty   = ty;
      sobj->size = size;
    } else {
      PyObject_DEL((PyObject *) sobj);
      sobj = 0;
    }
  }
  return (PyObject *) sobj;
}

SWIGRUNTIME swig_type_info *
SwigPyPacked_UnpackData(PyObject *obj, void *ptr, size_t size)
{
  if (SwigPyPacked_Check(obj)) {
    SwigPyPacked *sobj = (SwigPyPacked *)obj;
    if (sobj->size != size) return 0;
    memcpy(ptr, sobj->pack, size);
    return sobj->ty;
  } else {
    return 0;
  }
}

/* -----------------------------------------------------------------------------
 * pointers/data manipulation
 * ----------------------------------------------------------------------------- */

static PyObject *Swig_This_global = NULL;

SWIGRUNTIME PyObject *
SWIG_This(void)
{
  if (Swig_This_global == NULL)
    Swig_This_global = SWIG_Python_str_FromChar("this");
  return Swig_This_global;
}

/* #define SWIG_PYTHON_SLOW_GETSET_THIS */

/* TODO: I don't know how to implement the fast getset in Python 3 right now */
#if PY_VERSION_HEX>=0x03000000
#define SWIG_PYTHON_SLOW_GETSET_THIS 
#endif

SWIGRUNTIME SwigPyObject *
SWIG_Python_GetSwigThis(PyObject *pyobj) 
{
  PyObject *obj;

  if (SwigPyObject_Check(pyobj))
    return (SwigPyObject *) pyobj;

#ifdef SWIGPYTHON_BUILTIN
  (void)obj;
# ifdef PyWeakref_CheckProxy
  if (PyWeakref_CheckProxy(pyobj)) {
    pyobj = PyWeakref_GET_OBJECT(pyobj);
    if (pyobj && SwigPyObject_Check(pyobj))
      return (SwigPyObject*) pyobj;
  }
# endif
  return NULL;
#else

  obj = 0;

#if !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
  if (PyInstance_Check(pyobj)) {
    obj = _PyInstance_Lookup(pyobj, SWIG_This());      
  } else {
    PyObject **dictptr = _PyObject_GetDictPtr(pyobj);
    if (dictptr != NULL) {
      PyObject *dict = *dictptr;
      obj = dict ? PyDict_GetItem(dict, SWIG_This()) : 0;
    } else {
#ifdef PyWeakref_CheckProxy
      if (PyWeakref_CheckProxy(pyobj)) {
	PyObject *wobj = PyWeakref_GET_OBJECT(pyobj);
	return wobj ? SWIG_Python_GetSwigThis(wobj) : 0;
      }
#endif
      obj = PyObject_GetAttr(pyobj,SWIG_This());
      if (obj) {
	Py_DECREF(obj);
      } else {
	if (PyErr_Occurred()) PyErr_Clear();
	return 0;
      }
    }
  }
#else
  obj = PyObject_GetAttr(pyobj,SWIG_This());
  if (obj) {
    Py_DECREF(obj);
  } else {
    if (PyErr_Occurred()) PyErr_Clear();
    return 0;
  }
#endif
  if (obj && !SwigPyObject_Check(obj)) {
    /* a PyObject is called 'this', try to get the 'real this'
       SwigPyObject from it */ 
    return SWIG_Python_GetSwigThis(obj);
  }
  return (SwigPyObject *)obj;
#endif
}

/* Acquire a pointer value */

SWIGRUNTIME int
SWIG_Python_AcquirePtr(PyObject *obj, int own) {
  if (own == SWIG_POINTER_OWN) {
    SwigPyObject *sobj = SWIG_Python_GetSwigThis(obj);
    if (sobj) {
      int oldown = sobj->own;
      sobj->own = own;
      return oldown;
    }
  }
  return 0;
}

/* Convert a pointer value */

SWIGRUNTIME int
SWIG_Python_ConvertPtrAndOwn(PyObject *obj, void **ptr, swig_type_info *ty, int flags, int *own) {
  int res;
  SwigPyObject *sobj;
  int implicit_conv = (flags & SWIG_POINTER_IMPLICIT_CONV) != 0;

  if (!obj)
    return SWIG_ERROR;
  if (obj == Py_None && !implicit_conv) {
    if (ptr)
      *ptr = 0;
    return (flags & SWIG_POINTER_NO_NULL) ? SWIG_NullReferenceError : SWIG_OK;
  }

  res = SWIG_ERROR;

  sobj = SWIG_Python_GetSwigThis(obj);
  if (own)
    *own = 0;
  while (sobj) {
    void *vptr = sobj->ptr;
    if (ty) {
      swig_type_info *to = sobj->ty;
      if (to == ty) {
        /* no type cast needed */
        if (ptr) *ptr = vptr;
        break;
      } else {
        swig_cast_info *tc = SWIG_TypeCheck(to->name,ty);
        if (!tc) {
          sobj = (SwigPyObject *)sobj->next;
        } else {
          if (ptr) {
            int newmemory = 0;
            *ptr = SWIG_TypeCast(tc,vptr,&newmemory);
            if (newmemory == SWIG_CAST_NEW_MEMORY) {
              assert(own); /* badly formed typemap which will lead to a memory leak - it must set and use own to delete *ptr */
              if (own)
                *own = *own | SWIG_CAST_NEW_MEMORY;
            }
          }
          break;
        }
      }
    } else {
      if (ptr) *ptr = vptr;
      break;
    }
  }
  if (sobj) {
    if (own)
      *own = *own | sobj->own;
    if (flags & SWIG_POINTER_DISOWN) {
      sobj->own = 0;
    }
    res = SWIG_OK;
  } else {
    if (implicit_conv) {
      SwigPyClientData *data = ty ? (SwigPyClientData *) ty->clientdata : 0;
      if (data && !data->implicitconv) {
        PyObject *klass = data->klass;
        if (klass) {
          PyObject *impconv;
          data->implicitconv = 1; /* avoid recursion and call 'explicit' constructors*/
          impconv = SWIG_Python_CallFunctor(klass, obj);
          data->implicitconv = 0;
          if (PyErr_Occurred()) {
            PyErr_Clear();
            impconv = 0;
          }
          if (impconv) {
            SwigPyObject *iobj = SWIG_Python_GetSwigThis(impconv);
            if (iobj) {
              void *vptr;
              res = SWIG_Python_ConvertPtrAndOwn((PyObject*)iobj, &vptr, ty, 0, 0);
              if (SWIG_IsOK(res)) {
                if (ptr) {
                  *ptr = vptr;
                  /* transfer the ownership to 'ptr' */
                  iobj->own = 0;
                  res = SWIG_AddCast(res);
                  res = SWIG_AddNewMask(res);
                } else {
                  res = SWIG_AddCast(res);		    
                }
              }
            }
            Py_DECREF(impconv);
          }
        }
      }
      if (!SWIG_IsOK(res) && obj == Py_None) {
        if (ptr)
          *ptr = 0;
        if (PyErr_Occurred())
          PyErr_Clear();
        res = SWIG_OK;
      }
    }
  }
  return res;
}

/* Convert a function ptr value */

SWIGRUNTIME int
SWIG_Python_ConvertFunctionPtr(PyObject *obj, void **ptr, swig_type_info *ty) {
  if (!PyCFunction_Check(obj)) {
    return SWIG_ConvertPtr(obj, ptr, ty, 0);
  } else {
    void *vptr = 0;
    swig_cast_info *tc;

    /* here we get the method pointer for callbacks */
    const char *doc = (((PyCFunctionObject *)obj) -> m_ml -> ml_doc);
    const char *desc = doc ? strstr(doc, "swig_ptr: ") : 0;
    if (desc)
      desc = ty ? SWIG_UnpackVoidPtr(desc + 10, &vptr, ty->name) : 0;
    if (!desc)
      return SWIG_ERROR;
    tc = SWIG_TypeCheck(desc,ty);
    if (tc) {
      int newmemory = 0;
      *ptr = SWIG_TypeCast(tc,vptr,&newmemory);
      assert(!newmemory); /* newmemory handling not yet implemented */
    } else {
      return SWIG_ERROR;
    }
    return SWIG_OK;
  }
}

/* Convert a packed pointer value */

SWIGRUNTIME int
SWIG_Python_ConvertPacked(PyObject *obj, void *ptr, size_t sz, swig_type_info *ty) {
  swig_type_info *to = SwigPyPacked_UnpackData(obj, ptr, sz);
  if (!to) return SWIG_ERROR;
  if (ty) {
    if (to != ty) {
      /* check type cast? */
      swig_cast_info *tc = SWIG_TypeCheck(to->name,ty);
      if (!tc) return SWIG_ERROR;
    }
  }
  return SWIG_OK;
}  

/* -----------------------------------------------------------------------------
 * Create a new pointer object
 * ----------------------------------------------------------------------------- */

/*
  Create a new instance object, without calling __init__, and set the
  'this' attribute.
*/

SWIGRUNTIME PyObject* 
SWIG_Python_NewShadowInstance(SwigPyClientData *data, PyObject *swig_this)
{
  PyObject *inst = 0;
  PyObject *newraw = data->newraw;
  if (newraw) {
    inst = PyObject_Call(newraw, data->newargs, NULL);
    if (inst) {
#if !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
      PyObject **dictptr = _PyObject_GetDictPtr(inst);
      if (dictptr != NULL) {
	PyObject *dict = *dictptr;
	if (dict == NULL) {
	  dict = PyDict_New();
	  *dictptr = dict;
	  PyDict_SetItem(dict, SWIG_This(), swig_this);
	}
      }
#else
      PyObject *key = SWIG_This();
      PyObject_SetAttr(inst, key, swig_this);
#endif
    }
  } else {
#if PY_VERSION_HEX >= 0x03000000
    PyObject *empty_args = PyTuple_New(0);
    if (empty_args) {
      PyObject *empty_kwargs = PyDict_New();
      if (empty_kwargs) {
        inst = ((PyTypeObject *)data->newargs)->tp_new((PyTypeObject *)data->newargs, empty_args, empty_kwargs);
        Py_DECREF(empty_kwargs);
        if (inst) {
          PyObject_SetAttr(inst, SWIG_This(), swig_this);
          Py_TYPE(inst)->tp_flags &= ~Py_TPFLAGS_VALID_VERSION_TAG;
        }
      }
      Py_DECREF(empty_args);
    }
#else
    PyObject *dict = PyDict_New();
    if (dict) {
      PyDict_SetItem(dict, SWIG_This(), swig_this);
      inst = PyInstance_NewRaw(data->newargs, dict);
      Py_DECREF(dict);
    }
#endif
  }
  return inst;
}

SWIGRUNTIME void
SWIG_Python_SetSwigThis(PyObject *inst, PyObject *swig_this)
{
 PyObject *dict;
#if !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
 PyObject **dictptr = _PyObject_GetDictPtr(inst);
 if (dictptr != NULL) {
   dict = *dictptr;
   if (dict == NULL) {
     dict = PyDict_New();
     *dictptr = dict;
   }
   PyDict_SetItem(dict, SWIG_This(), swig_this);
   return;
 }
#endif
 dict = PyObject_GetAttrString(inst, "__dict__");
 PyDict_SetItem(dict, SWIG_This(), swig_this);
 Py_DECREF(dict);
} 


SWIGINTERN PyObject *
SWIG_Python_InitShadowInstance(PyObject *args) {
  PyObject *obj[2];
  if (!SWIG_Python_UnpackTuple(args, "swiginit", 2, 2, obj)) {
    return NULL;
  } else {
    SwigPyObject *sthis = SWIG_Python_GetSwigThis(obj[0]);
    if (sthis) {
      SwigPyObject_append((PyObject*) sthis, obj[1]);
    } else {
      SWIG_Python_SetSwigThis(obj[0], obj[1]);
    }
    return SWIG_Py_Void();
  }
}

/* Create a new pointer object */

SWIGRUNTIME PyObject *
SWIG_Python_NewPointerObj(PyObject *self, void *ptr, swig_type_info *type, int flags) {
  SwigPyClientData *clientdata;
  PyObject * robj;
  int own;

  if (!ptr)
    return SWIG_Py_Void();

  clientdata = type ? (SwigPyClientData *)(type->clientdata) : 0;
  own = (flags & SWIG_POINTER_OWN) ? SWIG_POINTER_OWN : 0;
  if (clientdata && clientdata->pytype) {
    SwigPyObject *newobj;
    if (flags & SWIG_BUILTIN_TP_INIT) {
      newobj = (SwigPyObject*) self;
      if (newobj->ptr) {
        PyObject *next_self = clientdata->pytype->tp_alloc(clientdata->pytype, 0);
        while (newobj->next)
	  newobj = (SwigPyObject *) newobj->next;
        newobj->next = next_self;
        newobj = (SwigPyObject *)next_self;
#ifdef SWIGPYTHON_BUILTIN
        newobj->dict = 0;
#endif
      }
    } else {
      newobj = PyObject_New(SwigPyObject, clientdata->pytype);
#ifdef SWIGPYTHON_BUILTIN
      newobj->dict = 0;
#endif
    }
    if (newobj) {
      newobj->ptr = ptr;
      newobj->ty = type;
      newobj->own = own;
      newobj->next = 0;
      return (PyObject*) newobj;
    }
    return SWIG_Py_Void();
  }

  assert(!(flags & SWIG_BUILTIN_TP_INIT));

  robj = SwigPyObject_New(ptr, type, own);
  if (robj && clientdata && !(flags & SWIG_POINTER_NOSHADOW)) {
    PyObject *inst = SWIG_Python_NewShadowInstance(clientdata, robj);
    Py_DECREF(robj);
    robj = inst;
  }
  return robj;
}

/* Create a new packed object */

SWIGRUNTIMEINLINE PyObject *
SWIG_Python_NewPackedObj(void *ptr, size_t sz, swig_type_info *type) {
  return ptr ? SwigPyPacked_New((void *) ptr, sz, type) : SWIG_Py_Void();
}

/* -----------------------------------------------------------------------------*
 *  Get type list 
 * -----------------------------------------------------------------------------*/

#ifdef SWIG_LINK_RUNTIME
void *SWIG_ReturnGlobalTypeList(void *);
#endif

SWIGRUNTIME swig_module_info *
SWIG_Python_GetModule(void *SWIGUNUSEDPARM(clientdata)) {
  static void *type_pointer = (void *)0;
  /* first check if module already created */
  if (!type_pointer) {
#ifdef SWIG_LINK_RUNTIME
    type_pointer = SWIG_ReturnGlobalTypeList((void *)0);
#else
    type_pointer = PyCapsule_Import(SWIGPY_CAPSULE_NAME, 0);
    if (PyErr_Occurred()) {
      PyErr_Clear();
      type_pointer = (void *)0;
    }
#endif
  }
  return (swig_module_info *) type_pointer;
}

SWIGRUNTIME void
SWIG_Python_DestroyModule(PyObject *obj)
{
  swig_module_info *swig_module = (swig_module_info *) PyCapsule_GetPointer(obj, SWIGPY_CAPSULE_NAME);
  swig_type_info **types = swig_module->types;
  size_t i;
  for (i =0; i < swig_module->size; ++i) {
    swig_type_info *ty = types[i];
    if (ty->owndata) {
      SwigPyClientData *data = (SwigPyClientData *) ty->clientdata;
      if (data) SwigPyClientData_Del(data);
    }
  }
  Py_DECREF(SWIG_This());
  Swig_This_global = NULL;
}

SWIGRUNTIME void
SWIG_Python_SetModule(swig_module_info *swig_module) {
#if PY_VERSION_HEX >= 0x03000000
 /* Add a dummy module object into sys.modules */
  PyObject *module = PyImport_AddModule("swig_runtime_data" SWIG_RUNTIME_VERSION);
#else
  static PyMethodDef swig_empty_runtime_method_table[] = { {NULL, NULL, 0, NULL} }; /* Sentinel */
  PyObject *module = Py_InitModule("swig_runtime_data" SWIG_RUNTIME_VERSION, swig_empty_runtime_method_table);
#endif
  PyObject *pointer = PyCapsule_New((void *) swig_module, SWIGPY_CAPSULE_NAME, SWIG_Python_DestroyModule);
  if (pointer && module) {
    PyModule_AddObject(module, "type_pointer_capsule" SWIG_TYPE_TABLE_NAME, pointer);
  } else {
    Py_XDECREF(pointer);
  }
}

/* The python cached type query */
SWIGRUNTIME PyObject *
SWIG_Python_TypeCache(void) {
  static PyObject *SWIG_STATIC_POINTER(cache) = PyDict_New();
  return cache;
}

SWIGRUNTIME swig_type_info *
SWIG_Python_TypeQuery(const char *type)
{
  PyObject *cache = SWIG_Python_TypeCache();
  PyObject *key = SWIG_Python_str_FromChar(type); 
  PyObject *obj = PyDict_GetItem(cache, key);
  swig_type_info *descriptor;
  if (obj) {
    descriptor = (swig_type_info *) PyCapsule_GetPointer(obj, NULL);
  } else {
    swig_module_info *swig_module = SWIG_GetModule(0);
    descriptor = SWIG_TypeQueryModule(swig_module, swig_module, type);
    if (descriptor) {
      obj = PyCapsule_New((void*) descriptor, NULL, NULL);
      PyDict_SetItem(cache, key, obj);
      Py_DECREF(obj);
    }
  }
  Py_DECREF(key);
  return descriptor;
}

/* 
   For backward compatibility only
*/
#define SWIG_POINTER_EXCEPTION  0
#define SWIG_arg_fail(arg)      SWIG_Python_ArgFail(arg)
#define SWIG_MustGetPtr(p, type, argnum, flags)  SWIG_Python_MustGetPtr(p, type, argnum, flags)

SWIGRUNTIME int
SWIG_Python_AddErrMesg(const char* mesg, int infront)
{  
  if (PyErr_Occurred()) {
    PyObject *type = 0;
    PyObject *value = 0;
    PyObject *traceback = 0;
    PyErr_Fetch(&type, &value, &traceback);
    if (value) {
      PyObject *old_str = PyObject_Str(value);
      const char *tmp = SWIG_Python_str_AsChar(old_str);
      const char *errmesg = tmp ? tmp : "Invalid error message";
      Py_XINCREF(type);
      PyErr_Clear();
      if (infront) {
	PyErr_Format(type, "%s %s", mesg, errmesg);
      } else {
	PyErr_Format(type, "%s %s", errmesg, mesg);
      }
      SWIG_Python_str_DelForPy3(tmp);
      Py_DECREF(old_str);
    }
    return 1;
  } else {
    return 0;
  }
}
  
SWIGRUNTIME int
SWIG_Python_ArgFail(int argnum)
{
  if (PyErr_Occurred()) {
    /* add information about failing argument */
    char mesg[256];
    PyOS_snprintf(mesg, sizeof(mesg), "argument number %d:", argnum);
    return SWIG_Python_AddErrMesg(mesg, 1);
  } else {
    return 0;
  }
}

SWIGRUNTIMEINLINE const char *
SwigPyObject_GetDesc(PyObject *self)
{
  SwigPyObject *v = (SwigPyObject *)self;
  swig_type_info *ty = v ? v->ty : 0;
  return ty ? ty->str : "";
}

SWIGRUNTIME void
SWIG_Python_TypeError(const char *type, PyObject *obj)
{
  if (type) {
#if defined(SWIG_COBJECT_TYPES)
    if (obj && SwigPyObject_Check(obj)) {
      const char *otype = (const char *) SwigPyObject_GetDesc(obj);
      if (otype) {
	PyErr_Format(PyExc_TypeError, "a '%s' is expected, 'SwigPyObject(%s)' is received",
		     type, otype);
	return;
      }
    } else 
#endif      
    {
      const char *otype = (obj ? obj->ob_type->tp_name : 0); 
      if (otype) {
	PyObject *str = PyObject_Str(obj);
	const char *cstr = str ? SWIG_Python_str_AsChar(str) : 0;
	if (cstr) {
	  PyErr_Format(PyExc_TypeError, "a '%s' is expected, '%s(%s)' is received",
		       type, otype, cstr);
          SWIG_Python_str_DelForPy3(cstr);
	} else {
	  PyErr_Format(PyExc_TypeError, "a '%s' is expected, '%s' is received",
		       type, otype);
	}
	Py_XDECREF(str);
	return;
      }
    }   
    PyErr_Format(PyExc_TypeError, "a '%s' is expected", type);
  } else {
    PyErr_Format(PyExc_TypeError, "unexpected type is received");
  }
}


/* Convert a pointer value, signal an exception on a type mismatch */
SWIGRUNTIME void *
SWIG_Python_MustGetPtr(PyObject *obj, swig_type_info *ty, int SWIGUNUSEDPARM(argnum), int flags) {
  void *result;
  if (SWIG_Python_ConvertPtr(obj, &result, ty, flags) == -1) {
    PyErr_Clear();
#if SWIG_POINTER_EXCEPTION
    if (flags) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(ty), obj);
      SWIG_Python_ArgFail(argnum);
    }
#endif
  }
  return result;
}

#ifdef SWIGPYTHON_BUILTIN
SWIGRUNTIME int
SWIG_Python_NonDynamicSetAttr(PyObject *obj, PyObject *name, PyObject *value) {
  PyTypeObject *tp = obj->ob_type;
  PyObject *descr;
  PyObject *encoded_name;
  descrsetfunc f;
  int res = -1;

# ifdef Py_USING_UNICODE
  if (PyString_Check(name)) {
    name = PyUnicode_Decode(PyString_AsString(name), PyString_Size(name), NULL, NULL);
    if (!name)
      return -1;
  } else if (!PyUnicode_Check(name))
# else
  if (!PyString_Check(name))
# endif
  {
    PyErr_Format(PyExc_TypeError, "attribute name must be string, not '%.200s'", name->ob_type->tp_name);
    return -1;
  } else {
    Py_INCREF(name);
  }

  if (!tp->tp_dict) {
    if (PyType_Ready(tp) < 0)
      goto done;
  }

  descr = _PyType_Lookup(tp, name);
  f = NULL;
  if (descr != NULL)
    f = descr->ob_type->tp_descr_set;
  if (!f) {
    if (PyString_Check(name)) {
      encoded_name = name;
      Py_INCREF(name);
    } else {
      encoded_name = PyUnicode_AsUTF8String(name);
      if (!encoded_name)
        return -1;
    }
    PyErr_Format(PyExc_AttributeError, "'%.100s' object has no attribute '%.200s'", tp->tp_name, PyString_AsString(encoded_name));
    Py_DECREF(encoded_name);
  } else {
    res = f(descr, obj, value);
  }
  
  done:
  Py_DECREF(name);
  return res;
}
#endif


#ifdef __cplusplus
}
#endif



#define SWIG_exception_fail(code, msg) do { SWIG_Error(code, msg); SWIG_fail; } while(0) 

#define SWIG_contract_assert(expr, msg) if (!(expr)) { SWIG_Error(SWIG_RuntimeError, msg); SWIG_fail; } else 



#ifdef __cplusplus
extern "C" {
#endif

/* Method creation and docstring support functions */

SWIGINTERN PyMethodDef *SWIG_PythonGetProxyDoc(const char *name);
SWIGINTERN PyObject *SWIG_PyInstanceMethod_New(PyObject *SWIGUNUSEDPARM(self), PyObject *func);
SWIGINTERN PyObject *SWIG_PyStaticMethod_New(PyObject *SWIGUNUSEDPARM(self), PyObject *func);

#ifdef __cplusplus
}
#endif


/* -------- TYPES TABLE (BEGIN) -------- */

#define SWIGTYPE_p_char swig_types[0]
#define SWIGTYPE_p_double swig_types[1]
#define SWIGTYPE_p_helics_complex swig_types[2]
#define SWIGTYPE_p_helics_core_type swig_types[3]
#define SWIGTYPE_p_helics_data_type swig_types[4]
#define SWIGTYPE_p_helics_error swig_types[5]
#define SWIGTYPE_p_helics_error_types swig_types[6]
#define SWIGTYPE_p_helics_federate_flags swig_types[7]
#define SWIGTYPE_p_helics_federate_state swig_types[8]
#define SWIGTYPE_p_helics_filter_type swig_types[9]
#define SWIGTYPE_p_helics_handle_options swig_types[10]
#define SWIGTYPE_p_helics_iteration_request swig_types[11]
#define SWIGTYPE_p_helics_iteration_result swig_types[12]
#define SWIGTYPE_p_helics_log_levels swig_types[13]
#define SWIGTYPE_p_helics_message swig_types[14]
#define SWIGTYPE_p_helics_multi_input_mode swig_types[15]
#define SWIGTYPE_p_helics_properties swig_types[16]
#define SWIGTYPE_p_int swig_types[17]
#define SWIGTYPE_p_int16_t swig_types[18]
#define SWIGTYPE_p_int32_t swig_types[19]
#define SWIGTYPE_p_p_char swig_types[20]
#define SWIGTYPE_p_void swig_types[21]
static swig_type_info *swig_types[23];
static swig_module_info swig_module = {swig_types, 22, 0, 0, 0, 0};
#define SWIG_TypeQuery(name) SWIG_TypeQueryModule(&swig_module, &swig_module, name)
#define SWIG_MangledTypeQuery(name) SWIG_MangledTypeQueryModule(&swig_module, &swig_module, name)

/* -------- TYPES TABLE (END) -------- */

#ifdef SWIG_TypeQuery
# undef SWIG_TypeQuery
#endif
#define SWIG_TypeQuery SWIG_Python_TypeQuery

/*-----------------------------------------------
              @(target):= _helics.so
  ------------------------------------------------*/
#if PY_VERSION_HEX >= 0x03000000
#  define SWIG_init    PyInit__helics

#else
#  define SWIG_init    init_helics

#endif
#define SWIG_name    "_helics"

#define SWIGVERSION 0x040001 
#define SWIG_VERSION SWIGVERSION


#define SWIG_as_voidptr(a) (void *)((const void *)(a)) 
#define SWIG_as_voidptrptr(a) ((void)SWIG_as_voidptr(*a),(void**)(a)) 


#include "api-data.h"

static PyObject* pHelicsException;

/* throw a helics error */
static void throwHelicsPythonException(helics_error *err) {
  char str[256];
  switch (err->error_code)
  {
  case helics_ok:
    return;
  case helics_error_registration_failure:
    strcat(str, "helics:registration_failure - ");
    strcat(str, err->message);
    PyErr_SetString(pHelicsException, str);
    break;
  case   helics_error_connection_failure:
    strcat(str, "helics:connection_failure - ");
    strcat(str, err->message);
    PyErr_SetString(pHelicsException, str);
    break;
  case   helics_error_invalid_object:
    strcat(str, "helics:invalid_object - ");
    strcat(str, err->message);
    PyErr_SetString(pHelicsException, str);
    break;
  case   helics_error_invalid_argument:
    strcat(str, "helics:invalid_argument - ");
    strcat(str, err->message);
    PyErr_SetString(pHelicsException, str);
    break;
  case   helics_error_discard:
    strcat(str, "helics:discard - ");
    strcat(str, err->message);
    PyErr_SetString(pHelicsException, str);
    break;
  case helics_error_system_failure:
    strcat(str, "helics:system_failure - ");
    strcat(str, err->message);
    PyErr_SetString(pHelicsException, str);
    break;
  case   helics_error_invalid_state_transition:
    strcat(str, "helics:invalid_state_transition - ");
    strcat(str, err->message);
    PyErr_SetString(pHelicsException, str);
    break;
  case   helics_error_invalid_function_call:
    strcat(str, "helics:invalid_function_call - ");
    strcat(str, err->message);
    PyErr_SetString(pHelicsException, str);
    break;
  case   helics_error_execution_failure:
    strcat(str, "helics:execution_failure - ");
    strcat(str, err->message);
    PyErr_SetString(pHelicsException, str);
    break;
  case   helics_error_insufficient_space:
    strcat(str, "helics:insufficient_space - ");
    strcat(str, err->message);
    PyErr_SetString(pHelicsException, str);
    break;
  case   helics_error_other:
  case   helics_error_external_type:
  default:
    strcat(str, "helics:error - ");
    strcat(str, err->message);
    PyErr_SetString(pHelicsException, str);
    break;
  }
}



#include "ValueFederate.h"
#include "MessageFederate.h"
#include "MessageFilters.h"


SWIGINTERNINLINE PyObject*
  SWIG_From_int  (int value)
{
  return PyInt_FromLong((long) value);
}


  #define SWIG_From_double   PyFloat_FromDouble 


SWIGINTERN int
SWIG_AsVal_double (PyObject *obj, double *val)
{
  int res = SWIG_TypeError;
  if (PyFloat_Check(obj)) {
    if (val) *val = PyFloat_AsDouble(obj);
    return SWIG_OK;
#if PY_VERSION_HEX < 0x03000000
  } else if (PyInt_Check(obj)) {
    if (val) *val = (double) PyInt_AsLong(obj);
    return SWIG_OK;
#endif
  } else if (PyLong_Check(obj)) {
    double v = PyLong_AsDouble(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    double d = PyFloat_AsDouble(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = d;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      long v = PyLong_AsLong(obj);
      if (!PyErr_Occurred()) {
	if (val) *val = v;
	return SWIG_AddCast(SWIG_AddCast(SWIG_OK));
      } else {
	PyErr_Clear();
      }
    }
  }
#endif
  return res;
}


SWIGINTERN swig_type_info*
SWIG_pchar_descriptor(void)
{
  static int init = 0;
  static swig_type_info* info = 0;
  if (!init) {
    info = SWIG_TypeQuery("_p_char");
    init = 1;
  }
  return info;
}


SWIGINTERN int
SWIG_AsCharPtrAndSize(PyObject *obj, char** cptr, size_t* psize, int *alloc)
{
#if PY_VERSION_HEX>=0x03000000
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
  if (PyBytes_Check(obj))
#else
  if (PyUnicode_Check(obj))
#endif
#else  
  if (PyString_Check(obj))
#endif
  {
    char *cstr; Py_ssize_t len;
    int ret = SWIG_OK;
#if PY_VERSION_HEX>=0x03000000
#if !defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
    if (!alloc && cptr) {
        /* We can't allow converting without allocation, since the internal
           representation of string in Python 3 is UCS-2/UCS-4 but we require
           a UTF-8 representation.
           TODO(bhy) More detailed explanation */
        return SWIG_RuntimeError;
    }
    obj = PyUnicode_AsUTF8String(obj);
    if (!obj)
      return SWIG_TypeError;
    if (alloc)
      *alloc = SWIG_NEWOBJ;
#endif
    PyBytes_AsStringAndSize(obj, &cstr, &len);
#else
    PyString_AsStringAndSize(obj, &cstr, &len);
#endif
    if (cptr) {
      if (alloc) {
	if (*alloc == SWIG_NEWOBJ) {
	  *cptr = (char *)memcpy(malloc((len + 1)*sizeof(char)), cstr, sizeof(char)*(len + 1));
	  *alloc = SWIG_NEWOBJ;
	} else {
	  *cptr = cstr;
	  *alloc = SWIG_OLDOBJ;
	}
      } else {
#if PY_VERSION_HEX>=0x03000000
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
	*cptr = PyBytes_AsString(obj);
#else
	assert(0); /* Should never reach here with Unicode strings in Python 3 */
#endif
#else
	*cptr = SWIG_Python_str_AsChar(obj);
        if (!*cptr)
          ret = SWIG_TypeError;
#endif
      }
    }
    if (psize) *psize = len + 1;
#if PY_VERSION_HEX>=0x03000000 && !defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
    Py_XDECREF(obj);
#endif
    return ret;
  } else {
#if defined(SWIG_PYTHON_2_UNICODE)
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
#error "Cannot use both SWIG_PYTHON_2_UNICODE and SWIG_PYTHON_STRICT_BYTE_CHAR at once"
#endif
#if PY_VERSION_HEX<0x03000000
    if (PyUnicode_Check(obj)) {
      char *cstr; Py_ssize_t len;
      if (!alloc && cptr) {
        return SWIG_RuntimeError;
      }
      obj = PyUnicode_AsUTF8String(obj);
      if (!obj)
        return SWIG_TypeError;
      if (PyString_AsStringAndSize(obj, &cstr, &len) != -1) {
        if (cptr) {
          if (alloc) *alloc = SWIG_NEWOBJ;
          *cptr = (char *)memcpy(malloc((len + 1)*sizeof(char)), cstr, sizeof(char)*(len + 1));
        }
        if (psize) *psize = len + 1;

        Py_XDECREF(obj);
        return SWIG_OK;
      } else {
        Py_XDECREF(obj);
      }
    }
#endif
#endif

    swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
    if (pchar_descriptor) {
      void* vptr = 0;
      if (SWIG_ConvertPtr(obj, &vptr, pchar_descriptor, 0) == SWIG_OK) {
	if (cptr) *cptr = (char *) vptr;
	if (psize) *psize = vptr ? (strlen((char *)vptr) + 1) : 0;
	if (alloc) *alloc = SWIG_OLDOBJ;
	return SWIG_OK;
      }
    }
  }
  return SWIG_TypeError;
}





SWIGINTERNINLINE PyObject *
SWIG_FromCharPtrAndSize(const char* carray, size_t size)
{
  if (carray) {
    if (size > INT_MAX) {
      swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
      return pchar_descriptor ? 
	SWIG_InternalNewPointerObj((char *)(carray), pchar_descriptor, 0) : SWIG_Py_Void();
    } else {
#if PY_VERSION_HEX >= 0x03000000
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
      return PyBytes_FromStringAndSize(carray, (Py_ssize_t)(size));
#else
      return PyUnicode_DecodeUTF8(carray, (Py_ssize_t)(size), "surrogateescape");
#endif
#else
      return PyString_FromStringAndSize(carray, (Py_ssize_t)(size));
#endif
    }
  } else {
    return SWIG_Py_Void();
  }
}


SWIGINTERNINLINE PyObject * 
SWIG_FromCharPtr(const char *cptr)
{ 
  return SWIG_FromCharPtrAndSize(cptr, (cptr ? strlen(cptr) : 0));
}


#include <float.h>


#include <math.h>


SWIGINTERNINLINE int
SWIG_CanCastAsInteger(double *d, double min, double max) {
  double x = *d;
  if ((min <= x && x <= max)) {
   double fx = floor(x);
   double cx = ceil(x);
   double rd =  ((x - fx) < 0.5) ? fx : cx; /* simple rint */
   if ((errno == EDOM) || (errno == ERANGE)) {
     errno = 0;
   } else {
     double summ, reps, diff;
     if (rd < x) {
       diff = x - rd;
     } else if (rd > x) {
       diff = rd - x;
     } else {
       return 1;
     }
     summ = rd + x;
     reps = diff/summ;
     if (reps < 8*DBL_EPSILON) {
       *d = rd;
       return 1;
     }
   }
  }
  return 0;
}


SWIGINTERN int
SWIG_AsVal_long (PyObject *obj, long* val)
{
#if PY_VERSION_HEX < 0x03000000
  if (PyInt_Check(obj)) {
    if (val) *val = PyInt_AsLong(obj);
    return SWIG_OK;
  } else
#endif
  if (PyLong_Check(obj)) {
    long v = PyLong_AsLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      return SWIG_OverflowError;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    long v = PyInt_AsLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      double d;
      int res = SWIG_AddCast(SWIG_AsVal_double (obj,&d));
      if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, LONG_MIN, LONG_MAX)) {
	if (val) *val = (long)(d);
	return res;
      }
    }
  }
#endif
  return SWIG_TypeError;
}


#include <limits.h>
#if !defined(SWIG_NO_LLONG_MAX)
# if !defined(LLONG_MAX) && defined(__GNUC__) && defined (__LONG_LONG_MAX__)
#   define LLONG_MAX __LONG_LONG_MAX__
#   define LLONG_MIN (-LLONG_MAX - 1LL)
#   define ULLONG_MAX (LLONG_MAX * 2ULL + 1ULL)
# endif
#endif


#if defined(LLONG_MAX) && !defined(SWIG_LONG_LONG_AVAILABLE)
#  define SWIG_LONG_LONG_AVAILABLE
#endif


#ifdef SWIG_LONG_LONG_AVAILABLE
SWIGINTERN int
SWIG_AsVal_long_SS_long (PyObject *obj, long long *val)
{
  int res = SWIG_TypeError;
  if (PyLong_Check(obj)) {
    long long v = PyLong_AsLongLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      res = SWIG_OverflowError;
    }
  } else {
    long v;
    res = SWIG_AsVal_long (obj,&v);
    if (SWIG_IsOK(res)) {
      if (val) *val = v;
      return res;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    const double mant_max = 1LL << DBL_MANT_DIG;
    const double mant_min = -mant_max;
    double d;
    res = SWIG_AsVal_double (obj,&d);
    if (SWIG_IsOK(res) && !SWIG_CanCastAsInteger(&d, mant_min, mant_max))
      return SWIG_OverflowError;
    if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, mant_min, mant_max)) {
      if (val) *val = (long long)(d);
      return SWIG_AddCast(res);
    }
    res = SWIG_TypeError;
  }
#endif
  return res;
}
#endif


#ifdef SWIG_LONG_LONG_AVAILABLE
SWIGINTERNINLINE PyObject* 
SWIG_From_long_SS_long  (long long value)
{
  return ((value < LONG_MIN) || (value > LONG_MAX)) ?
    PyLong_FromLongLong(value) : PyInt_FromLong((long)(value));
}
#endif


SWIGINTERN int
SWIG_AsVal_int (PyObject * obj, int *val)
{
  long v;
  int res = SWIG_AsVal_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v < INT_MIN || v > INT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (int)(v);
    }
  }  
  return res;
}


SWIGINTERN int
SWIG_AsCharArray(PyObject * obj, char *val, size_t size)
{ 
  char* cptr = 0; size_t csize = 0; int alloc = SWIG_OLDOBJ;
  int res = SWIG_AsCharPtrAndSize(obj, &cptr, &csize, &alloc);
  if (SWIG_IsOK(res)) {
    /* special case of single char conversion when we don't need space for NUL */
    if (size == 1 && csize == 2 && cptr && !cptr[1]) --csize;
    if (csize <= size) {
      if (val) {
	if (csize) memcpy(val, cptr, csize*sizeof(char));
	if (csize < size) memset(val + csize, 0, (size - csize)*sizeof(char));
      }
      if (alloc == SWIG_NEWOBJ) {
	free((char*)cptr);
	res = SWIG_DelNewMask(res);
      }      
      return res;
    }
    if (alloc == SWIG_NEWOBJ) free((char*)cptr);
  }
  return SWIG_TypeError;
}


SWIGINTERN int
SWIG_AsVal_char (PyObject * obj, char *val)
{    
  int res = SWIG_AsCharArray(obj, val, 1);
  if (!SWIG_IsOK(res)) {
    long v;
    res = SWIG_AddCast(SWIG_AsVal_long (obj, &v));
    if (SWIG_IsOK(res)) {
      if ((CHAR_MIN <= v) && (v <= CHAR_MAX)) {
	if (val) *val = (char)(v);
      } else {
	res = SWIG_OverflowError;
      }
    }
  }
  return res;
}


SWIGINTERNINLINE PyObject *
SWIG_From_char  (char c) 
{ 
  return SWIG_FromCharPtrAndSize(&c,1);
}

#ifdef __cplusplus
extern "C" {
#endif
SWIGINTERN int Swig_var_helics_time_zero_set(PyObject *_val SWIGUNUSED) {
  SWIG_Error(SWIG_AttributeError,"Variable helics_time_zero is read-only.");
  return 1;
}


SWIGINTERN PyObject *Swig_var_helics_time_zero_get(void) {
  PyObject *pyobj = 0;
  
  pyobj = SWIG_From_double((double)(helics_time_zero));
  return pyobj;
}


SWIGINTERN int Swig_var_helics_time_epsilon_set(PyObject *_val SWIGUNUSED) {
  SWIG_Error(SWIG_AttributeError,"Variable helics_time_epsilon is read-only.");
  return 1;
}


SWIGINTERN PyObject *Swig_var_helics_time_epsilon_get(void) {
  PyObject *pyobj = 0;
  
  pyobj = SWIG_From_double((double)(helics_time_epsilon));
  return pyobj;
}


SWIGINTERN int Swig_var_helics_time_invalid_set(PyObject *_val SWIGUNUSED) {
  SWIG_Error(SWIG_AttributeError,"Variable helics_time_invalid is read-only.");
  return 1;
}


SWIGINTERN PyObject *Swig_var_helics_time_invalid_get(void) {
  PyObject *pyobj = 0;
  
  pyobj = SWIG_From_double((double)(helics_time_invalid));
  return pyobj;
}


SWIGINTERN int Swig_var_helics_time_maxtime_set(PyObject *_val SWIGUNUSED) {
  SWIG_Error(SWIG_AttributeError,"Variable helics_time_maxtime is read-only.");
  return 1;
}


SWIGINTERN PyObject *Swig_var_helics_time_maxtime_get(void) {
  PyObject *pyobj = 0;
  
  pyobj = SWIG_From_double((double)(helics_time_maxtime));
  return pyobj;
}


SWIGINTERN int Swig_var_helics_true_set(PyObject *_val SWIGUNUSED) {
  SWIG_Error(SWIG_AttributeError,"Variable helics_true is read-only.");
  return 1;
}


SWIGINTERN PyObject *Swig_var_helics_true_get(void) {
  PyObject *pyobj = 0;
  
  pyobj = SWIG_From_int((int)(helics_true));
  return pyobj;
}


SWIGINTERN int Swig_var_helics_false_set(PyObject *_val SWIGUNUSED) {
  SWIG_Error(SWIG_AttributeError,"Variable helics_false is read-only.");
  return 1;
}


SWIGINTERN PyObject *Swig_var_helics_false_get(void) {
  PyObject *pyobj = 0;
  
  pyobj = SWIG_From_int((int)(helics_false));
  return pyobj;
}


SWIGINTERN PyObject *_wrap_helics_complex_real_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct helics_complex *arg1 = (struct helics_complex *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!SWIG_Python_UnpackTuple(args, "helics_complex_real_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_helics_complex, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helics_complex_real_set" "', argument " "1"" of type '" "struct helics_complex *""'"); 
  }
  arg1 = (struct helics_complex *)(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helics_complex_real_set" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = (double)(val2);
  if (arg1) (arg1)->real = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helics_complex_real_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct helics_complex *arg1 = (struct helics_complex *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_helics_complex, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helics_complex_real_get" "', argument " "1"" of type '" "struct helics_complex *""'"); 
  }
  arg1 = (struct helics_complex *)(argp1);
  result = (double) ((arg1)->real);
  resultobj = SWIG_From_double((double)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helics_complex_imag_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct helics_complex *arg1 = (struct helics_complex *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!SWIG_Python_UnpackTuple(args, "helics_complex_imag_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_helics_complex, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helics_complex_imag_set" "', argument " "1"" of type '" "struct helics_complex *""'"); 
  }
  arg1 = (struct helics_complex *)(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helics_complex_imag_set" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = (double)(val2);
  if (arg1) (arg1)->imag = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helics_complex_imag_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct helics_complex *arg1 = (struct helics_complex *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_helics_complex, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helics_complex_imag_get" "', argument " "1"" of type '" "struct helics_complex *""'"); 
  }
  arg1 = (struct helics_complex *)(argp1);
  result = (double) ((arg1)->imag);
  resultobj = SWIG_From_double((double)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_helics_complex(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct helics_complex *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "new_helics_complex", 0, 0, 0)) SWIG_fail;
  result = (struct helics_complex *)calloc(1, sizeof(struct helics_complex));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_helics_complex, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_helics_complex(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct helics_complex *arg1 = (struct helics_complex *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_helics_complex, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_helics_complex" "', argument " "1"" of type '" "struct helics_complex *""'"); 
  }
  arg1 = (struct helics_complex *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *helics_complex_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_helics_complex, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *helics_complex_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_helics_message_time_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct helics_message *arg1 = (struct helics_message *) 0 ;
  helics_time arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!SWIG_Python_UnpackTuple(args, "helics_message_time_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_helics_message, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helics_message_time_set" "', argument " "1"" of type '" "struct helics_message *""'"); 
  }
  arg1 = (struct helics_message *)(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helics_message_time_set" "', argument " "2"" of type '" "helics_time""'");
  } 
  arg2 = (helics_time)(val2);
  if (arg1) (arg1)->time = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helics_message_time_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct helics_message *arg1 = (struct helics_message *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  helics_time result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_helics_message, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helics_message_time_get" "', argument " "1"" of type '" "struct helics_message *""'"); 
  }
  arg1 = (struct helics_message *)(argp1);
  result = (helics_time) ((arg1)->time);
  resultobj = SWIG_From_double((double)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helics_message_data_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct helics_message *arg1 = (struct helics_message *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!SWIG_Python_UnpackTuple(args, "helics_message_data_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_helics_message, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helics_message_data_set" "', argument " "1"" of type '" "struct helics_message *""'"); 
  }
  arg1 = (struct helics_message *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helics_message_data_set" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (arg2) {
    size_t size = strlen((const char *)((const char *)(arg2))) + 1;
    arg1->data = (char const *)(char *)memcpy(malloc((size)*sizeof(char)), arg2, sizeof(char)*(size));
  } else {
    arg1->data = 0;
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_helics_message_data_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct helics_message *arg1 = (struct helics_message *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_helics_message, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helics_message_data_get" "', argument " "1"" of type '" "struct helics_message *""'"); 
  }
  arg1 = (struct helics_message *)(argp1);
  result = (char *) ((arg1)->data);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helics_message_length_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct helics_message *arg1 = (struct helics_message *) 0 ;
  int64_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!SWIG_Python_UnpackTuple(args, "helics_message_length_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_helics_message, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helics_message_length_set" "', argument " "1"" of type '" "struct helics_message *""'"); 
  }
  arg1 = (struct helics_message *)(argp1);
  ecode2 = SWIG_AsVal_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helics_message_length_set" "', argument " "2"" of type '" "int64_t""'");
  } 
  arg2 = (int64_t)(val2);
  if (arg1) (arg1)->length = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helics_message_length_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct helics_message *arg1 = (struct helics_message *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int64_t result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_helics_message, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helics_message_length_get" "', argument " "1"" of type '" "struct helics_message *""'"); 
  }
  arg1 = (struct helics_message *)(argp1);
  result =  ((arg1)->length);
  resultobj = SWIG_From_long_SS_long((long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helics_message_messageID_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct helics_message *arg1 = (struct helics_message *) 0 ;
  int32_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!SWIG_Python_UnpackTuple(args, "helics_message_messageID_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_helics_message, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helics_message_messageID_set" "', argument " "1"" of type '" "struct helics_message *""'"); 
  }
  arg1 = (struct helics_message *)(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_int32_t,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helics_message_messageID_set" "', argument " "2"" of type '" "int32_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "helics_message_messageID_set" "', argument " "2"" of type '" "int32_t""'");
    } else {
      arg2 = *((int32_t *)(argp2));
    }
  }
  if (arg1) (arg1)->messageID = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helics_message_messageID_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct helics_message *arg1 = (struct helics_message *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int32_t result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_helics_message, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helics_message_messageID_get" "', argument " "1"" of type '" "struct helics_message *""'"); 
  }
  arg1 = (struct helics_message *)(argp1);
  result =  ((arg1)->messageID);
  resultobj = SWIG_NewPointerObj((int32_t *)memcpy((int32_t *)calloc(1,sizeof(int32_t)),&result,sizeof(int32_t)), SWIGTYPE_p_int32_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helics_message_flags_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct helics_message *arg1 = (struct helics_message *) 0 ;
  int16_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!SWIG_Python_UnpackTuple(args, "helics_message_flags_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_helics_message, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helics_message_flags_set" "', argument " "1"" of type '" "struct helics_message *""'"); 
  }
  arg1 = (struct helics_message *)(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_int16_t,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helics_message_flags_set" "', argument " "2"" of type '" "int16_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "helics_message_flags_set" "', argument " "2"" of type '" "int16_t""'");
    } else {
      arg2 = *((int16_t *)(argp2));
    }
  }
  if (arg1) (arg1)->flags = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helics_message_flags_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct helics_message *arg1 = (struct helics_message *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int16_t result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_helics_message, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helics_message_flags_get" "', argument " "1"" of type '" "struct helics_message *""'"); 
  }
  arg1 = (struct helics_message *)(argp1);
  result =  ((arg1)->flags);
  resultobj = SWIG_NewPointerObj((int16_t *)memcpy((int16_t *)calloc(1,sizeof(int16_t)),&result,sizeof(int16_t)), SWIGTYPE_p_int16_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helics_message_original_source_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct helics_message *arg1 = (struct helics_message *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!SWIG_Python_UnpackTuple(args, "helics_message_original_source_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_helics_message, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helics_message_original_source_set" "', argument " "1"" of type '" "struct helics_message *""'"); 
  }
  arg1 = (struct helics_message *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helics_message_original_source_set" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (arg2) {
    size_t size = strlen((const char *)((const char *)(arg2))) + 1;
    arg1->original_source = (char const *)(char *)memcpy(malloc((size)*sizeof(char)), arg2, sizeof(char)*(size));
  } else {
    arg1->original_source = 0;
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_helics_message_original_source_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct helics_message *arg1 = (struct helics_message *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_helics_message, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helics_message_original_source_get" "', argument " "1"" of type '" "struct helics_message *""'"); 
  }
  arg1 = (struct helics_message *)(argp1);
  result = (char *) ((arg1)->original_source);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helics_message_source_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct helics_message *arg1 = (struct helics_message *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!SWIG_Python_UnpackTuple(args, "helics_message_source_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_helics_message, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helics_message_source_set" "', argument " "1"" of type '" "struct helics_message *""'"); 
  }
  arg1 = (struct helics_message *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helics_message_source_set" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (arg2) {
    size_t size = strlen((const char *)((const char *)(arg2))) + 1;
    arg1->source = (char const *)(char *)memcpy(malloc((size)*sizeof(char)), arg2, sizeof(char)*(size));
  } else {
    arg1->source = 0;
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_helics_message_source_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct helics_message *arg1 = (struct helics_message *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_helics_message, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helics_message_source_get" "', argument " "1"" of type '" "struct helics_message *""'"); 
  }
  arg1 = (struct helics_message *)(argp1);
  result = (char *) ((arg1)->source);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helics_message_dest_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct helics_message *arg1 = (struct helics_message *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!SWIG_Python_UnpackTuple(args, "helics_message_dest_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_helics_message, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helics_message_dest_set" "', argument " "1"" of type '" "struct helics_message *""'"); 
  }
  arg1 = (struct helics_message *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helics_message_dest_set" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (arg2) {
    size_t size = strlen((const char *)((const char *)(arg2))) + 1;
    arg1->dest = (char const *)(char *)memcpy(malloc((size)*sizeof(char)), arg2, sizeof(char)*(size));
  } else {
    arg1->dest = 0;
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_helics_message_dest_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct helics_message *arg1 = (struct helics_message *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_helics_message, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helics_message_dest_get" "', argument " "1"" of type '" "struct helics_message *""'"); 
  }
  arg1 = (struct helics_message *)(argp1);
  result = (char *) ((arg1)->dest);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helics_message_original_dest_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct helics_message *arg1 = (struct helics_message *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!SWIG_Python_UnpackTuple(args, "helics_message_original_dest_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_helics_message, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helics_message_original_dest_set" "', argument " "1"" of type '" "struct helics_message *""'"); 
  }
  arg1 = (struct helics_message *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helics_message_original_dest_set" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (arg2) {
    size_t size = strlen((const char *)((const char *)(arg2))) + 1;
    arg1->original_dest = (char const *)(char *)memcpy(malloc((size)*sizeof(char)), arg2, sizeof(char)*(size));
  } else {
    arg1->original_dest = 0;
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_helics_message_original_dest_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct helics_message *arg1 = (struct helics_message *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_helics_message, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helics_message_original_dest_get" "', argument " "1"" of type '" "struct helics_message *""'"); 
  }
  arg1 = (struct helics_message *)(argp1);
  result = (char *) ((arg1)->original_dest);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_helics_message(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct helics_message *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "new_helics_message", 0, 0, 0)) SWIG_fail;
  result = (struct helics_message *)calloc(1, sizeof(struct helics_message));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_helics_message, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_helics_message(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  struct helics_message *arg1 = (struct helics_message *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_helics_message, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_helics_message" "', argument " "1"" of type '" "struct helics_message *""'"); 
  }
  arg1 = (struct helics_message *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *helics_message_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_helics_message, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *helics_message_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_helicsGetVersion(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "helicsGetVersion", 0, 0, 0)) SWIG_fail;
  result = (char *)helicsGetVersion();
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsGetBuildFlags(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "helicsGetBuildFlags", 0, 0, 0)) SWIG_fail;
  result = (char *)helicsGetBuildFlags();
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsGetCompilerVersion(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "helicsGetCompilerVersion", 0, 0, 0)) SWIG_fail;
  result = (char *)helicsGetCompilerVersion();
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsIsCoreTypeAvailable(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  PyObject *swig_obj[1] ;
  helics_bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsIsCoreTypeAvailable" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = (char *)(buf1);
  result = (helics_bool)helicsIsCoreTypeAvailable((char const *)arg1);
  resultobj = SWIG_From_int((int)(result));
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return resultobj;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCreateCore(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  helics_core result;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsCreateCore", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCreateCore" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsCreateCore" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "helicsCreateCore" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  result = (helics_core)helicsCreateCore((char const *)arg1,(char const *)arg2,(char const *)arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCreateCoreFromArgs(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  int arg3 ;
  char **arg4 = (char **) 0 ;
  helics_error *arg5 = (helics_error *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp5 ;
  PyObject *swig_obj[3] ;
  helics_core result;
  
  {
    etemp5=helicsErrorInitialize();
    arg5=&etemp5;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsCreateCoreFromArgs", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCreateCoreFromArgs" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsCreateCoreFromArgs" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  {
    /* Check if is a list */
    if (PyList_Check(swig_obj[2])) {
      int i;
      arg3 = (int)(PyList_Size(swig_obj[2]));
      arg4 = (char **) malloc((arg3+1)*sizeof(char *));
      for (i = 0; i < arg3; i++) {
        PyObject *o = PyList_GetItem(swig_obj[2],i);
        if (PyString_Check(o))
        arg4[i] = PyString_AsString(PyList_GetItem(swig_obj[2],i));
        else {
          PyErr_SetString(PyExc_TypeError,"list must contain strings");
          free(arg4);
          return NULL;
        }
      }
      arg4[i] = 0;
    } else {
      PyErr_SetString(PyExc_TypeError,"not a list");
      return NULL;
    }
  }
  result = (helics_core)helicsCreateCoreFromArgs((char const *)arg1,(char const *)arg2,arg3,(char const *const *)arg4,arg5);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    free((char *) arg4);
  }
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    free((char *) arg4);
  }
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCoreClone(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_core arg1 = (helics_core) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  helics_core result;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCoreClone" "', argument " "1"" of type '" "helics_core""'"); 
  }
  result = (helics_core)helicsCoreClone(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCoreIsValid(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_core arg1 = (helics_core) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  helics_bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCoreIsValid" "', argument " "1"" of type '" "helics_core""'"); 
  }
  result = (helics_bool)helicsCoreIsValid(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCreateBroker(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  helics_broker result;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsCreateBroker", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCreateBroker" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsCreateBroker" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "helicsCreateBroker" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  result = (helics_broker)helicsCreateBroker((char const *)arg1,(char const *)arg2,(char const *)arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCreateBrokerFromArgs(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  int arg3 ;
  char **arg4 = (char **) 0 ;
  helics_error *arg5 = (helics_error *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp5 ;
  PyObject *swig_obj[3] ;
  helics_broker result;
  
  {
    etemp5=helicsErrorInitialize();
    arg5=&etemp5;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsCreateBrokerFromArgs", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCreateBrokerFromArgs" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsCreateBrokerFromArgs" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  {
    /* Check if is a list */
    if (PyList_Check(swig_obj[2])) {
      int i;
      arg3 = (int)(PyList_Size(swig_obj[2]));
      arg4 = (char **) malloc((arg3+1)*sizeof(char *));
      for (i = 0; i < arg3; i++) {
        PyObject *o = PyList_GetItem(swig_obj[2],i);
        if (PyString_Check(o))
        arg4[i] = PyString_AsString(PyList_GetItem(swig_obj[2],i));
        else {
          PyErr_SetString(PyExc_TypeError,"list must contain strings");
          free(arg4);
          return NULL;
        }
      }
      arg4[i] = 0;
    } else {
      PyErr_SetString(PyExc_TypeError,"not a list");
      return NULL;
    }
  }
  result = (helics_broker)helicsCreateBrokerFromArgs((char const *)arg1,(char const *)arg2,arg3,(char const *const *)arg4,arg5);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    free((char *) arg4);
  }
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    free((char *) arg4);
  }
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsBrokerClone(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_broker arg1 = (helics_broker) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  helics_broker result;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsBrokerClone" "', argument " "1"" of type '" "helics_broker""'"); 
  }
  result = (helics_broker)helicsBrokerClone(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsBrokerIsValid(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_broker arg1 = (helics_broker) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  helics_bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsBrokerIsValid" "', argument " "1"" of type '" "helics_broker""'"); 
  }
  result = (helics_bool)helicsBrokerIsValid(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsBrokerIsConnected(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_broker arg1 = (helics_broker) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  helics_bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsBrokerIsConnected" "', argument " "1"" of type '" "helics_broker""'"); 
  }
  result = (helics_bool)helicsBrokerIsConnected(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsBrokerDataLink(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_broker arg1 = (helics_broker) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsBrokerDataLink", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsBrokerDataLink" "', argument " "1"" of type '" "helics_broker""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsBrokerDataLink" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "helicsBrokerDataLink" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  helicsBrokerDataLink(arg1,(char const *)arg2,(char const *)arg3,arg4);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsBrokerAddSourceFilterToEndpoint(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_broker arg1 = (helics_broker) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsBrokerAddSourceFilterToEndpoint", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsBrokerAddSourceFilterToEndpoint" "', argument " "1"" of type '" "helics_broker""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsBrokerAddSourceFilterToEndpoint" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "helicsBrokerAddSourceFilterToEndpoint" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  helicsBrokerAddSourceFilterToEndpoint(arg1,(char const *)arg2,(char const *)arg3,arg4);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsBrokerAddDestinationFilterToEndpoint(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_broker arg1 = (helics_broker) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsBrokerAddDestinationFilterToEndpoint", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsBrokerAddDestinationFilterToEndpoint" "', argument " "1"" of type '" "helics_broker""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsBrokerAddDestinationFilterToEndpoint" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "helicsBrokerAddDestinationFilterToEndpoint" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  helicsBrokerAddDestinationFilterToEndpoint(arg1,(char const *)arg2,(char const *)arg3,arg4);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsBrokerMakeConnections(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_broker arg1 = (helics_broker) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsBrokerMakeConnections", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsBrokerMakeConnections" "', argument " "1"" of type '" "helics_broker""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsBrokerMakeConnections" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsBrokerMakeConnections(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCoreWaitForDisconnect(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_core arg1 = (helics_core) 0 ;
  int arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  helics_bool result;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsCoreWaitForDisconnect", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCoreWaitForDisconnect" "', argument " "1"" of type '" "helics_core""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsCoreWaitForDisconnect" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  result = (helics_bool)helicsCoreWaitForDisconnect(arg1,arg2,arg3);
  resultobj = SWIG_From_int((int)(result));
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsBrokerWaitForDisconnect(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_broker arg1 = (helics_broker) 0 ;
  int arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  helics_bool result;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsBrokerWaitForDisconnect", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsBrokerWaitForDisconnect" "', argument " "1"" of type '" "helics_broker""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsBrokerWaitForDisconnect" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  result = (helics_bool)helicsBrokerWaitForDisconnect(arg1,arg2,arg3);
  resultobj = SWIG_From_int((int)(result));
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCoreIsConnected(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_core arg1 = (helics_core) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  helics_bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCoreIsConnected" "', argument " "1"" of type '" "helics_core""'"); 
  }
  result = (helics_bool)helicsCoreIsConnected(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCoreDataLink(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_core arg1 = (helics_core) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsCoreDataLink", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCoreDataLink" "', argument " "1"" of type '" "helics_core""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsCoreDataLink" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "helicsCoreDataLink" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  helicsCoreDataLink(arg1,(char const *)arg2,(char const *)arg3,arg4);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCoreAddSourceFilterToEndpoint(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_core arg1 = (helics_core) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsCoreAddSourceFilterToEndpoint", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCoreAddSourceFilterToEndpoint" "', argument " "1"" of type '" "helics_core""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsCoreAddSourceFilterToEndpoint" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "helicsCoreAddSourceFilterToEndpoint" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  helicsCoreAddSourceFilterToEndpoint(arg1,(char const *)arg2,(char const *)arg3,arg4);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCoreAddDestinationFilterToEndpoint(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_core arg1 = (helics_core) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsCoreAddDestinationFilterToEndpoint", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCoreAddDestinationFilterToEndpoint" "', argument " "1"" of type '" "helics_core""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsCoreAddDestinationFilterToEndpoint" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "helicsCoreAddDestinationFilterToEndpoint" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  helicsCoreAddDestinationFilterToEndpoint(arg1,(char const *)arg2,(char const *)arg3,arg4);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCoreMakeConnections(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_core arg1 = (helics_core) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsCoreMakeConnections", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCoreMakeConnections" "', argument " "1"" of type '" "helics_core""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsCoreMakeConnections" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsCoreMakeConnections(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsBrokerGetIdentifier(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_broker arg1 = (helics_broker) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsBrokerGetIdentifier" "', argument " "1"" of type '" "helics_broker""'"); 
  }
  result = (char *)helicsBrokerGetIdentifier(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCoreGetIdentifier(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_core arg1 = (helics_core) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCoreGetIdentifier" "', argument " "1"" of type '" "helics_core""'"); 
  }
  result = (char *)helicsCoreGetIdentifier(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsBrokerGetAddress(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_broker arg1 = (helics_broker) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsBrokerGetAddress" "', argument " "1"" of type '" "helics_broker""'"); 
  }
  result = (char *)helicsBrokerGetAddress(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCoreGetAddress(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_core arg1 = (helics_core) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCoreGetAddress" "', argument " "1"" of type '" "helics_core""'"); 
  }
  result = (char *)helicsCoreGetAddress(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCoreSetReadyToInit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_core arg1 = (helics_core) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCoreSetReadyToInit" "', argument " "1"" of type '" "helics_core""'"); 
  }
  helicsCoreSetReadyToInit(arg1,arg2);
  resultobj = SWIG_Py_Void();
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCoreConnect(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_core arg1 = (helics_core) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  helics_bool result;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCoreConnect" "', argument " "1"" of type '" "helics_core""'"); 
  }
  result = (helics_bool)helicsCoreConnect(arg1,arg2);
  resultobj = SWIG_From_int((int)(result));
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCoreDisconnect(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_core arg1 = (helics_core) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCoreDisconnect" "', argument " "1"" of type '" "helics_core""'"); 
  }
  helicsCoreDisconnect(arg1,arg2);
  resultobj = SWIG_Py_Void();
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsGetFederateByName(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  helics_federate result;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsGetFederateByName" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = (char *)(buf1);
  result = (helics_federate)helicsGetFederateByName((char const *)arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsBrokerDisconnect(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_broker arg1 = (helics_broker) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsBrokerDisconnect" "', argument " "1"" of type '" "helics_broker""'"); 
  }
  helicsBrokerDisconnect(arg1,arg2);
  resultobj = SWIG_Py_Void();
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateDestroy(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateDestroy" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  helicsFederateDestroy(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsBrokerDestroy(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_broker arg1 = (helics_broker) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsBrokerDestroy" "', argument " "1"" of type '" "helics_broker""'"); 
  }
  helicsBrokerDestroy(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCoreDestroy(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_core arg1 = (helics_core) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCoreDestroy" "', argument " "1"" of type '" "helics_core""'"); 
  }
  helicsCoreDestroy(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCoreFree(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_core arg1 = (helics_core) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCoreFree" "', argument " "1"" of type '" "helics_core""'"); 
  }
  helicsCoreFree(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsBrokerFree(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_broker arg1 = (helics_broker) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsBrokerFree" "', argument " "1"" of type '" "helics_broker""'"); 
  }
  helicsBrokerFree(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCreateValueFederate(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  helics_federate_info arg2 = (helics_federate_info) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  helics_federate result;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsCreateValueFederate", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCreateValueFederate" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsCreateValueFederate" "', argument " "2"" of type '" "helics_federate_info""'"); 
  }
  result = (helics_federate)helicsCreateValueFederate((char const *)arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCreateValueFederateFromConfig(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  helics_federate result;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCreateValueFederateFromConfig" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = (char *)(buf1);
  result = (helics_federate)helicsCreateValueFederateFromConfig((char const *)arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCreateMessageFederate(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  helics_federate_info arg2 = (helics_federate_info) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  helics_federate result;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsCreateMessageFederate", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCreateMessageFederate" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsCreateMessageFederate" "', argument " "2"" of type '" "helics_federate_info""'"); 
  }
  result = (helics_federate)helicsCreateMessageFederate((char const *)arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCreateMessageFederateFromConfig(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  helics_federate result;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCreateMessageFederateFromConfig" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = (char *)(buf1);
  result = (helics_federate)helicsCreateMessageFederateFromConfig((char const *)arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCreateCombinationFederate(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  helics_federate_info arg2 = (helics_federate_info) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  helics_federate result;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsCreateCombinationFederate", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCreateCombinationFederate" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsCreateCombinationFederate" "', argument " "2"" of type '" "helics_federate_info""'"); 
  }
  result = (helics_federate)helicsCreateCombinationFederate((char const *)arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCreateCombinationFederateFromConfig(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  helics_federate result;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCreateCombinationFederateFromConfig" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = (char *)(buf1);
  result = (helics_federate)helicsCreateCombinationFederateFromConfig((char const *)arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateClone(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  helics_federate result;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateClone" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  result = (helics_federate)helicsFederateClone(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCreateFederateInfo(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate_info result;
  
  if (!SWIG_Python_UnpackTuple(args, "helicsCreateFederateInfo", 0, 0, 0)) SWIG_fail;
  result = (helics_federate_info)helicsCreateFederateInfo();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateInfoClone(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate_info arg1 = (helics_federate_info) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  helics_federate_info result;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateInfoClone" "', argument " "1"" of type '" "helics_federate_info""'"); 
  }
  result = (helics_federate_info)helicsFederateInfoClone(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateInfoLoadFromArgs(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate_info arg1 = (helics_federate_info) 0 ;
  int arg2 ;
  char **arg3 = (char **) 0 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp4 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateInfoLoadFromArgs", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateInfoLoadFromArgs" "', argument " "1"" of type '" "helics_federate_info""'"); 
  }
  {
    /* Check if is a list */
    if (PyList_Check(swig_obj[1])) {
      int i;
      arg2 = (int)(PyList_Size(swig_obj[1]));
      arg3 = (char **) malloc((arg2+1)*sizeof(char *));
      for (i = 0; i < arg2; i++) {
        PyObject *o = PyList_GetItem(swig_obj[1],i);
        if (PyString_Check(o))
        arg3[i] = PyString_AsString(PyList_GetItem(swig_obj[1],i));
        else {
          PyErr_SetString(PyExc_TypeError,"list must contain strings");
          free(arg3);
          return NULL;
        }
      }
      arg3[i] = 0;
    } else {
      PyErr_SetString(PyExc_TypeError,"not a list");
      return NULL;
    }
  }
  helicsFederateInfoLoadFromArgs(arg1,arg2,(char const *const *)arg3,arg4);
  resultobj = SWIG_Py_Void();
  {
    free((char *) arg3);
  }
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    free((char *) arg3);
  }
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateInfoFree(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate_info arg1 = (helics_federate_info) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateInfoFree" "', argument " "1"" of type '" "helics_federate_info""'"); 
  }
  helicsFederateInfoFree(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateIsValid(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  helics_bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateIsValid" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  result = (helics_bool)helicsFederateIsValid(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateInfoSetCoreName(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate_info arg1 = (helics_federate_info) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateInfoSetCoreName", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateInfoSetCoreName" "', argument " "1"" of type '" "helics_federate_info""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateInfoSetCoreName" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsFederateInfoSetCoreName(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateInfoSetCoreInitString(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate_info arg1 = (helics_federate_info) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateInfoSetCoreInitString", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateInfoSetCoreInitString" "', argument " "1"" of type '" "helics_federate_info""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateInfoSetCoreInitString" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsFederateInfoSetCoreInitString(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateInfoSetBrokerInitString(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate_info arg1 = (helics_federate_info) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateInfoSetBrokerInitString", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateInfoSetBrokerInitString" "', argument " "1"" of type '" "helics_federate_info""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateInfoSetBrokerInitString" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsFederateInfoSetBrokerInitString(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateInfoSetCoreType(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate_info arg1 = (helics_federate_info) 0 ;
  int arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateInfoSetCoreType", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateInfoSetCoreType" "', argument " "1"" of type '" "helics_federate_info""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFederateInfoSetCoreType" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  helicsFederateInfoSetCoreType(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateInfoSetCoreTypeFromString(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate_info arg1 = (helics_federate_info) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateInfoSetCoreTypeFromString", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateInfoSetCoreTypeFromString" "', argument " "1"" of type '" "helics_federate_info""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateInfoSetCoreTypeFromString" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsFederateInfoSetCoreTypeFromString(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateInfoSetBroker(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate_info arg1 = (helics_federate_info) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateInfoSetBroker", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateInfoSetBroker" "', argument " "1"" of type '" "helics_federate_info""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateInfoSetBroker" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsFederateInfoSetBroker(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateInfoSetBrokerKey(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate_info arg1 = (helics_federate_info) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateInfoSetBrokerKey", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateInfoSetBrokerKey" "', argument " "1"" of type '" "helics_federate_info""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateInfoSetBrokerKey" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsFederateInfoSetBrokerKey(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateInfoSetBrokerPort(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate_info arg1 = (helics_federate_info) 0 ;
  int arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateInfoSetBrokerPort", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateInfoSetBrokerPort" "', argument " "1"" of type '" "helics_federate_info""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFederateInfoSetBrokerPort" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  helicsFederateInfoSetBrokerPort(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateInfoSetLocalPort(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate_info arg1 = (helics_federate_info) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateInfoSetLocalPort", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateInfoSetLocalPort" "', argument " "1"" of type '" "helics_federate_info""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateInfoSetLocalPort" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsFederateInfoSetLocalPort(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsGetPropertyIndex(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsGetPropertyIndex" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = (char *)(buf1);
  result = (int)helicsGetPropertyIndex((char const *)arg1);
  resultobj = SWIG_From_int((int)(result));
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return resultobj;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsGetFlagIndex(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsGetFlagIndex" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = (char *)(buf1);
  result = (int)helicsGetFlagIndex((char const *)arg1);
  resultobj = SWIG_From_int((int)(result));
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return resultobj;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsGetOptionIndex(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsGetOptionIndex" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = (char *)(buf1);
  result = (int)helicsGetOptionIndex((char const *)arg1);
  resultobj = SWIG_From_int((int)(result));
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return resultobj;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsGetOptionValue(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsGetOptionValue" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = (char *)(buf1);
  result = (int)helicsGetOptionValue((char const *)arg1);
  resultobj = SWIG_From_int((int)(result));
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return resultobj;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateInfoSetFlagOption(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate_info arg1 = (helics_federate_info) 0 ;
  int arg2 ;
  helics_bool arg3 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateInfoSetFlagOption", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateInfoSetFlagOption" "', argument " "1"" of type '" "helics_federate_info""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFederateInfoSetFlagOption" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "helicsFederateInfoSetFlagOption" "', argument " "3"" of type '" "helics_bool""'");
  } 
  arg3 = (helics_bool)(val3);
  helicsFederateInfoSetFlagOption(arg1,arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateInfoSetSeparator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate_info arg1 = (helics_federate_info) 0 ;
  char arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  char val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateInfoSetSeparator", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateInfoSetSeparator" "', argument " "1"" of type '" "helics_federate_info""'"); 
  }
  ecode2 = SWIG_AsVal_char(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFederateInfoSetSeparator" "', argument " "2"" of type '" "char""'");
  } 
  arg2 = (char)(val2);
  helicsFederateInfoSetSeparator(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateInfoSetTimeProperty(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate_info arg1 = (helics_federate_info) 0 ;
  int arg2 ;
  helics_time arg3 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateInfoSetTimeProperty", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateInfoSetTimeProperty" "', argument " "1"" of type '" "helics_federate_info""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFederateInfoSetTimeProperty" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  ecode3 = SWIG_AsVal_double(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "helicsFederateInfoSetTimeProperty" "', argument " "3"" of type '" "helics_time""'");
  } 
  arg3 = (helics_time)(val3);
  helicsFederateInfoSetTimeProperty(arg1,arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateInfoSetIntegerProperty(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate_info arg1 = (helics_federate_info) 0 ;
  int arg2 ;
  int arg3 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateInfoSetIntegerProperty", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateInfoSetIntegerProperty" "', argument " "1"" of type '" "helics_federate_info""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFederateInfoSetIntegerProperty" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "helicsFederateInfoSetIntegerProperty" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  helicsFederateInfoSetIntegerProperty(arg1,arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateRegisterInterfaces(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateRegisterInterfaces", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateRegisterInterfaces" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateRegisterInterfaces" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsFederateRegisterInterfaces(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateGlobalError(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  int arg2 ;
  char *arg3 = (char *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateGlobalError", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateGlobalError" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFederateGlobalError" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "helicsFederateGlobalError" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  helicsFederateGlobalError(arg1,arg2,(char const *)arg3);
  resultobj = SWIG_Py_Void();
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateLocalError(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  int arg2 ;
  char *arg3 = (char *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateLocalError", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateLocalError" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFederateLocalError" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "helicsFederateLocalError" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  helicsFederateLocalError(arg1,arg2,(char const *)arg3);
  resultobj = SWIG_Py_Void();
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateFinalize(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateFinalize" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  helicsFederateFinalize(arg1,arg2);
  resultobj = SWIG_Py_Void();
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateFinalizeAsync(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateFinalizeAsync" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  helicsFederateFinalizeAsync(arg1,arg2);
  resultobj = SWIG_Py_Void();
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateFinalizeComplete(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateFinalizeComplete" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  helicsFederateFinalizeComplete(arg1,arg2);
  resultobj = SWIG_Py_Void();
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateFree(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateFree" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  helicsFederateFree(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCloseLibrary(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  
  if (!SWIG_Python_UnpackTuple(args, "helicsCloseLibrary", 0, 0, 0)) SWIG_fail;
  helicsCloseLibrary();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateEnterInitializingMode(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateEnterInitializingMode" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  helicsFederateEnterInitializingMode(arg1,arg2);
  resultobj = SWIG_Py_Void();
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateEnterInitializingModeAsync(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateEnterInitializingModeAsync" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  helicsFederateEnterInitializingModeAsync(arg1,arg2);
  resultobj = SWIG_Py_Void();
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateIsAsyncOperationCompleted(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  helics_bool result;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateIsAsyncOperationCompleted" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  result = (helics_bool)helicsFederateIsAsyncOperationCompleted(arg1,arg2);
  resultobj = SWIG_From_int((int)(result));
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateEnterInitializingModeComplete(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateEnterInitializingModeComplete" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  helicsFederateEnterInitializingModeComplete(arg1,arg2);
  resultobj = SWIG_Py_Void();
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateEnterExecutingMode(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateEnterExecutingMode" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  helicsFederateEnterExecutingMode(arg1,arg2);
  resultobj = SWIG_Py_Void();
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateEnterExecutingModeAsync(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateEnterExecutingModeAsync" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  helicsFederateEnterExecutingModeAsync(arg1,arg2);
  resultobj = SWIG_Py_Void();
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateEnterExecutingModeComplete(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateEnterExecutingModeComplete" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  helicsFederateEnterExecutingModeComplete(arg1,arg2);
  resultobj = SWIG_Py_Void();
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateEnterExecutingModeIterative(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  helics_iteration_request arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  helics_iteration_result result;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateEnterExecutingModeIterative", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateEnterExecutingModeIterative" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFederateEnterExecutingModeIterative" "', argument " "2"" of type '" "helics_iteration_request""'");
  } 
  arg2 = (helics_iteration_request)(val2);
  result = (helics_iteration_result)helicsFederateEnterExecutingModeIterative(arg1,arg2,arg3);
  resultobj = SWIG_From_int((int)(result));
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateEnterExecutingModeIterativeAsync(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  helics_iteration_request arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateEnterExecutingModeIterativeAsync", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateEnterExecutingModeIterativeAsync" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFederateEnterExecutingModeIterativeAsync" "', argument " "2"" of type '" "helics_iteration_request""'");
  } 
  arg2 = (helics_iteration_request)(val2);
  helicsFederateEnterExecutingModeIterativeAsync(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateEnterExecutingModeIterativeComplete(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  helics_iteration_result result;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateEnterExecutingModeIterativeComplete" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  result = (helics_iteration_result)helicsFederateEnterExecutingModeIterativeComplete(arg1,arg2);
  resultobj = SWIG_From_int((int)(result));
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateGetState(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  helics_federate_state result;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateGetState" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  result = (helics_federate_state)helicsFederateGetState(arg1,arg2);
  resultobj = SWIG_From_int((int)(result));
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateGetCoreObject(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  helics_core result;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateGetCoreObject" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  result = (helics_core)helicsFederateGetCoreObject(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateRequestTime(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  helics_time arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  double val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  helics_time result;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateRequestTime", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateRequestTime" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFederateRequestTime" "', argument " "2"" of type '" "helics_time""'");
  } 
  arg2 = (helics_time)(val2);
  result = (helics_time)helicsFederateRequestTime(arg1,arg2,arg3);
  resultobj = SWIG_From_double((double)(result));
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateRequestTimeAdvance(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  helics_time arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  double val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  helics_time result;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateRequestTimeAdvance", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateRequestTimeAdvance" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFederateRequestTimeAdvance" "', argument " "2"" of type '" "helics_time""'");
  } 
  arg2 = (helics_time)(val2);
  result = (helics_time)helicsFederateRequestTimeAdvance(arg1,arg2,arg3);
  resultobj = SWIG_From_double((double)(result));
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateRequestNextStep(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  helics_time result;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateRequestNextStep" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  result = (helics_time)helicsFederateRequestNextStep(arg1,arg2);
  resultobj = SWIG_From_double((double)(result));
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateRequestTimeIterative(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  helics_time arg2 ;
  helics_iteration_request arg3 ;
  helics_iteration_result *arg4 = (helics_iteration_result *) 0 ;
  helics_error *arg5 = (helics_error *) 0 ;
  int res1 ;
  double val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  helics_iteration_result temp4 ;
  int res4 = SWIG_TMPOBJ ;
  helics_error etemp5 ;
  PyObject *swig_obj[3] ;
  helics_time result;
  
  arg4 = &temp4;
  {
    etemp5=helicsErrorInitialize();
    arg5=&etemp5;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateRequestTimeIterative", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateRequestTimeIterative" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFederateRequestTimeIterative" "', argument " "2"" of type '" "helics_time""'");
  } 
  arg2 = (helics_time)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "helicsFederateRequestTimeIterative" "', argument " "3"" of type '" "helics_iteration_request""'");
  } 
  arg3 = (helics_iteration_request)(val3);
  result = (helics_time)helicsFederateRequestTimeIterative(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_double((double)(result));
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_helics_iteration_result, new_flags));
  }
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateRequestTimeAsync(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  helics_time arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  double val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateRequestTimeAsync", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateRequestTimeAsync" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFederateRequestTimeAsync" "', argument " "2"" of type '" "helics_time""'");
  } 
  arg2 = (helics_time)(val2);
  helicsFederateRequestTimeAsync(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateRequestTimeComplete(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  helics_time result;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateRequestTimeComplete" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  result = (helics_time)helicsFederateRequestTimeComplete(arg1,arg2);
  resultobj = SWIG_From_double((double)(result));
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateRequestTimeIterativeAsync(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  helics_time arg2 ;
  helics_iteration_request arg3 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  double val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateRequestTimeIterativeAsync", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateRequestTimeIterativeAsync" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFederateRequestTimeIterativeAsync" "', argument " "2"" of type '" "helics_time""'");
  } 
  arg2 = (helics_time)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "helicsFederateRequestTimeIterativeAsync" "', argument " "3"" of type '" "helics_iteration_request""'");
  } 
  arg3 = (helics_iteration_request)(val3);
  helicsFederateRequestTimeIterativeAsync(arg1,arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateRequestTimeIterativeComplete(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  helics_iteration_result *arg2 = (helics_iteration_result *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  helics_iteration_result temp2 ;
  int res2 = SWIG_TMPOBJ ;
  helics_error etemp3 ;
  PyObject *swig_obj[1] ;
  helics_time result;
  
  arg2 = &temp2;
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateRequestTimeIterativeComplete" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  result = (helics_time)helicsFederateRequestTimeIterativeComplete(arg1,arg2,arg3);
  resultobj = SWIG_From_double((double)(result));
  if (SWIG_IsTmpObj(res2)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg2)));
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_helics_iteration_result, new_flags));
  }
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateGetName(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateGetName" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  result = (char *)helicsFederateGetName(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateSetTimeProperty(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  int arg2 ;
  helics_time arg3 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateSetTimeProperty", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateSetTimeProperty" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFederateSetTimeProperty" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  ecode3 = SWIG_AsVal_double(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "helicsFederateSetTimeProperty" "', argument " "3"" of type '" "helics_time""'");
  } 
  arg3 = (helics_time)(val3);
  helicsFederateSetTimeProperty(arg1,arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateSetFlagOption(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  int arg2 ;
  helics_bool arg3 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateSetFlagOption", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateSetFlagOption" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFederateSetFlagOption" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "helicsFederateSetFlagOption" "', argument " "3"" of type '" "helics_bool""'");
  } 
  arg3 = (helics_bool)(val3);
  helicsFederateSetFlagOption(arg1,arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateSetSeparator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  char arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  char val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateSetSeparator", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateSetSeparator" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  ecode2 = SWIG_AsVal_char(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFederateSetSeparator" "', argument " "2"" of type '" "char""'");
  } 
  arg2 = (char)(val2);
  helicsFederateSetSeparator(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateSetIntegerProperty(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  int arg2 ;
  int arg3 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateSetIntegerProperty", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateSetIntegerProperty" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFederateSetIntegerProperty" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "helicsFederateSetIntegerProperty" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  helicsFederateSetIntegerProperty(arg1,arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateGetTimeProperty(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  int arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  helics_time result;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateGetTimeProperty", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateGetTimeProperty" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFederateGetTimeProperty" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  result = (helics_time)helicsFederateGetTimeProperty(arg1,arg2,arg3);
  resultobj = SWIG_From_double((double)(result));
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateGetFlagOption(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  int arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  helics_bool result;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateGetFlagOption", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateGetFlagOption" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFederateGetFlagOption" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  result = (helics_bool)helicsFederateGetFlagOption(arg1,arg2,arg3);
  resultobj = SWIG_From_int((int)(result));
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateGetIntegerProperty(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  int arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  int result;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateGetIntegerProperty", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateGetIntegerProperty" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFederateGetIntegerProperty" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  result = (int)helicsFederateGetIntegerProperty(arg1,arg2,arg3);
  resultobj = SWIG_From_int((int)(result));
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateGetCurrentTime(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  helics_time result;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateGetCurrentTime" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  result = (helics_time)helicsFederateGetCurrentTime(arg1,arg2);
  resultobj = SWIG_From_double((double)(result));
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateSetGlobal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateSetGlobal", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateSetGlobal" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateSetGlobal" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "helicsFederateSetGlobal" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  helicsFederateSetGlobal(arg1,(char const *)arg2,(char const *)arg3,arg4);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateAddDependency(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateAddDependency", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateAddDependency" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateAddDependency" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsFederateAddDependency(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateSetLogFile(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateSetLogFile", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateSetLogFile" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateSetLogFile" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsFederateSetLogFile(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateLogErrorMessage(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateLogErrorMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateLogErrorMessage" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateLogErrorMessage" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsFederateLogErrorMessage(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateLogWarningMessage(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateLogWarningMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateLogWarningMessage" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateLogWarningMessage" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsFederateLogWarningMessage(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateLogInfoMessage(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateLogInfoMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateLogInfoMessage" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateLogInfoMessage" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsFederateLogInfoMessage(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateLogDebugMessage(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateLogDebugMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateLogDebugMessage" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateLogDebugMessage" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsFederateLogDebugMessage(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateLogLevelMessage(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  int arg2 ;
  char *arg3 = (char *) 0 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateLogLevelMessage", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateLogLevelMessage" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFederateLogLevelMessage" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "helicsFederateLogLevelMessage" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  helicsFederateLogLevelMessage(arg1,arg2,(char const *)arg3,arg4);
  resultobj = SWIG_Py_Void();
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCoreSetGlobal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_core arg1 = (helics_core) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsCoreSetGlobal", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCoreSetGlobal" "', argument " "1"" of type '" "helics_core""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsCoreSetGlobal" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "helicsCoreSetGlobal" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  helicsCoreSetGlobal(arg1,(char const *)arg2,(char const *)arg3,arg4);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsBrokerSetGlobal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_broker arg1 = (helics_broker) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsBrokerSetGlobal", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsBrokerSetGlobal" "', argument " "1"" of type '" "helics_broker""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsBrokerSetGlobal" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "helicsBrokerSetGlobal" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  helicsBrokerSetGlobal(arg1,(char const *)arg2,(char const *)arg3,arg4);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCoreSetLogFile(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_core arg1 = (helics_core) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsCoreSetLogFile", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCoreSetLogFile" "', argument " "1"" of type '" "helics_core""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsCoreSetLogFile" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsCoreSetLogFile(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsBrokerSetLogFile(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_broker arg1 = (helics_broker) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsBrokerSetLogFile", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsBrokerSetLogFile" "', argument " "1"" of type '" "helics_broker""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsBrokerSetLogFile" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsBrokerSetLogFile(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCreateQuery(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  helics_query result;
  
  if (!SWIG_Python_UnpackTuple(args, "helicsCreateQuery", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCreateQuery" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsCreateQuery" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  result = (helics_query)helicsCreateQuery((char const *)arg1,(char const *)arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsQueryExecute(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_query arg1 = (helics_query) 0 ;
  helics_federate arg2 = (helics_federate) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsQueryExecute", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsQueryExecute" "', argument " "1"" of type '" "helics_query""'"); 
  }
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsQueryExecute" "', argument " "2"" of type '" "helics_federate""'"); 
  }
  result = (char *)helicsQueryExecute(arg1,arg2,arg3);
  resultobj = SWIG_FromCharPtr((const char *)result);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsQueryCoreExecute(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_query arg1 = (helics_query) 0 ;
  helics_core arg2 = (helics_core) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsQueryCoreExecute", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsQueryCoreExecute" "', argument " "1"" of type '" "helics_query""'"); 
  }
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsQueryCoreExecute" "', argument " "2"" of type '" "helics_core""'"); 
  }
  result = (char *)helicsQueryCoreExecute(arg1,arg2,arg3);
  resultobj = SWIG_FromCharPtr((const char *)result);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsQueryBrokerExecute(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_query arg1 = (helics_query) 0 ;
  helics_broker arg2 = (helics_broker) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsQueryBrokerExecute", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsQueryBrokerExecute" "', argument " "1"" of type '" "helics_query""'"); 
  }
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsQueryBrokerExecute" "', argument " "2"" of type '" "helics_broker""'"); 
  }
  result = (char *)helicsQueryBrokerExecute(arg1,arg2,arg3);
  resultobj = SWIG_FromCharPtr((const char *)result);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsQueryExecuteAsync(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_query arg1 = (helics_query) 0 ;
  helics_federate arg2 = (helics_federate) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsQueryExecuteAsync", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsQueryExecuteAsync" "', argument " "1"" of type '" "helics_query""'"); 
  }
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsQueryExecuteAsync" "', argument " "2"" of type '" "helics_federate""'"); 
  }
  helicsQueryExecuteAsync(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsQueryExecuteComplete(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_query arg1 = (helics_query) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsQueryExecuteComplete" "', argument " "1"" of type '" "helics_query""'"); 
  }
  result = (char *)helicsQueryExecuteComplete(arg1,arg2);
  resultobj = SWIG_FromCharPtr((const char *)result);
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsQueryIsCompleted(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_query arg1 = (helics_query) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  helics_bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsQueryIsCompleted" "', argument " "1"" of type '" "helics_query""'"); 
  }
  result = (helics_bool)helicsQueryIsCompleted(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsQuerySetTarget(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_query arg1 = (helics_query) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsQuerySetTarget", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsQuerySetTarget" "', argument " "1"" of type '" "helics_query""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsQuerySetTarget" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsQuerySetTarget(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsQuerySetQueryString(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_query arg1 = (helics_query) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsQuerySetQueryString", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsQuerySetQueryString" "', argument " "1"" of type '" "helics_query""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsQuerySetQueryString" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsQuerySetQueryString(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsQueryFree(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_query arg1 = (helics_query) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsQueryFree" "', argument " "1"" of type '" "helics_query""'"); 
  }
  helicsQueryFree(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCleanupLibrary(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  
  if (!SWIG_Python_UnpackTuple(args, "helicsCleanupLibrary", 0, 0, 0)) SWIG_fail;
  helicsCleanupLibrary();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateRegisterSubscription(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  helics_input result;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateRegisterSubscription", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateRegisterSubscription" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateRegisterSubscription" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "helicsFederateRegisterSubscription" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  result = (helics_input)helicsFederateRegisterSubscription(arg1,(char const *)arg2,(char const *)arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateRegisterPublication(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  char *arg2 = (char *) 0 ;
  helics_data_type arg3 ;
  char *arg4 = (char *) 0 ;
  helics_error *arg5 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  helics_error etemp5 ;
  PyObject *swig_obj[4] ;
  helics_publication result;
  
  {
    etemp5=helicsErrorInitialize();
    arg5=&etemp5;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateRegisterPublication", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateRegisterPublication" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateRegisterPublication" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "helicsFederateRegisterPublication" "', argument " "3"" of type '" "helics_data_type""'");
  } 
  arg3 = (helics_data_type)(val3);
  res4 = SWIG_AsCharPtrAndSize(swig_obj[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "helicsFederateRegisterPublication" "', argument " "4"" of type '" "char const *""'");
  }
  arg4 = (char *)(buf4);
  result = (helics_publication)helicsFederateRegisterPublication(arg1,(char const *)arg2,arg3,(char const *)arg4,arg5);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateRegisterTypePublication(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  helics_error *arg5 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  helics_error etemp5 ;
  PyObject *swig_obj[4] ;
  helics_publication result;
  
  {
    etemp5=helicsErrorInitialize();
    arg5=&etemp5;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateRegisterTypePublication", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateRegisterTypePublication" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateRegisterTypePublication" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "helicsFederateRegisterTypePublication" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(swig_obj[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "helicsFederateRegisterTypePublication" "', argument " "4"" of type '" "char const *""'");
  }
  arg4 = (char *)(buf4);
  result = (helics_publication)helicsFederateRegisterTypePublication(arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4,arg5);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateRegisterGlobalPublication(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  char *arg2 = (char *) 0 ;
  helics_data_type arg3 ;
  char *arg4 = (char *) 0 ;
  helics_error *arg5 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  helics_error etemp5 ;
  PyObject *swig_obj[4] ;
  helics_publication result;
  
  {
    etemp5=helicsErrorInitialize();
    arg5=&etemp5;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateRegisterGlobalPublication", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateRegisterGlobalPublication" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateRegisterGlobalPublication" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "helicsFederateRegisterGlobalPublication" "', argument " "3"" of type '" "helics_data_type""'");
  } 
  arg3 = (helics_data_type)(val3);
  res4 = SWIG_AsCharPtrAndSize(swig_obj[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "helicsFederateRegisterGlobalPublication" "', argument " "4"" of type '" "char const *""'");
  }
  arg4 = (char *)(buf4);
  result = (helics_publication)helicsFederateRegisterGlobalPublication(arg1,(char const *)arg2,arg3,(char const *)arg4,arg5);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateRegisterGlobalTypePublication(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  helics_error *arg5 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  helics_error etemp5 ;
  PyObject *swig_obj[4] ;
  helics_publication result;
  
  {
    etemp5=helicsErrorInitialize();
    arg5=&etemp5;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateRegisterGlobalTypePublication", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateRegisterGlobalTypePublication" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateRegisterGlobalTypePublication" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "helicsFederateRegisterGlobalTypePublication" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(swig_obj[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "helicsFederateRegisterGlobalTypePublication" "', argument " "4"" of type '" "char const *""'");
  }
  arg4 = (char *)(buf4);
  result = (helics_publication)helicsFederateRegisterGlobalTypePublication(arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4,arg5);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateRegisterInput(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  char *arg2 = (char *) 0 ;
  helics_data_type arg3 ;
  char *arg4 = (char *) 0 ;
  helics_error *arg5 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  helics_error etemp5 ;
  PyObject *swig_obj[4] ;
  helics_input result;
  
  {
    etemp5=helicsErrorInitialize();
    arg5=&etemp5;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateRegisterInput", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateRegisterInput" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateRegisterInput" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "helicsFederateRegisterInput" "', argument " "3"" of type '" "helics_data_type""'");
  } 
  arg3 = (helics_data_type)(val3);
  res4 = SWIG_AsCharPtrAndSize(swig_obj[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "helicsFederateRegisterInput" "', argument " "4"" of type '" "char const *""'");
  }
  arg4 = (char *)(buf4);
  result = (helics_input)helicsFederateRegisterInput(arg1,(char const *)arg2,arg3,(char const *)arg4,arg5);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateRegisterTypeInput(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  helics_error *arg5 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  helics_error etemp5 ;
  PyObject *swig_obj[4] ;
  helics_input result;
  
  {
    etemp5=helicsErrorInitialize();
    arg5=&etemp5;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateRegisterTypeInput", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateRegisterTypeInput" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateRegisterTypeInput" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "helicsFederateRegisterTypeInput" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(swig_obj[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "helicsFederateRegisterTypeInput" "', argument " "4"" of type '" "char const *""'");
  }
  arg4 = (char *)(buf4);
  result = (helics_input)helicsFederateRegisterTypeInput(arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4,arg5);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateRegisterGlobalInput(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  char *arg2 = (char *) 0 ;
  helics_data_type arg3 ;
  char *arg4 = (char *) 0 ;
  helics_error *arg5 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  helics_error etemp5 ;
  PyObject *swig_obj[4] ;
  helics_publication result;
  
  {
    etemp5=helicsErrorInitialize();
    arg5=&etemp5;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateRegisterGlobalInput", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateRegisterGlobalInput" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateRegisterGlobalInput" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "helicsFederateRegisterGlobalInput" "', argument " "3"" of type '" "helics_data_type""'");
  } 
  arg3 = (helics_data_type)(val3);
  res4 = SWIG_AsCharPtrAndSize(swig_obj[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "helicsFederateRegisterGlobalInput" "', argument " "4"" of type '" "char const *""'");
  }
  arg4 = (char *)(buf4);
  result = (helics_publication)helicsFederateRegisterGlobalInput(arg1,(char const *)arg2,arg3,(char const *)arg4,arg5);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateRegisterGlobalTypeInput(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  helics_error *arg5 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  helics_error etemp5 ;
  PyObject *swig_obj[4] ;
  helics_publication result;
  
  {
    etemp5=helicsErrorInitialize();
    arg5=&etemp5;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateRegisterGlobalTypeInput", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateRegisterGlobalTypeInput" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateRegisterGlobalTypeInput" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "helicsFederateRegisterGlobalTypeInput" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(swig_obj[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "helicsFederateRegisterGlobalTypeInput" "', argument " "4"" of type '" "char const *""'");
  }
  arg4 = (char *)(buf4);
  result = (helics_publication)helicsFederateRegisterGlobalTypeInput(arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4,arg5);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateGetPublication(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  helics_publication result;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateGetPublication", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateGetPublication" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateGetPublication" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  result = (helics_publication)helicsFederateGetPublication(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateGetPublicationByIndex(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  int arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  helics_publication result;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateGetPublicationByIndex", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateGetPublicationByIndex" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFederateGetPublicationByIndex" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  result = (helics_publication)helicsFederateGetPublicationByIndex(arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateGetInput(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  helics_input result;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateGetInput", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateGetInput" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateGetInput" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  result = (helics_input)helicsFederateGetInput(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateGetInputByIndex(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  int arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  helics_input result;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateGetInputByIndex", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateGetInputByIndex" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFederateGetInputByIndex" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  result = (helics_input)helicsFederateGetInputByIndex(arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateGetSubscription(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  helics_input result;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateGetSubscription", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateGetSubscription" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateGetSubscription" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  result = (helics_input)helicsFederateGetSubscription(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateClearUpdates(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateClearUpdates" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  helicsFederateClearUpdates(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateRegisterFromPublicationJSON(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateRegisterFromPublicationJSON", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateRegisterFromPublicationJSON" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateRegisterFromPublicationJSON" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsFederateRegisterFromPublicationJSON(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederatePublishJSON(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederatePublishJSON", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederatePublishJSON" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederatePublishJSON" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsFederatePublishJSON(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsPublicationIsValid(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_publication arg1 = (helics_publication) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  helics_bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsPublicationIsValid" "', argument " "1"" of type '" "helics_publication""'"); 
  }
  result = (helics_bool)helicsPublicationIsValid(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsPublicationPublishRaw(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_publication arg1 = (helics_publication) 0 ;
  void *arg2 = (void *) 0 ;
  int arg3 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp4 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsPublicationPublishRaw", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsPublicationPublishRaw" "', argument " "1"" of type '" "helics_publication""'"); 
  }
  {
    if (PyUnicode_Check(swig_obj[1])) {
      int kind=PyUnicode_KIND(swig_obj[1]);
      arg2=PyUnicode_DATA(swig_obj[1]);
      switch(kind)
      {
      case PyUnicode_1BYTE_KIND:
      default:
        arg3=(int)(PyUnicode_GetLength(swig_obj[1]));
        break;
      case PyUnicode_2BYTE_KIND:
      case PyUnicode_WCHAR_KIND:
        arg3=(int)(PyUnicode_GetLength(swig_obj[1])*2);
        break;
      case PyUnicode_4BYTE_KIND:
        arg3=(int)(PyUnicode_GetLength(swig_obj[1])*4);
        break;
      }
    }
    else if (PyBytes_Check(swig_obj[1])) {
      arg2=PyBytes_AsString(swig_obj[1]);
      arg3=(int)(PyBytes_Size(swig_obj[1]));
    }
    else
    {
      PyErr_SetString(PyExc_ValueError,"Expected a string or bytes");
      return NULL;
    }
  }
  helicsPublicationPublishRaw(arg1,(void const *)arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  {
    
  }
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsPublicationPublishString(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_publication arg1 = (helics_publication) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsPublicationPublishString", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsPublicationPublishString" "', argument " "1"" of type '" "helics_publication""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsPublicationPublishString" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsPublicationPublishString(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsPublicationPublishInteger(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_publication arg1 = (helics_publication) 0 ;
  int64_t arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  long long val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsPublicationPublishInteger", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsPublicationPublishInteger" "', argument " "1"" of type '" "helics_publication""'"); 
  }
  ecode2 = SWIG_AsVal_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsPublicationPublishInteger" "', argument " "2"" of type '" "int64_t""'");
  } 
  arg2 = (int64_t)(val2);
  helicsPublicationPublishInteger(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsPublicationPublishBoolean(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_publication arg1 = (helics_publication) 0 ;
  helics_bool arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsPublicationPublishBoolean", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsPublicationPublishBoolean" "', argument " "1"" of type '" "helics_publication""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsPublicationPublishBoolean" "', argument " "2"" of type '" "helics_bool""'");
  } 
  arg2 = (helics_bool)(val2);
  helicsPublicationPublishBoolean(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsPublicationPublishDouble(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_publication arg1 = (helics_publication) 0 ;
  double arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  double val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsPublicationPublishDouble", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsPublicationPublishDouble" "', argument " "1"" of type '" "helics_publication""'"); 
  }
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsPublicationPublishDouble" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = (double)(val2);
  helicsPublicationPublishDouble(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsPublicationPublishTime(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_publication arg1 = (helics_publication) 0 ;
  helics_time arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  double val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsPublicationPublishTime", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsPublicationPublishTime" "', argument " "1"" of type '" "helics_publication""'"); 
  }
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsPublicationPublishTime" "', argument " "2"" of type '" "helics_time""'");
  } 
  arg2 = (helics_time)(val2);
  helicsPublicationPublishTime(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsPublicationPublishChar(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_publication arg1 = (helics_publication) 0 ;
  char arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  char val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsPublicationPublishChar", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsPublicationPublishChar" "', argument " "1"" of type '" "helics_publication""'"); 
  }
  ecode2 = SWIG_AsVal_char(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsPublicationPublishChar" "', argument " "2"" of type '" "char""'");
  } 
  arg2 = (char)(val2);
  helicsPublicationPublishChar(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsPublicationPublishComplex(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_publication arg1 = (helics_publication) 0 ;
  double arg2 ;
  double arg3 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  double val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsPublicationPublishComplex", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsPublicationPublishComplex" "', argument " "1"" of type '" "helics_publication""'"); 
  }
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsPublicationPublishComplex" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = (double)(val2);
  ecode3 = SWIG_AsVal_double(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "helicsPublicationPublishComplex" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = (double)(val3);
  helicsPublicationPublishComplex(arg1,arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsPublicationPublishVector(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_publication arg1 = (helics_publication) 0 ;
  double *arg2 = (double *) 0 ;
  int arg3 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp4 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsPublicationPublishVector", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsPublicationPublishVector" "', argument " "1"" of type '" "helics_publication""'"); 
  }
  {
    int i;
    if (!PyList_Check(swig_obj[1])) {
      PyErr_SetString(PyExc_ValueError,"Expected a list");
      return NULL;
    }
    arg3=(int)(PyList_Size(swig_obj[1]));
    arg2 = (double *) malloc(arg3*sizeof(double));
    
    for (i = 0; i < arg3; i++) {
      PyObject *o = PyList_GetItem(swig_obj[1],i);
      if (PyFloat_Check(o)) {
        arg2[i] = PyFloat_AsDouble(o);
      }else if (PyInt_Check(o))
      {
        arg2[i] = (double)(PyInt_AsLong(o));
      } else {
        PyErr_SetString(PyExc_ValueError,"List elements must be numbers");
        free(arg2);
        return NULL;
      }
    }
  }
  helicsPublicationPublishVector(arg1,(double const *)arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  {
    
  }
  {
    if (arg2) free(arg2);
  }
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2) free(arg2);
  }
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsPublicationPublishNamedPoint(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_publication arg1 = (helics_publication) 0 ;
  char *arg2 = (char *) 0 ;
  double arg3 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsPublicationPublishNamedPoint", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsPublicationPublishNamedPoint" "', argument " "1"" of type '" "helics_publication""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsPublicationPublishNamedPoint" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  ecode3 = SWIG_AsVal_double(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "helicsPublicationPublishNamedPoint" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = (double)(val3);
  helicsPublicationPublishNamedPoint(arg1,(char const *)arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsPublicationAddTarget(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_publication arg1 = (helics_publication) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsPublicationAddTarget", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsPublicationAddTarget" "', argument " "1"" of type '" "helics_publication""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsPublicationAddTarget" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsPublicationAddTarget(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputIsValid(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  helics_bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputIsValid" "', argument " "1"" of type '" "helics_input""'"); 
  }
  result = (helics_bool)helicsInputIsValid(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputAddTarget(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsInputAddTarget", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputAddTarget" "', argument " "1"" of type '" "helics_input""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsInputAddTarget" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsInputAddTarget(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputGetRawValueSize(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputGetRawValueSize" "', argument " "1"" of type '" "helics_input""'"); 
  }
  result = (int)helicsInputGetRawValueSize(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputGetRawValue(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  void *arg2 = (void *) 0 ;
  int arg3 ;
  int *arg4 = (int *) 0 ;
  helics_error *arg5 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp5 ;
  PyObject *swig_obj[1] ;
  
  {
    arg4=&(arg3);
  }
  {
    etemp5=helicsErrorInitialize();
    arg5=&etemp5;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputGetRawValue" "', argument " "1"" of type '" "helics_input""'"); 
  }
  {
    arg3=helicsInputGetRawValueSize(arg1)+2;
    arg2 =  malloc(arg3);
  }
  helicsInputGetRawValue(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_Py_Void();
  {
    PyObject *o2=PyBytes_FromStringAndSize(arg2,*arg4);
    resultobj = SWIG_Python_AppendOutput(resultobj, o2);
  }
  {
    if (arg2) free(arg2);
  }
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2) free(arg2);
  }
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputGetStringSize(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputGetStringSize" "', argument " "1"" of type '" "helics_input""'"); 
  }
  result = (int)helicsInputGetStringSize(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputGetString(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  char *arg2 = (char *) 0 ;
  int arg3 ;
  int *arg4 = (int *) 0 ;
  helics_error *arg5 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp5 ;
  PyObject *swig_obj[1] ;
  
  {
    arg4=&(arg3);
  }
  {
    etemp5=helicsErrorInitialize();
    arg5=&etemp5;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputGetString" "', argument " "1"" of type '" "helics_input""'"); 
  }
  {
    arg3=helicsInputGetStringSize(arg1)+2;
    arg2 = (char *) malloc(arg3);
  }
  helicsInputGetString(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_Py_Void();
  {
    PyObject *o2=PyString_FromString(arg2);
    resultobj = SWIG_Python_AppendOutput(resultobj, o2);
  }
  {
    if (arg2) free(arg2);
  }
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2) free(arg2);
  }
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputGetInteger(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  int64_t result;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputGetInteger" "', argument " "1"" of type '" "helics_input""'"); 
  }
  result = helicsInputGetInteger(arg1,arg2);
  resultobj = SWIG_From_long_SS_long((long long)(result));
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputGetBoolean(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  helics_bool result;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputGetBoolean" "', argument " "1"" of type '" "helics_input""'"); 
  }
  result = (helics_bool)helicsInputGetBoolean(arg1,arg2);
  resultobj = SWIG_From_int((int)(result));
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputGetDouble(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  double result;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputGetDouble" "', argument " "1"" of type '" "helics_input""'"); 
  }
  result = (double)helicsInputGetDouble(arg1,arg2);
  resultobj = SWIG_From_double((double)(result));
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputGetTime(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  helics_time result;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputGetTime" "', argument " "1"" of type '" "helics_input""'"); 
  }
  result = (helics_time)helicsInputGetTime(arg1,arg2);
  resultobj = SWIG_From_double((double)(result));
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputGetChar(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  char result;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputGetChar" "', argument " "1"" of type '" "helics_input""'"); 
  }
  result = (char)helicsInputGetChar(arg1,arg2);
  resultobj = SWIG_From_char((char)(result));
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputGetComplexObject(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  helics_complex result;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputGetComplexObject" "', argument " "1"" of type '" "helics_input""'"); 
  }
  result = helicsInputGetComplexObject(arg1,arg2);
  resultobj = SWIG_NewPointerObj((helics_complex *)memcpy((helics_complex *)calloc(1,sizeof(helics_complex)),&result,sizeof(helics_complex)), SWIGTYPE_p_helics_complex, SWIG_POINTER_OWN |  0 );
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputGetComplex(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  double *arg2 = (double *) 0 ;
  double *arg3 = (double *) 0 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  double temp2 ;
  int res2 = SWIG_TMPOBJ ;
  double temp3 ;
  int res3 = SWIG_TMPOBJ ;
  helics_error etemp4 ;
  PyObject *swig_obj[1] ;
  
  arg2 = &temp2;
  arg3 = &temp3;
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputGetComplex" "', argument " "1"" of type '" "helics_input""'"); 
  }
  helicsInputGetComplex(arg1,arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsTmpObj(res2)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_double((*arg2)));
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_double, new_flags));
  }
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_double((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_double, new_flags));
  }
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputGetVectorSize(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputGetVectorSize" "', argument " "1"" of type '" "helics_input""'"); 
  }
  result = (int)helicsInputGetVectorSize(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputGetVector(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  double *arg2 ;
  int arg3 ;
  int *arg4 = (int *) 0 ;
  helics_error *arg5 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp5 ;
  PyObject *swig_obj[1] ;
  
  {
    arg2=(double *)(NULL);
  }
  {
    arg4=&(arg3);
  }
  {
    etemp5=helicsErrorInitialize();
    arg5=&etemp5;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputGetVector" "', argument " "1"" of type '" "helics_input""'"); 
  }
  {
    arg3=helicsInputGetVectorSize(arg1);
    arg2 = (double *) malloc(arg3*sizeof(double));
  }
  helicsInputGetVector(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_Py_Void();
  {
    int i;
    PyObject *o2=PyList_New(*arg4);
    for (i = 0; i < *arg4; i++) {
      PyObject *o_item=PyFloat_FromDouble(arg2[i]);
      PyList_SetItem(o2, i, o_item);
    }
    
    resultobj = SWIG_Python_AppendOutput(resultobj, o2);
  }
  {
    if (arg2) free(arg2);
  }
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2) free(arg2);
  }
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputGetNamedPoint(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  char *arg2 = (char *) 0 ;
  int arg3 ;
  int *arg4 = (int *) 0 ;
  double *arg5 = (double *) 0 ;
  helics_error *arg6 = (helics_error *) 0 ;
  int res1 ;
  double temp5 ;
  int res5 = SWIG_TMPOBJ ;
  helics_error etemp6 ;
  PyObject *swig_obj[1] ;
  
  {
    arg4=&(arg3);
  }
  arg5 = &temp5;
  {
    etemp6=helicsErrorInitialize();
    arg6=&etemp6;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputGetNamedPoint" "', argument " "1"" of type '" "helics_input""'"); 
  }
  {
    arg3=helicsInputGetStringSize(arg1)+2;
    arg2 = (char *) malloc(arg3);
  }
  helicsInputGetNamedPoint(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_Py_Void();
  {
    PyObject *o2=PyString_FromString(arg2);
    resultobj = SWIG_Python_AppendOutput(resultobj, o2);
  }
  if (SWIG_IsTmpObj(res5)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_double((*arg5)));
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_double, new_flags));
  }
  {
    if (arg2) free(arg2);
  }
  {
    if (arg6->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg6);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2) free(arg2);
  }
  {
    if (arg6->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg6);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputSetDefaultRaw(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  void *arg2 = (void *) 0 ;
  int arg3 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp4 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsInputSetDefaultRaw", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputSetDefaultRaw" "', argument " "1"" of type '" "helics_input""'"); 
  }
  {
    if (PyUnicode_Check(swig_obj[1])) {
      int kind=PyUnicode_KIND(swig_obj[1]);
      arg2=PyUnicode_DATA(swig_obj[1]);
      switch(kind)
      {
      case PyUnicode_1BYTE_KIND:
      default:
        arg3=(int)(PyUnicode_GetLength(swig_obj[1]));
        break;
      case PyUnicode_2BYTE_KIND:
      case PyUnicode_WCHAR_KIND:
        arg3=(int)(PyUnicode_GetLength(swig_obj[1])*2);
        break;
      case PyUnicode_4BYTE_KIND:
        arg3=(int)(PyUnicode_GetLength(swig_obj[1])*4);
        break;
      }
    }
    else if (PyBytes_Check(swig_obj[1])) {
      arg2=PyBytes_AsString(swig_obj[1]);
      arg3=(int)(PyBytes_Size(swig_obj[1]));
    }
    else
    {
      PyErr_SetString(PyExc_ValueError,"Expected a string or bytes");
      return NULL;
    }
  }
  helicsInputSetDefaultRaw(arg1,(void const *)arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  {
    
  }
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputSetDefaultString(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsInputSetDefaultString", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputSetDefaultString" "', argument " "1"" of type '" "helics_input""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsInputSetDefaultString" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsInputSetDefaultString(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputSetDefaultInteger(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  int64_t arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  long long val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsInputSetDefaultInteger", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputSetDefaultInteger" "', argument " "1"" of type '" "helics_input""'"); 
  }
  ecode2 = SWIG_AsVal_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsInputSetDefaultInteger" "', argument " "2"" of type '" "int64_t""'");
  } 
  arg2 = (int64_t)(val2);
  helicsInputSetDefaultInteger(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputSetDefaultBoolean(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  helics_bool arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsInputSetDefaultBoolean", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputSetDefaultBoolean" "', argument " "1"" of type '" "helics_input""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsInputSetDefaultBoolean" "', argument " "2"" of type '" "helics_bool""'");
  } 
  arg2 = (helics_bool)(val2);
  helicsInputSetDefaultBoolean(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputSetDefaultTime(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  helics_time arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  double val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsInputSetDefaultTime", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputSetDefaultTime" "', argument " "1"" of type '" "helics_input""'"); 
  }
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsInputSetDefaultTime" "', argument " "2"" of type '" "helics_time""'");
  } 
  arg2 = (helics_time)(val2);
  helicsInputSetDefaultTime(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputSetDefaultChar(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  char arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  char val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsInputSetDefaultChar", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputSetDefaultChar" "', argument " "1"" of type '" "helics_input""'"); 
  }
  ecode2 = SWIG_AsVal_char(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsInputSetDefaultChar" "', argument " "2"" of type '" "char""'");
  } 
  arg2 = (char)(val2);
  helicsInputSetDefaultChar(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputSetDefaultDouble(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  double arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  double val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsInputSetDefaultDouble", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputSetDefaultDouble" "', argument " "1"" of type '" "helics_input""'"); 
  }
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsInputSetDefaultDouble" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = (double)(val2);
  helicsInputSetDefaultDouble(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputSetDefaultComplex(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  double arg2 ;
  double arg3 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  double val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsInputSetDefaultComplex", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputSetDefaultComplex" "', argument " "1"" of type '" "helics_input""'"); 
  }
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsInputSetDefaultComplex" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = (double)(val2);
  ecode3 = SWIG_AsVal_double(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "helicsInputSetDefaultComplex" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = (double)(val3);
  helicsInputSetDefaultComplex(arg1,arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputSetDefaultVector(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  double *arg2 = (double *) 0 ;
  int arg3 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp4 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsInputSetDefaultVector", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputSetDefaultVector" "', argument " "1"" of type '" "helics_input""'"); 
  }
  {
    int i;
    if (!PyList_Check(swig_obj[1])) {
      PyErr_SetString(PyExc_ValueError,"Expected a list");
      return NULL;
    }
    arg3=(int)(PyList_Size(swig_obj[1]));
    arg2 = (double *) malloc(arg3*sizeof(double));
    
    for (i = 0; i < arg3; i++) {
      PyObject *o = PyList_GetItem(swig_obj[1],i);
      if (PyFloat_Check(o)) {
        arg2[i] = PyFloat_AsDouble(o);
      }else if (PyInt_Check(o))
      {
        arg2[i] = (double)(PyInt_AsLong(o));
      } else {
        PyErr_SetString(PyExc_ValueError,"List elements must be numbers");
        free(arg2);
        return NULL;
      }
    }
  }
  helicsInputSetDefaultVector(arg1,(double const *)arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  {
    
  }
  {
    if (arg2) free(arg2);
  }
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2) free(arg2);
  }
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputSetDefaultNamedPoint(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  char *arg2 = (char *) 0 ;
  double arg3 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsInputSetDefaultNamedPoint", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputSetDefaultNamedPoint" "', argument " "1"" of type '" "helics_input""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsInputSetDefaultNamedPoint" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  ecode3 = SWIG_AsVal_double(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "helicsInputSetDefaultNamedPoint" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = (double)(val3);
  helicsInputSetDefaultNamedPoint(arg1,(char const *)arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputGetType(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputGetType" "', argument " "1"" of type '" "helics_input""'"); 
  }
  result = (char *)helicsInputGetType(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputGetPublicationType(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputGetPublicationType" "', argument " "1"" of type '" "helics_input""'"); 
  }
  result = (char *)helicsInputGetPublicationType(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsPublicationGetType(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_publication arg1 = (helics_publication) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsPublicationGetType" "', argument " "1"" of type '" "helics_publication""'"); 
  }
  result = (char *)helicsPublicationGetType(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputGetKey(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputGetKey" "', argument " "1"" of type '" "helics_input""'"); 
  }
  result = (char *)helicsInputGetKey(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsSubscriptionGetKey(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsSubscriptionGetKey" "', argument " "1"" of type '" "helics_input""'"); 
  }
  result = (char *)helicsSubscriptionGetKey(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsPublicationGetKey(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_publication arg1 = (helics_publication) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsPublicationGetKey" "', argument " "1"" of type '" "helics_publication""'"); 
  }
  result = (char *)helicsPublicationGetKey(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputGetUnits(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputGetUnits" "', argument " "1"" of type '" "helics_input""'"); 
  }
  result = (char *)helicsInputGetUnits(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputGetInjectionUnits(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputGetInjectionUnits" "', argument " "1"" of type '" "helics_input""'"); 
  }
  result = (char *)helicsInputGetInjectionUnits(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputGetExtractionUnits(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputGetExtractionUnits" "', argument " "1"" of type '" "helics_input""'"); 
  }
  result = (char *)helicsInputGetExtractionUnits(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsPublicationGetUnits(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_publication arg1 = (helics_publication) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsPublicationGetUnits" "', argument " "1"" of type '" "helics_publication""'"); 
  }
  result = (char *)helicsPublicationGetUnits(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputGetInfo(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputGetInfo" "', argument " "1"" of type '" "helics_input""'"); 
  }
  result = (char *)helicsInputGetInfo(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputSetInfo(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsInputSetInfo", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputSetInfo" "', argument " "1"" of type '" "helics_input""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsInputSetInfo" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsInputSetInfo(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsPublicationGetInfo(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_publication arg1 = (helics_publication) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsPublicationGetInfo" "', argument " "1"" of type '" "helics_publication""'"); 
  }
  result = (char *)helicsPublicationGetInfo(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsPublicationSetInfo(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_publication arg1 = (helics_publication) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsPublicationSetInfo", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsPublicationSetInfo" "', argument " "1"" of type '" "helics_publication""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsPublicationSetInfo" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsPublicationSetInfo(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputGetOption(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  int arg2 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  int result;
  
  if (!SWIG_Python_UnpackTuple(args, "helicsInputGetOption", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputGetOption" "', argument " "1"" of type '" "helics_input""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsInputGetOption" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  result = (int)helicsInputGetOption(arg1,arg2);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputSetOption(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  int arg2 ;
  int arg3 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsInputSetOption", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputSetOption" "', argument " "1"" of type '" "helics_input""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsInputSetOption" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "helicsInputSetOption" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  helicsInputSetOption(arg1,arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsPublicationGetOption(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_publication arg1 = (helics_publication) 0 ;
  int arg2 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  int result;
  
  if (!SWIG_Python_UnpackTuple(args, "helicsPublicationGetOption", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsPublicationGetOption" "', argument " "1"" of type '" "helics_publication""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsPublicationGetOption" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  result = (int)helicsPublicationGetOption(arg1,arg2);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsPublicationSetOption(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_publication arg1 = (helics_publication) 0 ;
  int arg2 ;
  int arg3 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsPublicationSetOption", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsPublicationSetOption" "', argument " "1"" of type '" "helics_publication""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsPublicationSetOption" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "helicsPublicationSetOption" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  helicsPublicationSetOption(arg1,arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsPublicationSetMinimumChange(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_publication arg1 = (helics_publication) 0 ;
  double arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  double val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsPublicationSetMinimumChange", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsPublicationSetMinimumChange" "', argument " "1"" of type '" "helics_publication""'"); 
  }
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsPublicationSetMinimumChange" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = (double)(val2);
  helicsPublicationSetMinimumChange(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputSetMinimumChange(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  double arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  double val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsInputSetMinimumChange", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputSetMinimumChange" "', argument " "1"" of type '" "helics_input""'"); 
  }
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsInputSetMinimumChange" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = (double)(val2);
  helicsInputSetMinimumChange(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputIsUpdated(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  helics_bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputIsUpdated" "', argument " "1"" of type '" "helics_input""'"); 
  }
  result = (helics_bool)helicsInputIsUpdated(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputLastUpdateTime(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  helics_time result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputLastUpdateTime" "', argument " "1"" of type '" "helics_input""'"); 
  }
  result = (helics_time)helicsInputLastUpdateTime(arg1);
  resultobj = SWIG_From_double((double)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsInputClearUpdate(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_input arg1 = (helics_input) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsInputClearUpdate" "', argument " "1"" of type '" "helics_input""'"); 
  }
  helicsInputClearUpdate(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateGetPublicationCount(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateGetPublicationCount" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  result = (int)helicsFederateGetPublicationCount(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateGetInputCount(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateGetInputCount" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  result = (int)helicsFederateGetInputCount(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateRegisterEndpoint(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  helics_endpoint result;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateRegisterEndpoint", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateRegisterEndpoint" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateRegisterEndpoint" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "helicsFederateRegisterEndpoint" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  result = (helics_endpoint)helicsFederateRegisterEndpoint(arg1,(char const *)arg2,(char const *)arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateRegisterGlobalEndpoint(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  helics_endpoint result;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateRegisterGlobalEndpoint", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateRegisterGlobalEndpoint" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateRegisterGlobalEndpoint" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "helicsFederateRegisterGlobalEndpoint" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  result = (helics_endpoint)helicsFederateRegisterGlobalEndpoint(arg1,(char const *)arg2,(char const *)arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateGetEndpoint(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  helics_endpoint result;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateGetEndpoint", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateGetEndpoint" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateGetEndpoint" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  result = (helics_endpoint)helicsFederateGetEndpoint(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateGetEndpointByIndex(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  int arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  helics_endpoint result;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateGetEndpointByIndex", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateGetEndpointByIndex" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFederateGetEndpointByIndex" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  result = (helics_endpoint)helicsFederateGetEndpointByIndex(arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsEndpointIsValid(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_endpoint arg1 = (helics_endpoint) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  helics_bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsEndpointIsValid" "', argument " "1"" of type '" "helics_endpoint""'"); 
  }
  result = (helics_bool)helicsEndpointIsValid(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsEndpointSetDefaultDestination(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_endpoint arg1 = (helics_endpoint) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsEndpointSetDefaultDestination", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsEndpointSetDefaultDestination" "', argument " "1"" of type '" "helics_endpoint""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsEndpointSetDefaultDestination" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsEndpointSetDefaultDestination(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsEndpointGetDefaultDestination(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_endpoint arg1 = (helics_endpoint) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsEndpointGetDefaultDestination" "', argument " "1"" of type '" "helics_endpoint""'"); 
  }
  result = (char *)helicsEndpointGetDefaultDestination(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsEndpointSendMessageRaw(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_endpoint arg1 = (helics_endpoint) 0 ;
  char *arg2 = (char *) 0 ;
  void *arg3 = (void *) 0 ;
  int arg4 ;
  helics_error *arg5 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp5 ;
  PyObject *swig_obj[3] ;
  
  {
    etemp5=helicsErrorInitialize();
    arg5=&etemp5;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsEndpointSendMessageRaw", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsEndpointSendMessageRaw" "', argument " "1"" of type '" "helics_endpoint""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsEndpointSendMessageRaw" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  {
    if (PyUnicode_Check(swig_obj[2])) {
      int kind=PyUnicode_KIND(swig_obj[2]);
      arg3=PyUnicode_DATA(swig_obj[2]);
      switch(kind)
      {
      case PyUnicode_1BYTE_KIND:
      default:
        arg4=(int)(PyUnicode_GetLength(swig_obj[2]));
        break;
      case PyUnicode_2BYTE_KIND:
      case PyUnicode_WCHAR_KIND:
        arg4=(int)(PyUnicode_GetLength(swig_obj[2])*2);
        break;
      case PyUnicode_4BYTE_KIND:
        arg4=(int)(PyUnicode_GetLength(swig_obj[2])*4);
        break;
      }
    }
    else if (PyBytes_Check(swig_obj[2])) {
      arg3=PyBytes_AsString(swig_obj[2]);
      arg4=(int)(PyBytes_Size(swig_obj[2]));
    }
    else
    {
      PyErr_SetString(PyExc_ValueError,"Expected a string or bytes");
      return NULL;
    }
  }
  helicsEndpointSendMessageRaw(arg1,(char const *)arg2,(void const *)arg3,arg4,arg5);
  resultobj = SWIG_Py_Void();
  {
    
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsEndpointSendEventRaw(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_endpoint arg1 = (helics_endpoint) 0 ;
  char *arg2 = (char *) 0 ;
  void *arg3 = (void *) 0 ;
  int arg4 ;
  helics_time arg5 ;
  helics_error *arg6 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  double val5 ;
  int ecode5 = 0 ;
  helics_error etemp6 ;
  PyObject *swig_obj[4] ;
  
  {
    etemp6=helicsErrorInitialize();
    arg6=&etemp6;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsEndpointSendEventRaw", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsEndpointSendEventRaw" "', argument " "1"" of type '" "helics_endpoint""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsEndpointSendEventRaw" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  {
    if (PyUnicode_Check(swig_obj[2])) {
      int kind=PyUnicode_KIND(swig_obj[2]);
      arg3=PyUnicode_DATA(swig_obj[2]);
      switch(kind)
      {
      case PyUnicode_1BYTE_KIND:
      default:
        arg4=(int)(PyUnicode_GetLength(swig_obj[2]));
        break;
      case PyUnicode_2BYTE_KIND:
      case PyUnicode_WCHAR_KIND:
        arg4=(int)(PyUnicode_GetLength(swig_obj[2])*2);
        break;
      case PyUnicode_4BYTE_KIND:
        arg4=(int)(PyUnicode_GetLength(swig_obj[2])*4);
        break;
      }
    }
    else if (PyBytes_Check(swig_obj[2])) {
      arg3=PyBytes_AsString(swig_obj[2]);
      arg4=(int)(PyBytes_Size(swig_obj[2]));
    }
    else
    {
      PyErr_SetString(PyExc_ValueError,"Expected a string or bytes");
      return NULL;
    }
  }
  ecode5 = SWIG_AsVal_double(swig_obj[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "helicsEndpointSendEventRaw" "', argument " "5"" of type '" "helics_time""'");
  } 
  arg5 = (helics_time)(val5);
  helicsEndpointSendEventRaw(arg1,(char const *)arg2,(void const *)arg3,arg4,arg5,arg6);
  resultobj = SWIG_Py_Void();
  {
    
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg6->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg6);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg6->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg6);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsEndpointSendMessage(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_endpoint arg1 = (helics_endpoint) 0 ;
  helics_message *arg2 = (helics_message *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsEndpointSendMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsEndpointSendMessage" "', argument " "1"" of type '" "helics_endpoint""'"); 
  }
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_helics_message, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsEndpointSendMessage" "', argument " "2"" of type '" "helics_message *""'"); 
  }
  arg2 = (helics_message *)(argp2);
  helicsEndpointSendMessage(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsEndpointSendMessageObject(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_endpoint arg1 = (helics_endpoint) 0 ;
  helics_message_object arg2 = (helics_message_object) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsEndpointSendMessageObject", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsEndpointSendMessageObject" "', argument " "1"" of type '" "helics_endpoint""'"); 
  }
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsEndpointSendMessageObject" "', argument " "2"" of type '" "helics_message_object""'"); 
  }
  helicsEndpointSendMessageObject(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsEndpointSendMessageObjectZeroCopy(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_endpoint arg1 = (helics_endpoint) 0 ;
  helics_message_object arg2 = (helics_message_object) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsEndpointSendMessageObjectZeroCopy", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsEndpointSendMessageObjectZeroCopy" "', argument " "1"" of type '" "helics_endpoint""'"); 
  }
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsEndpointSendMessageObjectZeroCopy" "', argument " "2"" of type '" "helics_message_object""'"); 
  }
  helicsEndpointSendMessageObjectZeroCopy(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsEndpointSubscribe(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_endpoint arg1 = (helics_endpoint) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsEndpointSubscribe", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsEndpointSubscribe" "', argument " "1"" of type '" "helics_endpoint""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsEndpointSubscribe" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsEndpointSubscribe(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateHasMessage(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  helics_bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateHasMessage" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  result = (helics_bool)helicsFederateHasMessage(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsEndpointHasMessage(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_endpoint arg1 = (helics_endpoint) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  helics_bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsEndpointHasMessage" "', argument " "1"" of type '" "helics_endpoint""'"); 
  }
  result = (helics_bool)helicsEndpointHasMessage(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederatePendingMessages(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederatePendingMessages" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  result = (int)helicsFederatePendingMessages(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsEndpointPendingMessages(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_endpoint arg1 = (helics_endpoint) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsEndpointPendingMessages" "', argument " "1"" of type '" "helics_endpoint""'"); 
  }
  result = (int)helicsEndpointPendingMessages(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsEndpointGetMessage(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_endpoint arg1 = (helics_endpoint) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  helics_message result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsEndpointGetMessage" "', argument " "1"" of type '" "helics_endpoint""'"); 
  }
  result = helicsEndpointGetMessage(arg1);
  resultobj = SWIG_NewPointerObj((helics_message *)memcpy((helics_message *)calloc(1,sizeof(helics_message)),&result,sizeof(helics_message)), SWIGTYPE_p_helics_message, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsEndpointGetMessageObject(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_endpoint arg1 = (helics_endpoint) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  helics_message_object result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsEndpointGetMessageObject" "', argument " "1"" of type '" "helics_endpoint""'"); 
  }
  result = (helics_message_object)helicsEndpointGetMessageObject(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsEndpointCreateMessageObject(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_endpoint arg1 = (helics_endpoint) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  helics_message_object result;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsEndpointCreateMessageObject" "', argument " "1"" of type '" "helics_endpoint""'"); 
  }
  result = (helics_message_object)helicsEndpointCreateMessageObject(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateGetMessage(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  helics_message result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateGetMessage" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  result = helicsFederateGetMessage(arg1);
  resultobj = SWIG_NewPointerObj((helics_message *)memcpy((helics_message *)calloc(1,sizeof(helics_message)),&result,sizeof(helics_message)), SWIGTYPE_p_helics_message, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateGetMessageObject(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  helics_message_object result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateGetMessageObject" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  result = (helics_message_object)helicsFederateGetMessageObject(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateCreateMessageObject(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  helics_message_object result;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateCreateMessageObject" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  result = (helics_message_object)helicsFederateCreateMessageObject(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateClearMessages(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateClearMessages" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  helicsFederateClearMessages(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsEndpointClearMessages(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_endpoint arg1 = (helics_endpoint) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsEndpointClearMessages" "', argument " "1"" of type '" "helics_endpoint""'"); 
  }
  helicsEndpointClearMessages(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsEndpointGetType(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_endpoint arg1 = (helics_endpoint) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsEndpointGetType" "', argument " "1"" of type '" "helics_endpoint""'"); 
  }
  result = (char *)helicsEndpointGetType(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsEndpointGetName(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_endpoint arg1 = (helics_endpoint) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsEndpointGetName" "', argument " "1"" of type '" "helics_endpoint""'"); 
  }
  result = (char *)helicsEndpointGetName(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateGetEndpointCount(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateGetEndpointCount" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  result = (int)helicsFederateGetEndpointCount(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsEndpointGetInfo(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_endpoint arg1 = (helics_endpoint) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsEndpointGetInfo" "', argument " "1"" of type '" "helics_endpoint""'"); 
  }
  result = (char *)helicsEndpointGetInfo(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsEndpointSetInfo(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_endpoint arg1 = (helics_endpoint) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsEndpointSetInfo", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsEndpointSetInfo" "', argument " "1"" of type '" "helics_endpoint""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsEndpointSetInfo" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsEndpointSetInfo(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsEndpointSetOption(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_endpoint arg1 = (helics_endpoint) 0 ;
  int arg2 ;
  int arg3 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsEndpointSetOption", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsEndpointSetOption" "', argument " "1"" of type '" "helics_endpoint""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsEndpointSetOption" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "helicsEndpointSetOption" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  helicsEndpointSetOption(arg1,arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsEndpointGetOption(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_endpoint arg1 = (helics_endpoint) 0 ;
  int arg2 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  int result;
  
  if (!SWIG_Python_UnpackTuple(args, "helicsEndpointGetOption", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsEndpointGetOption" "', argument " "1"" of type '" "helics_endpoint""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsEndpointGetOption" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  result = (int)helicsEndpointGetOption(arg1,arg2);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsMessageGetSource(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_message_object arg1 = (helics_message_object) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsMessageGetSource" "', argument " "1"" of type '" "helics_message_object""'"); 
  }
  result = (char *)helicsMessageGetSource(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsMessageGetDestination(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_message_object arg1 = (helics_message_object) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsMessageGetDestination" "', argument " "1"" of type '" "helics_message_object""'"); 
  }
  result = (char *)helicsMessageGetDestination(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsMessageGetOriginalSource(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_message_object arg1 = (helics_message_object) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsMessageGetOriginalSource" "', argument " "1"" of type '" "helics_message_object""'"); 
  }
  result = (char *)helicsMessageGetOriginalSource(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsMessageGetOriginalDestination(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_message_object arg1 = (helics_message_object) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsMessageGetOriginalDestination" "', argument " "1"" of type '" "helics_message_object""'"); 
  }
  result = (char *)helicsMessageGetOriginalDestination(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsMessageGetTime(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_message_object arg1 = (helics_message_object) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  helics_time result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsMessageGetTime" "', argument " "1"" of type '" "helics_message_object""'"); 
  }
  result = (helics_time)helicsMessageGetTime(arg1);
  resultobj = SWIG_From_double((double)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsMessageGetString(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_message_object arg1 = (helics_message_object) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsMessageGetString" "', argument " "1"" of type '" "helics_message_object""'"); 
  }
  result = (char *)helicsMessageGetString(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsMessageGetMessageID(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_message_object arg1 = (helics_message_object) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsMessageGetMessageID" "', argument " "1"" of type '" "helics_message_object""'"); 
  }
  result = (int)helicsMessageGetMessageID(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsMessageCheckFlag(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_message_object arg1 = (helics_message_object) 0 ;
  int arg2 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  helics_bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "helicsMessageCheckFlag", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsMessageCheckFlag" "', argument " "1"" of type '" "helics_message_object""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsMessageCheckFlag" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  result = (helics_bool)helicsMessageCheckFlag(arg1,arg2);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsMessageGetRawDataSize(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_message_object arg1 = (helics_message_object) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsMessageGetRawDataSize" "', argument " "1"" of type '" "helics_message_object""'"); 
  }
  result = (int)helicsMessageGetRawDataSize(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsMessageGetRawData(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_message_object arg1 = (helics_message_object) 0 ;
  void *arg2 = (void *) 0 ;
  int arg3 ;
  int *arg4 = (int *) 0 ;
  helics_error *arg5 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp5 ;
  PyObject *swig_obj[1] ;
  
  {
    arg4=&(arg3);
  }
  {
    etemp5=helicsErrorInitialize();
    arg5=&etemp5;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsMessageGetRawData" "', argument " "1"" of type '" "helics_message_object""'"); 
  }
  {
    arg3=helicsMessageGetRawDataSize(arg1)+2;
    arg2 =  malloc(arg3);
  }
  helicsMessageGetRawData(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_Py_Void();
  {
    PyObject *o2=PyBytes_FromStringAndSize(arg2,*arg4);
    resultobj = SWIG_Python_AppendOutput(resultobj, o2);
  }
  {
    if (arg2) free(arg2);
  }
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2) free(arg2);
  }
  {
    if (arg5->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg5);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsMessageIsValid(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_message_object arg1 = (helics_message_object) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  helics_bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsMessageIsValid" "', argument " "1"" of type '" "helics_message_object""'"); 
  }
  result = (helics_bool)helicsMessageIsValid(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsMessageSetSource(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_message_object arg1 = (helics_message_object) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsMessageSetSource", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsMessageSetSource" "', argument " "1"" of type '" "helics_message_object""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsMessageSetSource" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsMessageSetSource(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsMessageSetDestination(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_message_object arg1 = (helics_message_object) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsMessageSetDestination", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsMessageSetDestination" "', argument " "1"" of type '" "helics_message_object""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsMessageSetDestination" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsMessageSetDestination(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsMessageSetOriginalSource(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_message_object arg1 = (helics_message_object) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsMessageSetOriginalSource", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsMessageSetOriginalSource" "', argument " "1"" of type '" "helics_message_object""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsMessageSetOriginalSource" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsMessageSetOriginalSource(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsMessageSetOriginalDestination(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_message_object arg1 = (helics_message_object) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsMessageSetOriginalDestination", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsMessageSetOriginalDestination" "', argument " "1"" of type '" "helics_message_object""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsMessageSetOriginalDestination" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsMessageSetOriginalDestination(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsMessageSetTime(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_message_object arg1 = (helics_message_object) 0 ;
  helics_time arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  double val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsMessageSetTime", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsMessageSetTime" "', argument " "1"" of type '" "helics_message_object""'"); 
  }
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsMessageSetTime" "', argument " "2"" of type '" "helics_time""'");
  } 
  arg2 = (helics_time)(val2);
  helicsMessageSetTime(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsMessageReserve(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_message_object arg1 = (helics_message_object) 0 ;
  int arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsMessageReserve", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsMessageReserve" "', argument " "1"" of type '" "helics_message_object""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsMessageReserve" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  helicsMessageReserve(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsMessageSetMessageID(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_message_object arg1 = (helics_message_object) 0 ;
  int32_t arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  void *argp2 ;
  int res2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsMessageSetMessageID", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsMessageSetMessageID" "', argument " "1"" of type '" "helics_message_object""'"); 
  }
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_int32_t,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsMessageSetMessageID" "', argument " "2"" of type '" "int32_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "helicsMessageSetMessageID" "', argument " "2"" of type '" "int32_t""'");
    } else {
      arg2 = *((int32_t *)(argp2));
    }
  }
  helicsMessageSetMessageID(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsMessageClearFlags(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_message_object arg1 = (helics_message_object) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsMessageClearFlags" "', argument " "1"" of type '" "helics_message_object""'"); 
  }
  helicsMessageClearFlags(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsMessageSetFlagOption(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_message_object arg1 = (helics_message_object) 0 ;
  int arg2 ;
  helics_bool arg3 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsMessageSetFlagOption", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsMessageSetFlagOption" "', argument " "1"" of type '" "helics_message_object""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsMessageSetFlagOption" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "helicsMessageSetFlagOption" "', argument " "3"" of type '" "helics_bool""'");
  } 
  arg3 = (helics_bool)(val3);
  helicsMessageSetFlagOption(arg1,arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsMessageSetString(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_message_object arg1 = (helics_message_object) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsMessageSetString", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsMessageSetString" "', argument " "1"" of type '" "helics_message_object""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsMessageSetString" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsMessageSetString(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsMessageSetData(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_message_object arg1 = (helics_message_object) 0 ;
  void *arg2 = (void *) 0 ;
  int arg3 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp4 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsMessageSetData", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsMessageSetData" "', argument " "1"" of type '" "helics_message_object""'"); 
  }
  {
    if (PyUnicode_Check(swig_obj[1])) {
      int kind=PyUnicode_KIND(swig_obj[1]);
      arg2=PyUnicode_DATA(swig_obj[1]);
      switch(kind)
      {
      case PyUnicode_1BYTE_KIND:
      default:
        arg3=(int)(PyUnicode_GetLength(swig_obj[1]));
        break;
      case PyUnicode_2BYTE_KIND:
      case PyUnicode_WCHAR_KIND:
        arg3=(int)(PyUnicode_GetLength(swig_obj[1])*2);
        break;
      case PyUnicode_4BYTE_KIND:
        arg3=(int)(PyUnicode_GetLength(swig_obj[1])*4);
        break;
      }
    }
    else if (PyBytes_Check(swig_obj[1])) {
      arg2=PyBytes_AsString(swig_obj[1]);
      arg3=(int)(PyBytes_Size(swig_obj[1]));
    }
    else
    {
      PyErr_SetString(PyExc_ValueError,"Expected a string or bytes");
      return NULL;
    }
  }
  helicsMessageSetData(arg1,(void const *)arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  {
    
  }
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsMessageAppendData(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_message_object arg1 = (helics_message_object) 0 ;
  void *arg2 = (void *) 0 ;
  int arg3 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp4 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsMessageAppendData", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsMessageAppendData" "', argument " "1"" of type '" "helics_message_object""'"); 
  }
  {
    if (PyUnicode_Check(swig_obj[1])) {
      int kind=PyUnicode_KIND(swig_obj[1]);
      arg2=PyUnicode_DATA(swig_obj[1]);
      switch(kind)
      {
      case PyUnicode_1BYTE_KIND:
      default:
        arg3=(int)(PyUnicode_GetLength(swig_obj[1]));
        break;
      case PyUnicode_2BYTE_KIND:
      case PyUnicode_WCHAR_KIND:
        arg3=(int)(PyUnicode_GetLength(swig_obj[1])*2);
        break;
      case PyUnicode_4BYTE_KIND:
        arg3=(int)(PyUnicode_GetLength(swig_obj[1])*4);
        break;
      }
    }
    else if (PyBytes_Check(swig_obj[1])) {
      arg2=PyBytes_AsString(swig_obj[1]);
      arg3=(int)(PyBytes_Size(swig_obj[1]));
    }
    else
    {
      PyErr_SetString(PyExc_ValueError,"Expected a string or bytes");
      return NULL;
    }
  }
  helicsMessageAppendData(arg1,(void const *)arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  {
    
  }
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsMessageCopy(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_message_object arg1 = (helics_message_object) 0 ;
  helics_message_object arg2 = (helics_message_object) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsMessageCopy", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsMessageCopy" "', argument " "1"" of type '" "helics_message_object""'"); 
  }
  res2 = SWIG_ConvertPtr(swig_obj[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsMessageCopy" "', argument " "2"" of type '" "helics_message_object""'"); 
  }
  helicsMessageCopy(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsMessageClone(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_message_object arg1 = (helics_message_object) 0 ;
  helics_error *arg2 = (helics_error *) 0 ;
  int res1 ;
  helics_error etemp2 ;
  PyObject *swig_obj[1] ;
  helics_message_object result;
  
  {
    etemp2=helicsErrorInitialize();
    arg2=&etemp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsMessageClone" "', argument " "1"" of type '" "helics_message_object""'"); 
  }
  result = (helics_message_object)helicsMessageClone(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg2->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg2);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsMessageFree(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_message_object arg1 = (helics_message_object) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsMessageFree" "', argument " "1"" of type '" "helics_message_object""'"); 
  }
  helicsMessageFree(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateRegisterFilter(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  helics_filter_type arg2 ;
  char *arg3 = (char *) 0 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  helics_filter result;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateRegisterFilter", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateRegisterFilter" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFederateRegisterFilter" "', argument " "2"" of type '" "helics_filter_type""'");
  } 
  arg2 = (helics_filter_type)(val2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "helicsFederateRegisterFilter" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  result = (helics_filter)helicsFederateRegisterFilter(arg1,arg2,(char const *)arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateRegisterGlobalFilter(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  helics_filter_type arg2 ;
  char *arg3 = (char *) 0 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  helics_filter result;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateRegisterGlobalFilter", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateRegisterGlobalFilter" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFederateRegisterGlobalFilter" "', argument " "2"" of type '" "helics_filter_type""'");
  } 
  arg2 = (helics_filter_type)(val2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "helicsFederateRegisterGlobalFilter" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  result = (helics_filter)helicsFederateRegisterGlobalFilter(arg1,arg2,(char const *)arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateRegisterCloningFilter(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  helics_filter result;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateRegisterCloningFilter", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateRegisterCloningFilter" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateRegisterCloningFilter" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  result = (helics_filter)helicsFederateRegisterCloningFilter(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateRegisterGlobalCloningFilter(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  helics_filter result;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateRegisterGlobalCloningFilter", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateRegisterGlobalCloningFilter" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateRegisterGlobalCloningFilter" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  result = (helics_filter)helicsFederateRegisterGlobalCloningFilter(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCoreRegisterFilter(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_core arg1 = (helics_core) 0 ;
  helics_filter_type arg2 ;
  char *arg3 = (char *) 0 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  helics_filter result;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsCoreRegisterFilter", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCoreRegisterFilter" "', argument " "1"" of type '" "helics_core""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsCoreRegisterFilter" "', argument " "2"" of type '" "helics_filter_type""'");
  } 
  arg2 = (helics_filter_type)(val2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "helicsCoreRegisterFilter" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  result = (helics_filter)helicsCoreRegisterFilter(arg1,arg2,(char const *)arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsCoreRegisterCloningFilter(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_core arg1 = (helics_core) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  helics_filter result;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsCoreRegisterCloningFilter", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsCoreRegisterCloningFilter" "', argument " "1"" of type '" "helics_core""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsCoreRegisterCloningFilter" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  result = (helics_filter)helicsCoreRegisterCloningFilter(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateGetFilterCount(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateGetFilterCount" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  result = (int)helicsFederateGetFilterCount(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateGetFilter(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  helics_filter result;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateGetFilter", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateGetFilter" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFederateGetFilter" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  result = (helics_filter)helicsFederateGetFilter(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFederateGetFilterByIndex(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_federate arg1 = (helics_federate) 0 ;
  int arg2 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  helics_filter result;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFederateGetFilterByIndex", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFederateGetFilterByIndex" "', argument " "1"" of type '" "helics_federate""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFederateGetFilterByIndex" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  result = (helics_filter)helicsFederateGetFilterByIndex(arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFilterIsValid(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_filter arg1 = (helics_filter) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  helics_bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFilterIsValid" "', argument " "1"" of type '" "helics_filter""'"); 
  }
  result = (helics_bool)helicsFilterIsValid(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFilterGetName(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_filter arg1 = (helics_filter) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFilterGetName" "', argument " "1"" of type '" "helics_filter""'"); 
  }
  result = (char *)helicsFilterGetName(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFilterSet(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_filter arg1 = (helics_filter) 0 ;
  char *arg2 = (char *) 0 ;
  double arg3 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFilterSet", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFilterSet" "', argument " "1"" of type '" "helics_filter""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFilterSet" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  ecode3 = SWIG_AsVal_double(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "helicsFilterSet" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = (double)(val3);
  helicsFilterSet(arg1,(char const *)arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFilterSetString(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_filter arg1 = (helics_filter) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFilterSetString", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFilterSetString" "', argument " "1"" of type '" "helics_filter""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFilterSetString" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "helicsFilterSetString" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  helicsFilterSetString(arg1,(char const *)arg2,(char const *)arg3,arg4);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFilterAddDestinationTarget(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_filter arg1 = (helics_filter) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFilterAddDestinationTarget", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFilterAddDestinationTarget" "', argument " "1"" of type '" "helics_filter""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFilterAddDestinationTarget" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsFilterAddDestinationTarget(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFilterAddSourceTarget(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_filter arg1 = (helics_filter) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFilterAddSourceTarget", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFilterAddSourceTarget" "', argument " "1"" of type '" "helics_filter""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFilterAddSourceTarget" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsFilterAddSourceTarget(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFilterAddDeliveryEndpoint(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_filter arg1 = (helics_filter) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFilterAddDeliveryEndpoint", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFilterAddDeliveryEndpoint" "', argument " "1"" of type '" "helics_filter""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFilterAddDeliveryEndpoint" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsFilterAddDeliveryEndpoint(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFilterRemoveTarget(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_filter arg1 = (helics_filter) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFilterRemoveTarget", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFilterRemoveTarget" "', argument " "1"" of type '" "helics_filter""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFilterRemoveTarget" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsFilterRemoveTarget(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFilterRemoveDeliveryEndpoint(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_filter arg1 = (helics_filter) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFilterRemoveDeliveryEndpoint", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFilterRemoveDeliveryEndpoint" "', argument " "1"" of type '" "helics_filter""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFilterRemoveDeliveryEndpoint" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsFilterRemoveDeliveryEndpoint(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFilterGetInfo(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_filter arg1 = (helics_filter) 0 ;
  int res1 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFilterGetInfo" "', argument " "1"" of type '" "helics_filter""'"); 
  }
  result = (char *)helicsFilterGetInfo(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFilterSetInfo(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_filter arg1 = (helics_filter) 0 ;
  char *arg2 = (char *) 0 ;
  helics_error *arg3 = (helics_error *) 0 ;
  int res1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  helics_error etemp3 ;
  PyObject *swig_obj[2] ;
  
  {
    etemp3=helicsErrorInitialize();
    arg3=&etemp3;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFilterSetInfo", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFilterSetInfo" "', argument " "1"" of type '" "helics_filter""'"); 
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "helicsFilterSetInfo" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  helicsFilterSetInfo(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    if (arg3->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg3);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFilterSetOption(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_filter arg1 = (helics_filter) 0 ;
  int arg2 ;
  int arg3 ;
  helics_error *arg4 = (helics_error *) 0 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  helics_error etemp4 ;
  PyObject *swig_obj[3] ;
  
  {
    etemp4=helicsErrorInitialize();
    arg4=&etemp4;
  }
  if (!SWIG_Python_UnpackTuple(args, "helicsFilterSetOption", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFilterSetOption" "', argument " "1"" of type '" "helics_filter""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFilterSetOption" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "helicsFilterSetOption" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  helicsFilterSetOption(arg1,arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return resultobj;
fail:
  {
    if (arg4->error_code!=helics_ok)
    {
      throwHelicsPythonException(arg4);
      return NULL;
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_helicsFilterGetOption(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  helics_filter arg1 = (helics_filter) 0 ;
  int arg2 ;
  int res1 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  int result;
  
  if (!SWIG_Python_UnpackTuple(args, "helicsFilterGetOption", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "helicsFilterGetOption" "', argument " "1"" of type '" "helics_filter""'"); 
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "helicsFilterGetOption" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  result = (int)helicsFilterGetOption(arg1,arg2);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


static PyMethodDef SwigMethods[] = {
	 { "SWIG_PyInstanceMethod_New", SWIG_PyInstanceMethod_New, METH_O, NULL},
	 { "helics_complex_real_set", _wrap_helics_complex_real_set, METH_VARARGS, NULL},
	 { "helics_complex_real_get", _wrap_helics_complex_real_get, METH_O, NULL},
	 { "helics_complex_imag_set", _wrap_helics_complex_imag_set, METH_VARARGS, NULL},
	 { "helics_complex_imag_get", _wrap_helics_complex_imag_get, METH_O, NULL},
	 { "new_helics_complex", _wrap_new_helics_complex, METH_NOARGS, NULL},
	 { "delete_helics_complex", _wrap_delete_helics_complex, METH_O, NULL},
	 { "helics_complex_swigregister", helics_complex_swigregister, METH_O, NULL},
	 { "helics_complex_swiginit", helics_complex_swiginit, METH_VARARGS, NULL},
	 { "helics_message_time_set", _wrap_helics_message_time_set, METH_VARARGS, " message time"},
	 { "helics_message_time_get", _wrap_helics_message_time_get, METH_O, " message time"},
	 { "helics_message_data_set", _wrap_helics_message_data_set, METH_VARARGS, " message data"},
	 { "helics_message_data_get", _wrap_helics_message_data_get, METH_O, " message data"},
	 { "helics_message_length_set", _wrap_helics_message_length_set, METH_VARARGS, " message length"},
	 { "helics_message_length_get", _wrap_helics_message_length_get, METH_O, " message length"},
	 { "helics_message_messageID_set", _wrap_helics_message_messageID_set, METH_VARARGS, " message identification information"},
	 { "helics_message_messageID_get", _wrap_helics_message_messageID_get, METH_O, " message identification information"},
	 { "helics_message_flags_set", _wrap_helics_message_flags_set, METH_VARARGS, " flags related to the message"},
	 { "helics_message_flags_get", _wrap_helics_message_flags_get, METH_O, " flags related to the message"},
	 { "helics_message_original_source_set", _wrap_helics_message_original_source_set, METH_VARARGS, " original source"},
	 { "helics_message_original_source_get", _wrap_helics_message_original_source_get, METH_O, " original source"},
	 { "helics_message_source_set", _wrap_helics_message_source_set, METH_VARARGS, " the most recent source"},
	 { "helics_message_source_get", _wrap_helics_message_source_get, METH_O, " the most recent source"},
	 { "helics_message_dest_set", _wrap_helics_message_dest_set, METH_VARARGS, " the final destination"},
	 { "helics_message_dest_get", _wrap_helics_message_dest_get, METH_O, " the final destination"},
	 { "helics_message_original_dest_set", _wrap_helics_message_original_dest_set, METH_VARARGS, " the original destination of the message"},
	 { "helics_message_original_dest_get", _wrap_helics_message_original_dest_get, METH_O, " the original destination of the message"},
	 { "new_helics_message", _wrap_new_helics_message, METH_NOARGS, NULL},
	 { "delete_helics_message", _wrap_delete_helics_message, METH_O, NULL},
	 { "helics_message_swigregister", helics_message_swigregister, METH_O, NULL},
	 { "helics_message_swiginit", helics_message_swiginit, METH_VARARGS, NULL},
	 { "helicsGetVersion", _wrap_helicsGetVersion, METH_NOARGS, "Get a version string for HELICS."},
	 { "helicsGetBuildFlags", _wrap_helicsGetBuildFlags, METH_NOARGS, "Get the build flags used to compile HELICS."},
	 { "helicsGetCompilerVersion", _wrap_helicsGetCompilerVersion, METH_NOARGS, "Get the compiler version used to compile HELICS."},
	 { "helicsIsCoreTypeAvailable", _wrap_helicsIsCoreTypeAvailable, METH_O, "\n"
		"Returns true if core/broker type specified is available in current compilation.\n"
		"\n"
		":type type: string\n"
		":param type: A string representing a core type.\n"
		"\n"
		"Options include \"zmq\", \"udp\", \"ipc\", \"interprocess\", \"tcp\", \"default\", \"mpi\".\n"
		""},
	 { "helicsCreateCore", _wrap_helicsCreateCore, METH_VARARGS, "\n"
		"Create a core object.\n"
		"\n"
		":type type: string\n"
		":param type: The type of the core to create.\n"
		":type name: string\n"
		":param name: The name of the core. It can be a nullptr or empty string to have a name automatically assigned.\n"
		":type initString: string\n"
		":param initString: An initialization string to send to the core. The format is similar to command line arguments.\n"
		"                      Typical options include a broker name, the broker address, the number of federates, etc.\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: A helics_core object.\n"
		""},
	 { "helicsCreateCoreFromArgs", _wrap_helicsCreateCoreFromArgs, METH_VARARGS, "\n"
		"Create a core object by passing command line arguments.\n"
		"\n"
		":type type: string\n"
		":param type: The type of the core to create.\n"
		":type name: string\n"
		":param name: The name of the core. It can be a nullptr or empty string to have a name automatically assigned.\n"
		"\n"
		":type argv: string\n"
		":param argv: The list of string values from a command line.\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: A helics_core object.\n"
		""},
	 { "helicsCoreClone", _wrap_helicsCoreClone, METH_O, "\n"
		"Create a new reference to an existing core.\n"
		"\n"
		"This will create a new broker object that references the existing broker. The new broker object must be freed as well.\n"
		"\n"
		":type core: void\n"
		":param core: An existing helics_core.\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: A new reference to the same broker.\n"
		""},
	 { "helicsCoreIsValid", _wrap_helicsCoreIsValid, METH_O, "\n"
		"Check if a core object is a valid object.\n"
		"\n"
		":type core: void\n"
		":param core: The helics_core object to test.\n"
		""},
	 { "helicsCreateBroker", _wrap_helicsCreateBroker, METH_VARARGS, "\n"
		"Create a broker object.\n"
		"\n"
		":type type: string\n"
		":param type: The type of the broker to create.\n"
		":type name: string\n"
		":param name: The name of the broker. It can be a nullptr or empty string to have a name automatically assigned.\n"
		":type initString: string\n"
		":param initString: An initialization string to send to the core-the format is similar to command line arguments.\n"
		"                      Typical options include a broker address such as --broker=\"XSSAF\" if this is a subbroker, or the number of federates,\n"
		"    or the address.\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: A helics_broker object.\n"
		""},
	 { "helicsCreateBrokerFromArgs", _wrap_helicsCreateBrokerFromArgs, METH_VARARGS, "\n"
		"Create a core object by passing command line arguments.\n"
		"\n"
		":type type: string\n"
		":param type: The type of the core to create.\n"
		":type name: string\n"
		":param name: The name of the core. It can be a nullptr or empty string to have a name automatically assigned.\n"
		"\n"
		":type argv: string\n"
		":param argv: The list of string values from a command line.\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: A helics_core object.\n"
		""},
	 { "helicsBrokerClone", _wrap_helicsBrokerClone, METH_O, "\n"
		"Create a new reference to an existing broker.\n"
		"\n"
		"This will create a new broker object that references the existing broker it must be freed as well.\n"
		"\n"
		":type broker: void\n"
		":param broker: An existing helics_broker.\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: A new reference to the same broker.\n"
		""},
	 { "helicsBrokerIsValid", _wrap_helicsBrokerIsValid, METH_O, "\n"
		"Check if a broker object is a valid object.\n"
		"\n"
		":type broker: void\n"
		":param broker: The helics_broker object to test.\n"
		""},
	 { "helicsBrokerIsConnected", _wrap_helicsBrokerIsConnected, METH_O, "\n"
		"Check if a broker is connected.\n"
		"\n"
		"A connected broker implies it is attached to cores or cores could reach out to communicate.\n"
		"\n"
		":rtype: int\n"
		":return: helics_false if not connected.\n"
		""},
	 { "helicsBrokerDataLink", _wrap_helicsBrokerDataLink, METH_VARARGS, "\n"
		"Link a named publication and named input using a broker.\n"
		"\n"
		":type broker: void\n"
		":param broker: The broker to generate the connection from.\n"
		":type source: string\n"
		":param source: The name of the publication (cannot be NULL).\n"
		":type target: string\n"
		":param target: The name of the target to send the publication data (cannot be NULL).\n"
		""},
	 { "helicsBrokerAddSourceFilterToEndpoint", _wrap_helicsBrokerAddSourceFilterToEndpoint, METH_VARARGS, "\n"
		"Link a named filter to a source endpoint.\n"
		"\n"
		":type broker: void\n"
		":param broker: The broker to generate the connection from.\n"
		":type filter: string\n"
		":param filter: The name of the filter (cannot be NULL).\n"
		":type endpoint: string\n"
		":param endpoint: The name of the endpoint to filter the data from (cannot be NULL).\n"
		""},
	 { "helicsBrokerAddDestinationFilterToEndpoint", _wrap_helicsBrokerAddDestinationFilterToEndpoint, METH_VARARGS, "\n"
		"Link a named filter to a destination endpoint.\n"
		"\n"
		":type broker: void\n"
		":param broker: The broker to generate the connection from.\n"
		":type filter: string\n"
		":param filter: The name of the filter (cannot be NULL).\n"
		":type endpoint: string\n"
		":param endpoint: The name of the endpoint to filter the data going to (cannot be NULL).\n"
		""},
	 { "helicsBrokerMakeConnections", _wrap_helicsBrokerMakeConnections, METH_VARARGS, "\n"
		"Load a file containing connection information.\n"
		"\n"
		":type broker: void\n"
		":param broker: The broker to generate the connections from.\n"
		":type file: string\n"
		":param file: A JSON or TOML file containing connection information.\n"
		""},
	 { "helicsCoreWaitForDisconnect", _wrap_helicsCoreWaitForDisconnect, METH_VARARGS, "\n"
		"Wait for the core to disconnect.\n"
		"\n"
		":type core: void\n"
		":param core: The core to wait for.\n"
		":type msToWait: int\n"
		":param msToWait: The time out in millisecond (<0 for infinite timeout).\n"
		"\n"
		"\n"
		":rtype: int\n"
		":return: helics_true if the disconnect was successful, helics_false if there was a timeout.\n"
		""},
	 { "helicsBrokerWaitForDisconnect", _wrap_helicsBrokerWaitForDisconnect, METH_VARARGS, "\n"
		"Wait for the broker to disconnect.\n"
		"\n"
		":type broker: void\n"
		":param broker: The broker to wait for.\n"
		":type msToWait: int\n"
		":param msToWait: The time out in millisecond (<0 for infinite timeout).\n"
		"\n"
		"\n"
		":rtype: int\n"
		":return: helics_true if the disconnect was successful, helics_false if there was a timeout.\n"
		""},
	 { "helicsCoreIsConnected", _wrap_helicsCoreIsConnected, METH_O, "\n"
		"Check if a core is connected.\n"
		"\n"
		"A connected core implies it is attached to federates or federates could be attached to it\n"
		"\n"
		":rtype: int\n"
		":return: helics_false if not connected, helics_true if it is connected.\n"
		""},
	 { "helicsCoreDataLink", _wrap_helicsCoreDataLink, METH_VARARGS, "\n"
		"Link a named publication and named input using a core.\n"
		"\n"
		":type core: void\n"
		":param core: The core to generate the connection from.\n"
		":type source: string\n"
		":param source: The name of the publication (cannot be NULL).\n"
		":type target: string\n"
		":param target: The name of the target to send the publication data (cannot be NULL).\n"
		""},
	 { "helicsCoreAddSourceFilterToEndpoint", _wrap_helicsCoreAddSourceFilterToEndpoint, METH_VARARGS, "\n"
		"Link a named filter to a source endpoint.\n"
		"\n"
		":type core: void\n"
		":param core: The core to generate the connection from.\n"
		":type filter: string\n"
		":param filter: The name of the filter (cannot be NULL).\n"
		":type endpoint: string\n"
		":param endpoint: The name of the endpoint to filter the data from (cannot be NULL).\n"
		""},
	 { "helicsCoreAddDestinationFilterToEndpoint", _wrap_helicsCoreAddDestinationFilterToEndpoint, METH_VARARGS, "\n"
		"Link a named filter to a destination endpoint.\n"
		"\n"
		":type core: void\n"
		":param core: The core to generate the connection from.\n"
		":type filter: string\n"
		":param filter: The name of the filter (cannot be NULL).\n"
		":type endpoint: string\n"
		":param endpoint: The name of the endpoint to filter the data going to (cannot be NULL).\n"
		""},
	 { "helicsCoreMakeConnections", _wrap_helicsCoreMakeConnections, METH_VARARGS, "\n"
		"Load a file containing connection information.\n"
		"\n"
		":type core: void\n"
		":param core: The core to generate the connections from.\n"
		":type file: string\n"
		":param file: A JSON or TOML file containing connection information.\n"
		""},
	 { "helicsBrokerGetIdentifier", _wrap_helicsBrokerGetIdentifier, METH_O, "\n"
		"Get an identifier for the broker.\n"
		"\n"
		":type broker: void\n"
		":param broker: The broker to query.\n"
		"\n"
		":rtype: string\n"
		":return: A string containing the identifier for the broker.\n"
		""},
	 { "helicsCoreGetIdentifier", _wrap_helicsCoreGetIdentifier, METH_O, "\n"
		"Get an identifier for the core.\n"
		"\n"
		":type core: void\n"
		":param core: The core to query.\n"
		"\n"
		":rtype: string\n"
		":return: A string with the identifier of the core.\n"
		""},
	 { "helicsBrokerGetAddress", _wrap_helicsBrokerGetAddress, METH_O, "\n"
		"Get the network address associated with a broker.\n"
		"\n"
		":type broker: void\n"
		":param broker: The broker to query.\n"
		"\n"
		":rtype: string\n"
		":return: A string with the network address of the broker.\n"
		""},
	 { "helicsCoreGetAddress", _wrap_helicsCoreGetAddress, METH_O, "\n"
		"Get the network address associated with a core.\n"
		"\n"
		":type core: void\n"
		":param core: The core to query.\n"
		"\n"
		":rtype: string\n"
		":return: A string with the network address of the broker.\n"
		""},
	 { "helicsCoreSetReadyToInit", _wrap_helicsCoreSetReadyToInit, METH_O, "\n"
		"Set the core to ready for init.\n"
		"\n"
		"This function is used for cores that have filters but no federates so there needs to be\n"
		"         a direct signal to the core to trigger the federation initialization.\n"
		"\n"
		":type core: void\n"
		":param core: The core object to enable init values for.\n"
		""},
	 { "helicsCoreConnect", _wrap_helicsCoreConnect, METH_O, "\n"
		"Connect a core to the federate based on current configuration.\n"
		"\n"
		":type core: void\n"
		":param core: The core to connect.\n"
		"\n"
		"\n"
		":rtype: int\n"
		":return: helics_false if not connected, helics_true if it is connected.\n"
		""},
	 { "helicsCoreDisconnect", _wrap_helicsCoreDisconnect, METH_O, "\n"
		"Disconnect a core from the federation.\n"
		"\n"
		":type core: void\n"
		":param core: The core to query.\n"
		""},
	 { "helicsGetFederateByName", _wrap_helicsGetFederateByName, METH_O, "\n"
		"Get an existing federate object from a core by name.\n"
		"\n"
		"The federate must have been created by one of the other functions and at least one of the objects referencing the created\n"
		"         federate must still be active in the process.\n"
		"\n"
		":type fedName: string\n"
		":param fedName: The name of the federate to retrieve.\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: NULL if no fed is available by that name otherwise a helics_federate with that name.\n"
		""},
	 { "helicsBrokerDisconnect", _wrap_helicsBrokerDisconnect, METH_O, "\n"
		"Disconnect a broker.\n"
		"\n"
		":type broker: void\n"
		":param broker: The broker to disconnect.\n"
		""},
	 { "helicsFederateDestroy", _wrap_helicsFederateDestroy, METH_O, "Disconnect and free a federate."},
	 { "helicsBrokerDestroy", _wrap_helicsBrokerDestroy, METH_O, "Disconnect and free a broker."},
	 { "helicsCoreDestroy", _wrap_helicsCoreDestroy, METH_O, "Disconnect and free a core."},
	 { "helicsCoreFree", _wrap_helicsCoreFree, METH_O, "Release the memory associated with a core."},
	 { "helicsBrokerFree", _wrap_helicsBrokerFree, METH_O, "Release the memory associated with a broker."},
	 { "helicsCreateValueFederate", _wrap_helicsCreateValueFederate, METH_VARARGS, "\n"
		"Create a value federate from a federate info object.\n"
		"\n"
		"helics_federate objects can be used in all functions that take a helics_federate or helics_federate object as an argument.\n"
		"\n"
		":type fedName: string\n"
		":param fedName: The name of the federate to create, can NULL or an empty string to use the default name from fi or an assigned name.\n"
		":type fi: void\n"
		":param fi: The federate info object that contains details on the federate.\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: An opaque value federate object.\n"
		""},
	 { "helicsCreateValueFederateFromConfig", _wrap_helicsCreateValueFederateFromConfig, METH_O, "\n"
		"Create a value federate from a JSON file, JSON string, or TOML file.\n"
		"\n"
		"helics_federate objects can be used in all functions that take a helics_federate or helics_federate object as an argument.\n"
		"\n"
		":type configFile: string\n"
		":param configFile: A JSON file or a JSON string or TOML file that contains setup and configuration information.\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: An opaque value federate object.\n"
		""},
	 { "helicsCreateMessageFederate", _wrap_helicsCreateMessageFederate, METH_VARARGS, "\n"
		"Create a message federate from a federate info object.\n"
		"\n"
		"helics_message_federate objects can be used in all functions that take a helics_message_federate or helics_federate object as an\n"
		"argument.\n"
		"\n"
		":type fedName: string\n"
		":param fedName: The name of the federate to create.\n"
		":type fi: void\n"
		":param fi: The federate info object that contains details on the federate.\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: An opaque message federate object.\n"
		""},
	 { "helicsCreateMessageFederateFromConfig", _wrap_helicsCreateMessageFederateFromConfig, METH_O, "\n"
		"Create a message federate from a JSON file or JSON string or TOML file.\n"
		"\n"
		"helics_message_federate objects can be used in all functions that take a helics_message_federate or helics_federate object as an\n"
		"argument.\n"
		"\n"
		":type configFile: string\n"
		":param configFile: A Config(JSON,TOML) file or a JSON string that contains setup and configuration information.\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: An opaque message federate object.\n"
		""},
	 { "helicsCreateCombinationFederate", _wrap_helicsCreateCombinationFederate, METH_VARARGS, "\n"
		"Create a combination federate from a federate info object.\n"
		"\n"
		"Combination federates are both value federates and message federates, objects can be used in all functions\n"
		"                     that take a helics_federate, helics_message_federate or helics_federate object as an argument\n"
		"\n"
		":type fedName: string\n"
		":param fedName: A string with the name of the federate, can be NULL or an empty string to pull the default name from fi.\n"
		":type fi: void\n"
		":param fi: The federate info object that contains details on the federate.\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: An opaque value federate object nullptr if the object creation failed.\n"
		""},
	 { "helicsCreateCombinationFederateFromConfig", _wrap_helicsCreateCombinationFederateFromConfig, METH_O, "\n"
		"Create a combination federate from a JSON file or JSON string or TOML file.\n"
		"\n"
		"Combination federates are both value federates and message federates, objects can be used in all functions\n"
		"         that take a helics_federate, helics_message_federate or helics_federate object as an argument\n"
		"\n"
		":type configFile: string\n"
		":param configFile: A JSON file or a JSON string or TOML file that contains setup and configuration information.\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: An opaque combination federate object.\n"
		""},
	 { "helicsFederateClone", _wrap_helicsFederateClone, METH_O, "\n"
		"Create a new reference to an existing federate.\n"
		"\n"
		"This will create a new helics_federate object that references the existing federate. The new object must be freed as well.\n"
		"\n"
		":type fed: void\n"
		":param fed: An existing helics_federate.\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: A new reference to the same federate.\n"
		""},
	 { "helicsCreateFederateInfo", _wrap_helicsCreateFederateInfo, METH_NOARGS, "\n"
		"Create a federate info object for specifying federate information when constructing a federate.\n"
		"\n"
		":rtype: void\n"
		":return: A helics_federate_info object which is a reference to the created object.\n"
		""},
	 { "helicsFederateInfoClone", _wrap_helicsFederateInfoClone, METH_O, "\n"
		"Create a federate info object from an existing one and clone the information.\n"
		"\n"
		":type fi: void\n"
		":param fi: A federateInfo object to duplicate.\n"
		"\n"
		"\n"
		" :rtype: void\n"
		" :return: A helics_federate_info object which is a reference to the created object.\n"
		""},
	 { "helicsFederateInfoLoadFromArgs", _wrap_helicsFederateInfoLoadFromArgs, METH_VARARGS, "\n"
		"Load federate info from command line arguments.\n"
		"\n"
		":type fi: void\n"
		":param fi: A federateInfo object.\n"
		":type argc: int\n"
		":param argc: The number of command line arguments.\n"
		":type argv: string\n"
		":param argv: An array of strings from the command line.\n"
		""},
	 { "helicsFederateInfoFree", _wrap_helicsFederateInfoFree, METH_O, "Delete the memory associated with a federate info object."},
	 { "helicsFederateIsValid", _wrap_helicsFederateIsValid, METH_O, "\n"
		"Check if a federate_object is valid.\n"
		"\n"
		":rtype: int\n"
		":return: helics_true if the federate is a valid active federate, helics_false otherwise\n"
		""},
	 { "helicsFederateInfoSetCoreName", _wrap_helicsFederateInfoSetCoreName, METH_VARARGS, "\n"
		"Set the name of the core to link to for a federate.\n"
		"\n"
		":type fi: void\n"
		":param fi: The federate info object to alter.\n"
		":type corename: string\n"
		":param corename: The identifier for a core to link to.\n"
		""},
	 { "helicsFederateInfoSetCoreInitString", _wrap_helicsFederateInfoSetCoreInitString, METH_VARARGS, "\n"
		"Set the initialization string for the core usually in the form of command line arguments.\n"
		"\n"
		":type fi: void\n"
		":param fi: The federate info object to alter.\n"
		":type coreInit: string\n"
		":param coreInit: A string containing command line arguments to be passed to the core.\n"
		""},
	 { "helicsFederateInfoSetBrokerInitString", _wrap_helicsFederateInfoSetBrokerInitString, METH_VARARGS, "\n"
		"Set the initialization string that a core will pass to a generated broker usually in the form of command line arguments.\n"
		"\n"
		":type fi: void\n"
		":param fi: The federate info object to alter.\n"
		":type brokerInit: string\n"
		":param brokerInit: A string with command line arguments for a generated broker.\n"
		""},
	 { "helicsFederateInfoSetCoreType", _wrap_helicsFederateInfoSetCoreType, METH_VARARGS, "\n"
		"Set the core type by integer code.\n"
		"\n"
		"Valid values available by definitions in api-data.h.\n"
		":type fi: void\n"
		":param fi: The federate info object to alter.\n"
		":type coretype: int\n"
		":param coretype: An numerical code for a core type see /ref helics_core_type.\n"
		""},
	 { "helicsFederateInfoSetCoreTypeFromString", _wrap_helicsFederateInfoSetCoreTypeFromString, METH_VARARGS, "\n"
		"Set the core type from a string.\n"
		"\n"
		":type fi: void\n"
		":param fi: The federate info object to alter.\n"
		":type coretype: string\n"
		":param coretype: A string naming a core type.\n"
		""},
	 { "helicsFederateInfoSetBroker", _wrap_helicsFederateInfoSetBroker, METH_VARARGS, "\n"
		"Set the name or connection information for a broker.\n"
		"\n"
		"This is only used if the core is automatically created, the broker information will be transferred to the core for connection.\n"
		":type fi: void\n"
		":param fi: The federate info object to alter.\n"
		":type broker: string\n"
		":param broker: A string which defines the connection information for a broker either a name or an address.\n"
		""},
	 { "helicsFederateInfoSetBrokerKey", _wrap_helicsFederateInfoSetBrokerKey, METH_VARARGS, "\n"
		"Set the key for a broker connection.\n"
		"\n"
		"This is only used if the core is automatically created, the broker information will be transferred to the core for connection.\n"
		":type fi: void\n"
		":param fi: The federate info object to alter.\n"
		":type brokerkey: string\n"
		":param brokerkey: A string containing a key for the broker to connect.\n"
		""},
	 { "helicsFederateInfoSetBrokerPort", _wrap_helicsFederateInfoSetBrokerPort, METH_VARARGS, "\n"
		"Set the port to use for the broker.\n"
		"\n"
		"This is only used if the core is automatically created, the broker information will be transferred to the core for connection.\n"
		"This will only be useful for network broker connections.\n"
		":type fi: void\n"
		":param fi: The federate info object to alter.\n"
		":type brokerPort: int\n"
		":param brokerPort: The integer port number to use for connection with a broker.\n"
		""},
	 { "helicsFederateInfoSetLocalPort", _wrap_helicsFederateInfoSetLocalPort, METH_VARARGS, "\n"
		"Set the local port to use.\n"
		"\n"
		"This is only used if the core is automatically created, the port information will be transferred to the core for connection.\n"
		":type fi: void\n"
		":param fi: The federate info object to alter.\n"
		":type localPort: string\n"
		":param localPort: A string with the port information to use as the local server port can be a number or \"auto\" or \"os_local\".\n"
		""},
	 { "helicsGetPropertyIndex", _wrap_helicsGetPropertyIndex, METH_O, "\n"
		"Get a property index for use in /ref helicsFederateInfoSetFlagOption, /ref helicsFederateInfoSetTimeProperty,\n"
		"or /ref helicsFederateInfoSetIntegerProperty\n"
		":type val: string\n"
		":param val: A string with the property name.\n"
		":rtype: int\n"
		":return: An int with the property code or (-1) if not a valid property.\n"
		""},
	 { "helicsGetFlagIndex", _wrap_helicsGetFlagIndex, METH_O, "\n"
		"Get a property index for use in /ref helicsFederateInfoSetFlagOption, /ref helicsFederateSetFlagOption,\n"
		":type val: string\n"
		":param val: A string with the option name.\n"
		":rtype: int\n"
		":return: An int with the property code or (-1) if not a valid property.\n"
		""},
	 { "helicsGetOptionIndex", _wrap_helicsGetOptionIndex, METH_O, "\n"
		"Get an option index for use in /ref helicsPublicationSetOption, /ref helicsInputSetOption, /ref helicsEndpointSetOption,\n"
		"/ref helicsFilterSetOption, and the corresponding get functions.\n"
		"\n"
		":type val: string\n"
		":param val: A string with the option name.\n"
		"\n"
		":rtype: int\n"
		":return: An int with the option index or (-1) if not a valid property.\n"
		""},
	 { "helicsGetOptionValue", _wrap_helicsGetOptionValue, METH_O, "\n"
		"Get an option value for use in /ref helicsPublicationSetOption, /ref helicsInputSetOption, /ref helicsEndpointSetOption,\n"
		"/ref helicsFilterSetOption.\n"
		"\n"
		":type val: string\n"
		":param val: A string representing the value.\n"
		"\n"
		":rtype: int\n"
		":return: An int with the option value or (-1) if not a valid value.\n"
		""},
	 { "helicsFederateInfoSetFlagOption", _wrap_helicsFederateInfoSetFlagOption, METH_VARARGS, "\n"
		"Set a flag in the info structure.\n"
		"\n"
		"Valid flags are available /ref helics_federate_flags.\n"
		":type fi: void\n"
		":param fi: The federate info object to alter.\n"
		":type flag: int\n"
		":param flag: A numerical index for a flag.\n"
		":type value: int\n"
		":param value: The desired value of the flag helics_true or helics_false.\n"
		""},
	 { "helicsFederateInfoSetSeparator", _wrap_helicsFederateInfoSetSeparator, METH_VARARGS, "\n"
		"Set the separator character in the info structure.\n"
		"\n"
		"The separator character is the separation character for local publications/endpoints in creating their global name.\n"
		"For example if the separator character is '/'  then a local endpoint would have a globally reachable name of fedName/localName.\n"
		":type fi: void\n"
		":param fi: The federate info object to alter.\n"
		":type separator: char\n"
		":param separator: The character to use as a separator.\n"
		""},
	 { "helicsFederateInfoSetTimeProperty", _wrap_helicsFederateInfoSetTimeProperty, METH_VARARGS, "\n"
		"Set the output delay for a federate.\n"
		"\n"
		":type fi: void\n"
		":param fi: The federate info object to alter.\n"
		":type timeProperty: int\n"
		":param timeProperty: An integer representation of the time based property to set see /ref helics_properties.\n"
		":type propertyValue: float\n"
		":param propertyValue: The value of the property to set the timeProperty to.\n"
		""},
	 { "helicsFederateInfoSetIntegerProperty", _wrap_helicsFederateInfoSetIntegerProperty, METH_VARARGS, "\n"
		"Set an integer property for a federate.\n"
		"\n"
		"Set known properties.\n"
		"\n"
		":type fi: void\n"
		":param fi: The federateInfo object to alter.\n"
		":type intProperty: int\n"
		":param intProperty: An int identifying the property.\n"
		":type propertyValue: int\n"
		":param propertyValue: The value to set the property to.\n"
		""},
	 { "helicsFederateRegisterInterfaces", _wrap_helicsFederateRegisterInterfaces, METH_VARARGS, "\n"
		"Load interfaces from a file.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to which to load interfaces.\n"
		":type file: string\n"
		":param file: The name of a file to load the interfaces from either JSON, or TOML.\n"
		""},
	 { "helicsFederateGlobalError", _wrap_helicsFederateGlobalError, METH_VARARGS, "\n"
		"Generate a global error from a federate.\n"
		"\n"
		"A global error halts the co-simulation completely.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to create an error in.\n"
		":type error_code: int\n"
		":param error_code: The integer code for the error.\n"
		":type error_string: string\n"
		":param error_string: A string describing the error.\n"
		""},
	 { "helicsFederateLocalError", _wrap_helicsFederateLocalError, METH_VARARGS, "\n"
		"Generate a local error in a federate.\n"
		"\n"
		"This will propagate through the co-simulation but not necessarily halt the co-simulation, it has a similar effect to finalize\n"
		"but does allow some interaction with a core for a brief time.\n"
		":type fed: void\n"
		":param fed: The federate to create an error in.\n"
		":type error_code: int\n"
		":param error_code: The integer code for the error.\n"
		":type error_string: string\n"
		":param error_string: A string describing the error.\n"
		""},
	 { "helicsFederateFinalize", _wrap_helicsFederateFinalize, METH_O, "Finalize the federate. This function halts all communication in the federate and disconnects it from the core."},
	 { "helicsFederateFinalizeAsync", _wrap_helicsFederateFinalizeAsync, METH_O, "Finalize the federate in an async call."},
	 { "helicsFederateFinalizeComplete", _wrap_helicsFederateFinalizeComplete, METH_O, "Complete the asynchronous finalize call."},
	 { "helicsFederateFree", _wrap_helicsFederateFree, METH_O, "Release the memory associated with a federate."},
	 { "helicsCloseLibrary", _wrap_helicsCloseLibrary, METH_NOARGS, "\n"
		"Call when done using the helics library.\n"
		"This function will ensure the threads are closed properly. If possible this should be the last call before exiting.\n"
		""},
	 { "helicsFederateEnterInitializingMode", _wrap_helicsFederateEnterInitializingMode, METH_O, "\n"
		"Enter the initialization state of a federate.\n"
		"\n"
		"The initialization state allows initial values to be set and received if the iteration is requested on entry to the execution\n"
		"state. This is a blocking call and will block until the core allows it to proceed.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to operate on.\n"
		""},
	 { "helicsFederateEnterInitializingModeAsync", _wrap_helicsFederateEnterInitializingModeAsync, METH_O, "\n"
		"Non blocking alternative to 'helicsFederateEnterInitializingMode'.\n"
		"\n"
		"The function helicsFederateEnterInitializationModeFinalize must be called to finish the operation.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to operate on.\n"
		""},
	 { "helicsFederateIsAsyncOperationCompleted", _wrap_helicsFederateIsAsyncOperationCompleted, METH_O, "\n"
		"Check if the current Asynchronous operation has completed.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to operate on.\n"
		"\n"
		"\n"
		":rtype: int\n"
		":return: helics_false if not completed, helics_true if completed.\n"
		""},
	 { "helicsFederateEnterInitializingModeComplete", _wrap_helicsFederateEnterInitializingModeComplete, METH_O, "\n"
		"Finalize the entry to initialize mode that was initiated with /ref heliceEnterInitializingModeAsync.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate desiring to complete the initialization step.\n"
		""},
	 { "helicsFederateEnterExecutingMode", _wrap_helicsFederateEnterExecutingMode, METH_O, "\n"
		"Request that the federate enter the Execution mode.\n"
		"\n"
		"This call is blocking until granted entry by the core object. On return from this call the federate will be at time 0.\n"
		"         For an asynchronous alternative call see /ref helicsFederateEnterExecutingModeAsync.\n"
		"\n"
		":type fed: void\n"
		":param fed: A federate to change modes.\n"
		""},
	 { "helicsFederateEnterExecutingModeAsync", _wrap_helicsFederateEnterExecutingModeAsync, METH_O, "\n"
		"Request that the federate enter the Execution mode.\n"
		"\n"
		"This call is non-blocking and will return immediately. Call /ref helicsFederateEnterExecutingModeComplete to finish the call\n"
		"sequence.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate object to complete the call.\n"
		""},
	 { "helicsFederateEnterExecutingModeComplete", _wrap_helicsFederateEnterExecutingModeComplete, METH_O, "\n"
		"Complete the call to /ref helicsFederateEnterExecutingModeAsync.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate object to complete the call.\n"
		""},
	 { "helicsFederateEnterExecutingModeIterative", _wrap_helicsFederateEnterExecutingModeIterative, METH_VARARGS, "\n"
		"Request an iterative time.\n"
		"\n"
		"This call allows for finer grain control of the iterative process than /ref helicsFederateRequestTime. It takes a time and\n"
		"         iteration request, and returns a time and iteration status.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to make the request of.\n"
		":type iterate: int\n"
		":param iterate: The requested iteration mode.\n"
		"\n"
		"\n"
		":rtype: int\n"
		":return: An iteration structure with field containing the time and iteration status.\n"
		""},
	 { "helicsFederateEnterExecutingModeIterativeAsync", _wrap_helicsFederateEnterExecutingModeIterativeAsync, METH_VARARGS, "\n"
		"Request an iterative entry to the execution mode.\n"
		"\n"
		"This call allows for finer grain control of the iterative process than /ref helicsFederateRequestTime. It takes a time and\n"
		"         iteration request, and returns a time and iteration status\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to make the request of.\n"
		":type iterate: int\n"
		":param iterate: The requested iteration mode.\n"
		""},
	 { "helicsFederateEnterExecutingModeIterativeComplete", _wrap_helicsFederateEnterExecutingModeIterativeComplete, METH_O, "\n"
		"Complete the asynchronous iterative call into ExecutionMode.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to make the request of.\n"
		"\n"
		"\n"
		":rtype: int\n"
		":return: An iteration object containing the iteration time and iteration_status.\n"
		""},
	 { "helicsFederateGetState", _wrap_helicsFederateGetState, METH_O, "\n"
		"Get the current state of a federate.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to query.\n"
		"\n"
		"\n"
		":rtype: int\n"
		":return: State the resulting state if void return helics_ok.\n"
		""},
	 { "helicsFederateGetCoreObject", _wrap_helicsFederateGetCoreObject, METH_O, "\n"
		"Get the core object associated with a federate.\n"
		"\n"
		":type fed: void\n"
		":param fed: A federate object.\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: A core object, nullptr if invalid.\n"
		""},
	 { "helicsFederateRequestTime", _wrap_helicsFederateRequestTime, METH_VARARGS, "\n"
		"Request the next time for federate execution.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to make the request of.\n"
		":type requestTime: float\n"
		":param requestTime: The next requested time.\n"
		"\n"
		"\n"
		":rtype: float\n"
		":return: The time granted to the federate, will return helics_time_maxtime if the simulation has terminated or is invalid.\n"
		""},
	 { "helicsFederateRequestTimeAdvance", _wrap_helicsFederateRequestTimeAdvance, METH_VARARGS, "\n"
		"Request the next time for federate execution.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to make the request of.\n"
		":type timeDelta: float\n"
		":param timeDelta: The requested amount of time to advance.\n"
		"\n"
		"\n"
		":rtype: float\n"
		":return: The time granted to the federate, will return helics_time_maxtime if the simulation has terminated or is invalid\n"
		""},
	 { "helicsFederateRequestNextStep", _wrap_helicsFederateRequestNextStep, METH_O, "\n"
		"Request the next time step for federate execution.\n"
		"\n"
		"Feds should have setup the period or minDelta for this to work well but it will request the next time step which is the current\n"
		"time plus the minimum time step.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to make the request of.\n"
		"\n"
		"\n"
		":rtype: float\n"
		":return: The time granted to the federate, will return helics_time_maxtime if the simulation has terminated or is invalid\n"
		""},
	 { "helicsFederateRequestTimeIterative", _wrap_helicsFederateRequestTimeIterative, METH_VARARGS, "\n"
		"Request an iterative time.\n"
		"\n"
		"This call allows for finer grain control of the iterative process than /ref helicsFederateRequestTime. It takes a time and and\n"
		"iteration request, and returns a time and iteration status.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to make the request of.\n"
		":type requestTime: float\n"
		":param requestTime: The next desired time.\n"
		":type iterate: int\n"
		":param iterate: The requested iteration mode.\n"
		"\n"
		"\n"
		"\n"
		":rtype: float\n"
		":return: The granted time, will return helics_time_maxtime if the simulation has terminated along with the appropriate iteration result.\n"
		"\n"
		"This function also returns the iteration specification of the result.\n"
		""},
	 { "helicsFederateRequestTimeAsync", _wrap_helicsFederateRequestTimeAsync, METH_VARARGS, "\n"
		"Request the next time for federate execution in an asynchronous call.\n"
		"\n"
		"Call /ref helicsFederateRequestTimeComplete to finish the call.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to make the request of.\n"
		":type requestTime: float\n"
		":param requestTime: The next requested time.\n"
		""},
	 { "helicsFederateRequestTimeComplete", _wrap_helicsFederateRequestTimeComplete, METH_O, "\n"
		"Complete an asynchronous requestTime call.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to make the request of.\n"
		"\n"
		"\n"
		":rtype: float\n"
		":return: The time granted to the federate, will return helics_time_maxtime if the simulation has terminated.\n"
		""},
	 { "helicsFederateRequestTimeIterativeAsync", _wrap_helicsFederateRequestTimeIterativeAsync, METH_VARARGS, "\n"
		"Request an iterative time through an asynchronous call.\n"
		"\n"
		"This call allows for finer grain control of the iterative process than /ref helicsFederateRequestTime. It takes a time and\n"
		"iteration request, and returns a time and iteration status. Call /ref helicsFederateRequestTimeIterativeComplete to finish the process.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to make the request of.\n"
		":type requestTime: float\n"
		":param requestTime: The next desired time.\n"
		":type iterate: int\n"
		":param iterate: The requested iteration mode.\n"
		""},
	 { "helicsFederateRequestTimeIterativeComplete", _wrap_helicsFederateRequestTimeIterativeComplete, METH_O, "\n"
		"Complete an iterative time request asynchronous call.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to make the request of.\n"
		"\n"
		"\n"
		":rtype: float\n"
		":return: The granted time, will return helics_time_maxtime if the simulation has terminated.\n"
		"\n"
		"This function also returns the iteration specification of the result.\n"
		""},
	 { "helicsFederateGetName", _wrap_helicsFederateGetName, METH_O, "\n"
		"Get the name of the federate.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate object to query.\n"
		"\n"
		":rtype: string\n"
		":return: A pointer to a string with the name.\n"
		""},
	 { "helicsFederateSetTimeProperty", _wrap_helicsFederateSetTimeProperty, METH_VARARGS, "\n"
		"Set a time based property for a federate.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate object to set the property for.\n"
		":type timeProperty: int\n"
		":param timeProperty: A integer code for a time property.\n"
		":type time: float\n"
		":param time: The requested value of the property.\n"
		""},
	 { "helicsFederateSetFlagOption", _wrap_helicsFederateSetFlagOption, METH_VARARGS, "\n"
		"Set a flag for the federate.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to alter a flag for.\n"
		":type flag: int\n"
		":param flag: The flag to change.\n"
		":type flagValue: int\n"
		":param flagValue: The new value of the flag. 0 for false, !=0 for true.\n"
		""},
	 { "helicsFederateSetSeparator", _wrap_helicsFederateSetSeparator, METH_VARARGS, "\n"
		"Set the separator character in a federate.\n"
		"\n"
		"The separator character is the separation character for local publications/endpoints in creating their global name.\n"
		"         For example if the separator character is '/' then a local endpoint would have a globally reachable name of fedName/localName.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate info object to alter.\n"
		":type separator: char\n"
		":param separator: The character to use as a separator.\n"
		""},
	 { "helicsFederateSetIntegerProperty", _wrap_helicsFederateSetIntegerProperty, METH_VARARGS, "\n"
		"Set an integer based property of a federate.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to change the property for.\n"
		":type intProperty: int\n"
		":param intProperty: The property to set.\n"
		":type propertyVal: int\n"
		":param propertyVal: The value of the property.\n"
		""},
	 { "helicsFederateGetTimeProperty", _wrap_helicsFederateGetTimeProperty, METH_VARARGS, "\n"
		"Get the current value of a time based property in a federate.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate query.\n"
		":type timeProperty: int\n"
		":param timeProperty: The property to query.\n"
		""},
	 { "helicsFederateGetFlagOption", _wrap_helicsFederateGetFlagOption, METH_VARARGS, "\n"
		"Get a flag value for a federate.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to get the flag for.\n"
		":type flag: int\n"
		":param flag: The flag to query.\n"
		"\n"
		"\n"
		":rtype: int\n"
		":return: The value of the flag.\n"
		""},
	 { "helicsFederateGetIntegerProperty", _wrap_helicsFederateGetIntegerProperty, METH_VARARGS, "\n"
		"Get the current value of an integer property (such as a logging level).\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to get the flag for.\n"
		":type intProperty: int\n"
		":param intProperty: A code for the property to set /ref helics_handle_options.\n"
		"\n"
		"\n"
		":rtype: int\n"
		":return: The value of the property.\n"
		""},
	 { "helicsFederateGetCurrentTime", _wrap_helicsFederateGetCurrentTime, METH_O, "\n"
		"Get the current time of the federate.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate object to query.\n"
		"\n"
		"\n"
		":rtype: float\n"
		":return: The current time of the federate.\n"
		""},
	 { "helicsFederateSetGlobal", _wrap_helicsFederateSetGlobal, METH_VARARGS, "\n"
		"Set a federation global value through a federate.\n"
		"\n"
		"This overwrites any previous value for this name.\n"
		":type fed: void\n"
		":param fed: The federate to set the global through.\n"
		":type valueName: string\n"
		":param valueName: The name of the global to set.\n"
		":type value: string\n"
		":param value: The value of the global.\n"
		""},
	 { "helicsFederateAddDependency", _wrap_helicsFederateAddDependency, METH_VARARGS, "\n"
		"Add a time dependency for a federate. The federate will depend on the given named federate for time synchronization.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to add the dependency for.\n"
		":type fedName: string\n"
		":param fedName: The name of the federate to depend on.\n"
		""},
	 { "helicsFederateSetLogFile", _wrap_helicsFederateSetLogFile, METH_VARARGS, "\n"
		"Set the logging file for a federate (actually on the core associated with a federate).\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to set the log file for.\n"
		":type logFile: string\n"
		":param logFile: The name of the log file.\n"
		""},
	 { "helicsFederateLogErrorMessage", _wrap_helicsFederateLogErrorMessage, METH_VARARGS, "\n"
		"Log an error message through a federate.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to log the error message through.\n"
		":type logmessage: string\n"
		":param logmessage: The message to put in the log.\n"
		""},
	 { "helicsFederateLogWarningMessage", _wrap_helicsFederateLogWarningMessage, METH_VARARGS, "\n"
		"Log a warning message through a federate.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to log the warning message through.\n"
		":type logmessage: string\n"
		":param logmessage: The message to put in the log.\n"
		""},
	 { "helicsFederateLogInfoMessage", _wrap_helicsFederateLogInfoMessage, METH_VARARGS, "\n"
		"Log an info message through a federate.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to log the info message through.\n"
		":type logmessage: string\n"
		":param logmessage: The message to put in the log.\n"
		""},
	 { "helicsFederateLogDebugMessage", _wrap_helicsFederateLogDebugMessage, METH_VARARGS, "\n"
		"Log a debug message through a federate.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to log the debug message through.\n"
		":type logmessage: string\n"
		":param logmessage: The message to put in the log.\n"
		""},
	 { "helicsFederateLogLevelMessage", _wrap_helicsFederateLogLevelMessage, METH_VARARGS, "\n"
		"Log a message through a federate.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to log the message through.\n"
		":type loglevel: int\n"
		":param loglevel: The level of the message to log see /ref helics_log_levels.\n"
		":type logmessage: string\n"
		":param logmessage: The message to put in the log.\n"
		""},
	 { "helicsCoreSetGlobal", _wrap_helicsCoreSetGlobal, METH_VARARGS, "\n"
		"Set a global value in a core.\n"
		"\n"
		"This overwrites any previous value for this name.\n"
		"\n"
		":type core: void\n"
		":param core: The core to set the global through.\n"
		":type valueName: string\n"
		":param valueName: The name of the global to set.\n"
		":type value: string\n"
		":param value: The value of the global.\n"
		""},
	 { "helicsBrokerSetGlobal", _wrap_helicsBrokerSetGlobal, METH_VARARGS, "\n"
		"Set a federation global value.\n"
		"\n"
		"This overwrites any previous value for this name.\n"
		"\n"
		":type broker: void\n"
		":param broker: The broker to set the global through.\n"
		":type valueName: string\n"
		":param valueName: The name of the global to set.\n"
		":type value: string\n"
		":param value: The value of the global.\n"
		""},
	 { "helicsCoreSetLogFile", _wrap_helicsCoreSetLogFile, METH_VARARGS, "\n"
		"Set the log file on a core.\n"
		"\n"
		":type core: void\n"
		":param core: The core to set the log file for.\n"
		":type logFileName: string\n"
		":param logFileName: The name of the file to log to.\n"
		""},
	 { "helicsBrokerSetLogFile", _wrap_helicsBrokerSetLogFile, METH_VARARGS, "\n"
		"Set the log file on a broker.\n"
		"\n"
		":type broker: void\n"
		":param broker: The broker to set the log file for.\n"
		":type logFileName: string\n"
		":param logFileName: The name of the file to log to.\n"
		""},
	 { "helicsCreateQuery", _wrap_helicsCreateQuery, METH_VARARGS, "\n"
		"Create a query object.\n"
		"\n"
		"A query object consists of a target and query string.\n"
		"\n"
		":type target: string\n"
		":param target: The name of the target to query.\n"
		":type query: string\n"
		":param query: The query to make of the target.\n"
		""},
	 { "helicsQueryExecute", _wrap_helicsQueryExecute, METH_VARARGS, "\n"
		"Execute a query.\n"
		"\n"
		"The call will block until the query finishes which may require communication or other delays.\n"
		"\n"
		":type query: void\n"
		":param query: The query object to use in the query.\n"
		":type fed: void\n"
		":param fed: A federate to send the query through.\n"
		"\n"
		"\n"
		":rtype: string\n"
		":return: A pointer to a string.  The string will remain valid until the query is freed or executed again.\n"
		""},
	 { "helicsQueryCoreExecute", _wrap_helicsQueryCoreExecute, METH_VARARGS, "\n"
		"Execute a query directly on a core.\n"
		"\n"
		"The call will block until the query finishes which may require communication or other delays.\n"
		"\n"
		":type query: void\n"
		":param query: The query object to use in the query.\n"
		":type core: void\n"
		":param core: The core to send the query to.\n"
		"\n"
		"\n"
		":rtype: string\n"
		":return: A pointer to a string.  The string will remain valid until the query is freed or executed again.\n"
		""},
	 { "helicsQueryBrokerExecute", _wrap_helicsQueryBrokerExecute, METH_VARARGS, "\n"
		"Execute a query directly on a broker.\n"
		"\n"
		"The call will block until the query finishes which may require communication or other delays.\n"
		"\n"
		":type query: void\n"
		":param query: The query object to use in the query.\n"
		":type broker: void\n"
		":param broker: The broker to send the query to.\n"
		"\n"
		"\n"
		":rtype: string\n"
		":return: A pointer to a string.  The string will remain valid until the query is freed or executed again.\n"
		""},
	 { "helicsQueryExecuteAsync", _wrap_helicsQueryExecuteAsync, METH_VARARGS, "\n"
		"Execute a query in a non-blocking call.\n"
		"\n"
		":type query: void\n"
		":param query: The query object to use in the query.\n"
		":type fed: void\n"
		":param fed: A federate to send the query through.\n"
		""},
	 { "helicsQueryExecuteComplete", _wrap_helicsQueryExecuteComplete, METH_O, "\n"
		"Complete the return from a query called with /ref helicsExecuteQueryAsync.\n"
		"\n"
		"The function will block until the query completes /ref isQueryComplete can be called to determine if a query has completed or\n"
		"not.\n"
		"\n"
		":type query: void\n"
		":param query: The query object to complete execution of.\n"
		"\n"
		"\n"
		":rtype: string\n"
		":return: A pointer to a string. The string will remain valid until the query is freed or executed again.\n"
		""},
	 { "helicsQueryIsCompleted", _wrap_helicsQueryIsCompleted, METH_O, "\n"
		"Check if an asynchronously executed query has completed.\n"
		"\n"
		"This function should usually be called after a QueryExecuteAsync function has been called.\n"
		"\n"
		":type query: void\n"
		":param query: The query object to check if completed.\n"
		"\n"
		":rtype: int\n"
		":return: Will return helics_true if an asynchronous query has completed or a regular query call was made with a result,\n"
		"            and false if an asynchronous query has not completed or is invalid\n"
		""},
	 { "helicsQuerySetTarget", _wrap_helicsQuerySetTarget, METH_VARARGS, "\n"
		"Update the target of a query.\n"
		"\n"
		":type query: void\n"
		":param query: The query object to change the target of.\n"
		":type target: string\n"
		":param target: the name of the target to query\n"
		"\n"
		""},
	 { "helicsQuerySetQueryString", _wrap_helicsQuerySetQueryString, METH_VARARGS, "\n"
		"Update the queryString of a query.\n"
		"\n"
		":type query: void\n"
		":param query: The query object to change the target of.\n"
		":type queryString: string\n"
		":param queryString: the new queryString\n"
		""},
	 { "helicsQueryFree", _wrap_helicsQueryFree, METH_O, "Free the memory associated with a query object."},
	 { "helicsCleanupLibrary", _wrap_helicsCleanupLibrary, METH_NOARGS, "\n"
		"Function to do some housekeeping work.\n"
		"\n"
		"This runs some cleanup routines and tries to close out any residual thread that haven't been shutdown yet.\n"
		""},
	 { "helicsFederateRegisterSubscription", _wrap_helicsFederateRegisterSubscription, METH_VARARGS, "\n"
		"sub/pub registration\n"
		"\n"
		"Create a subscription.\n"
		"\n"
		"The subscription becomes part of the federate and is destroyed when the federate is freed so there are no separate free\n"
		"functions for subscriptions and publications.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate object in which to create a subscription, must have been created with /ref helicsCreateValueFederate or\n"
		"    /ref helicsCreateCombinationFederate.\n"
		":type key: string\n"
		":param key: The identifier matching a publication to get a subscription for.\n"
		":type units: string\n"
		":param units: A string listing the units of the subscription (may be NULL).\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: An object containing the subscription.\n"
		""},
	 { "helicsFederateRegisterPublication", _wrap_helicsFederateRegisterPublication, METH_VARARGS, "\n"
		"Register a publication with a known type.\n"
		"\n"
		"The publication becomes part of the federate and is destroyed when the federate is freed so there are no separate free\n"
		"functions for subscriptions and publications.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate object in which to create a publication.\n"
		":type key: string\n"
		":param key: The identifier for the publication the global publication key will be prepended with the federate name.\n"
		":type type: int\n"
		":param type: A code identifying the type of the input see /ref helics_data_type for available options.\n"
		":type units: string\n"
		":param units: A string listing the units of the subscription (may be NULL).\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: An object containing the publication.\n"
		""},
	 { "helicsFederateRegisterTypePublication", _wrap_helicsFederateRegisterTypePublication, METH_VARARGS, "\n"
		"Register a publication with a defined type.\n"
		"\n"
		"The publication becomes part of the federate and is destroyed when the federate is freed so there are no separate free\n"
		"functions for subscriptions and publications.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate object in which to create a publication.\n"
		":type key: string\n"
		":param key: The identifier for the publication.\n"
		":type type: string\n"
		":param type: A string labeling the type of the publication.\n"
		":type units: string\n"
		":param units: A string listing the units of the subscription (may be NULL).\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: An object containing the publication.\n"
		""},
	 { "helicsFederateRegisterGlobalPublication", _wrap_helicsFederateRegisterGlobalPublication, METH_VARARGS, "\n"
		"Register a global named publication with an arbitrary type.\n"
		"\n"
		"The publication becomes part of the federate and is destroyed when the federate is freed so there are no separate free\n"
		"functions for subscriptions and publications.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate object in which to create a publication.\n"
		":type key: string\n"
		":param key: The identifier for the publication.\n"
		":type type: int\n"
		":param type: A code identifying the type of the input see /ref helics_data_type for available options.\n"
		":type units: string\n"
		":param units: A string listing the units of the subscription (may be NULL).\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: An object containing the publication.\n"
		""},
	 { "helicsFederateRegisterGlobalTypePublication", _wrap_helicsFederateRegisterGlobalTypePublication, METH_VARARGS, "\n"
		"Register a global publication with a defined type.\n"
		"\n"
		"The publication becomes part of the federate and is destroyed when the federate is freed so there are no separate free\n"
		"functions for subscriptions and publications.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate object in which to create a publication.\n"
		":type key: string\n"
		":param key: The identifier for the publication.\n"
		":type type: string\n"
		":param type: A string describing the expected type of the publication.\n"
		":type units: string\n"
		":param units: A string listing the units of the subscription (may be NULL).\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: An object containing the publication.\n"
		""},
	 { "helicsFederateRegisterInput", _wrap_helicsFederateRegisterInput, METH_VARARGS, "\n"
		"Register a named input.\n"
		"\n"
		"The input becomes part of the federate and is destroyed when the federate is freed so there are no separate free\n"
		"functions for subscriptions, inputs, and publications.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate object in which to create an input.\n"
		":type key: string\n"
		":param key: The identifier for the publication the global input key will be prepended with the federate name.\n"
		":type type: int\n"
		":param type: A code identifying the type of the input see /ref helics_data_type for available options.\n"
		":type units: string\n"
		":param units: A string listing the units of the input (may be NULL).\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: An object containing the input.\n"
		""},
	 { "helicsFederateRegisterTypeInput", _wrap_helicsFederateRegisterTypeInput, METH_VARARGS, "\n"
		"Register an input with a defined type.\n"
		"\n"
		"The input becomes part of the federate and is destroyed when the federate is freed so there are no separate free\n"
		"functions for subscriptions, inputs, and publications.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate object in which to create an input.\n"
		":type key: string\n"
		":param key: The identifier for the input.\n"
		":type type: string\n"
		":param type: A string describing the expected type of the input.\n"
		":type units: string\n"
		":param units: A string listing the units of the input maybe NULL.\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: An object containing the publication.\n"
		""},
	 { "helicsFederateRegisterGlobalInput", _wrap_helicsFederateRegisterGlobalInput, METH_VARARGS, "\n"
		"Register a global named input.\n"
		"\n"
		"The publication becomes part of the federate and is destroyed when the federate is freed so there are no separate free\n"
		"functions for subscriptions and publications.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate object in which to create a publication.\n"
		":type key: string\n"
		":param key: The identifier for the publication.\n"
		":type type: int\n"
		":param type: A code identifying the type of the input see /ref helics_data_type for available options.\n"
		":type units: string\n"
		":param units: A string listing the units of the subscription maybe NULL.\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: An object containing the publication.\n"
		""},
	 { "helicsFederateRegisterGlobalTypeInput", _wrap_helicsFederateRegisterGlobalTypeInput, METH_VARARGS, "\n"
		"Register a global publication with an arbitrary type.\n"
		"\n"
		"The publication becomes part of the federate and is destroyed when the federate is freed so there are no separate free\n"
		"functions for subscriptions and publications.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate object in which to create a publication.\n"
		":type key: string\n"
		":param key: The identifier for the publication.\n"
		":type type: string\n"
		":param type: A string defining the type of the input.\n"
		":type units: string\n"
		":param units: A string listing the units of the subscription maybe NULL.\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: An object containing the publication.\n"
		""},
	 { "helicsFederateGetPublication", _wrap_helicsFederateGetPublication, METH_VARARGS, "\n"
		"Get a publication object from a key.\n"
		"\n"
		":type fed: void\n"
		":param fed: The value federate object to use to get the publication.\n"
		":type key: string\n"
		":param key: The name of the publication.\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: A helics_publication object, the object will not be valid and err will contain an error code if no publication with the\n"
		"    specified key exists.\n"
		""},
	 { "helicsFederateGetPublicationByIndex", _wrap_helicsFederateGetPublicationByIndex, METH_VARARGS, "\n"
		"Get a publication by its index, typically already created via registerInterfaces file or something of that nature.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate object in which to create a publication.\n"
		":type index: int\n"
		":param index: The index of the publication to get.\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: A helics_publication.\n"
		""},
	 { "helicsFederateGetInput", _wrap_helicsFederateGetInput, METH_VARARGS, "\n"
		"Get an input object from a key.\n"
		"\n"
		":type fed: void\n"
		":param fed: The value federate object to use to get the publication.\n"
		":type key: string\n"
		":param key: The name of the input.\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: A helics_input object, the object will not be valid and err will contain an error code if no input with the specified\n"
		"    key exists.\n"
		""},
	 { "helicsFederateGetInputByIndex", _wrap_helicsFederateGetInputByIndex, METH_VARARGS, "\n"
		"Get an input by its index, typically already created via registerInterfaces file or something of that nature.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate object in which to create a publication.\n"
		":type index: int\n"
		":param index: The index of the publication to get.\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: A helics_input, which will be NULL if an invalid index.\n"
		""},
	 { "helicsFederateGetSubscription", _wrap_helicsFederateGetSubscription, METH_VARARGS, "\n"
		"Get an input object from a subscription target.\n"
		"\n"
		":type fed: void\n"
		":param fed: The value federate object to use to get the publication.\n"
		":type key: string\n"
		":param key: The name of the publication that a subscription is targeting.\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: A helics_input object, the object will not be valid and err will contain an error code if no input with the specified\n"
		"    key exists.\n"
		""},
	 { "helicsFederateClearUpdates", _wrap_helicsFederateClearUpdates, METH_O, "\n"
		"Clear all the update flags from a federates inputs.\n"
		"\n"
		":type fed: void\n"
		":param fed: The value federate object for which to clear update flags.\n"
		""},
	 { "helicsFederateRegisterFromPublicationJSON", _wrap_helicsFederateRegisterFromPublicationJSON, METH_VARARGS, "\n"
		"Register the publications via JSON publication string.\n"
		"\n"
		":type fed: void\n"
		":param fed: The value federate object to use to register the publications.\n"
		":type json: string\n"
		":param json: The JSON publication string.\n"
		"\n"
		"\n"
		"This would be the same JSON that would be used to publish data.\n"
		""},
	 { "helicsFederatePublishJSON", _wrap_helicsFederatePublishJSON, METH_VARARGS, "\n"
		"Publish data contained in a JSON file or string.\n"
		"\n"
		":type fed: void\n"
		":param fed: The value federate object through which to publish the data.\n"
		":type json: string\n"
		":param json: The publication file name or literal JSON data string.\n"
		""},
	 { "helicsPublicationIsValid", _wrap_helicsPublicationIsValid, METH_O, "\n"
		"Check if a publication is valid.\n"
		"\n"
		":type pub: void\n"
		":param pub: The publication to check.\n"
		"\n"
		":rtype: int\n"
		":return: helics_true if the publication is a valid publication.\n"
		""},
	 { "helicsPublicationPublishRaw", _wrap_helicsPublicationPublishRaw, METH_VARARGS, "\n"
		"Publish raw data from a char * and length.\n"
		"\n"
		":type pub: void\n"
		":param pub: The publication to publish for.\n"
		":type data: void\n"
		":param data: A pointer to the raw data.\n"
		":type inputDataLength: int\n"
		":param inputDataLength: The size in bytes of the data to publish.\n"
		""},
	 { "helicsPublicationPublishString", _wrap_helicsPublicationPublishString, METH_VARARGS, "\n"
		"Publish a string.\n"
		"\n"
		":type pub: void\n"
		":param pub: The publication to publish for.\n"
		":type str: string\n"
		":param str: The string to publish.\n"
		""},
	 { "helicsPublicationPublishInteger", _wrap_helicsPublicationPublishInteger, METH_VARARGS, "\n"
		"Publish an integer value.\n"
		"\n"
		":type pub: void\n"
		":param pub: The publication to publish for.\n"
		":type val: int\n"
		":param val: The numerical value to publish.\n"
		""},
	 { "helicsPublicationPublishBoolean", _wrap_helicsPublicationPublishBoolean, METH_VARARGS, "\n"
		"Publish a Boolean Value.\n"
		"\n"
		":type pub: void\n"
		":param pub: The publication to publish for.\n"
		":type val: int\n"
		":param val: The boolean value to publish.\n"
		""},
	 { "helicsPublicationPublishDouble", _wrap_helicsPublicationPublishDouble, METH_VARARGS, "\n"
		"Publish a double floating point value.\n"
		"\n"
		":type pub: void\n"
		":param pub: The publication to publish for.\n"
		":type val: float\n"
		":param val: The numerical value to publish.\n"
		""},
	 { "helicsPublicationPublishTime", _wrap_helicsPublicationPublishTime, METH_VARARGS, "\n"
		"Publish a time value.\n"
		"\n"
		":type pub: void\n"
		":param pub: The publication to publish for.\n"
		":type val: float\n"
		":param val: The numerical value to publish.\n"
		""},
	 { "helicsPublicationPublishChar", _wrap_helicsPublicationPublishChar, METH_VARARGS, "\n"
		"Publish a single character.\n"
		"\n"
		":type pub: void\n"
		":param pub: The publication to publish for.\n"
		":type val: char\n"
		":param val: The numerical value to publish.\n"
		""},
	 { "helicsPublicationPublishComplex", _wrap_helicsPublicationPublishComplex, METH_VARARGS, "\n"
		"Publish a complex value (or pair of values).\n"
		"\n"
		":type pub: void\n"
		":param pub: The publication to publish for.\n"
		":type real: float\n"
		":param real: The real part of a complex number to publish.\n"
		":type imag: float\n"
		":param imag: The imaginary part of a complex number to publish.\n"
		""},
	 { "helicsPublicationPublishVector", _wrap_helicsPublicationPublishVector, METH_VARARGS, "\n"
		"Publish a vector of doubles.\n"
		"\n"
		":type pub: void\n"
		":param pub: The publication to publish for.\n"
		":type vectorInput: float\n"
		":param vectorInput: A pointer to an array of double data.\n"
		""},
	 { "helicsPublicationPublishNamedPoint", _wrap_helicsPublicationPublishNamedPoint, METH_VARARGS, "\n"
		"Publish a named point.\n"
		"\n"
		":type pub: void\n"
		":param pub: The publication to publish for.\n"
		":type str: string\n"
		":param str: A string for the name to publish.\n"
		":type val: float\n"
		":param val: A double for the value to publish.\n"
		""},
	 { "helicsPublicationAddTarget", _wrap_helicsPublicationAddTarget, METH_VARARGS, "\n"
		"Add a named input to the list of targets a publication publishes to.\n"
		"\n"
		":type pub: void\n"
		":param pub: The publication to add the target for.\n"
		":type target: string\n"
		":param target: The name of an input that the data should be sent to.\n"
		""},
	 { "helicsInputIsValid", _wrap_helicsInputIsValid, METH_O, "\n"
		"Check if an input is valid.\n"
		"\n"
		":type ipt: void\n"
		":param ipt: The input to check.\n"
		"\n"
		":rtype: int\n"
		":return: helics_true if the Input object represents a valid input.\n"
		""},
	 { "helicsInputAddTarget", _wrap_helicsInputAddTarget, METH_VARARGS, "\n"
		"Add a publication to the list of data that an input subscribes to.\n"
		"\n"
		":type ipt: void\n"
		":param ipt: The named input to modify.\n"
		":type target: string\n"
		":param target: The name of a publication that an input should subscribe to.\n"
		""},
	 { "helicsInputGetRawValueSize", _wrap_helicsInputGetRawValueSize, METH_O, "\n"
		"Get the size of the raw value for subscription.\n"
		"\n"
		":rtype: int\n"
		":return: The size of the raw data/string in bytes.\n"
		""},
	 { "helicsInputGetRawValue", _wrap_helicsInputGetRawValue, METH_O, "\n"
		"Get the raw data for the latest value of a subscription.\n"
		"\n"
		":type ipt: void\n"
		":param ipt: The input to get the data for.\n"
		"\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: Raw string data.\n"
		""},
	 { "helicsInputGetStringSize", _wrap_helicsInputGetStringSize, METH_O, "\n"
		"Get the size of a value for subscription assuming return as a string.\n"
		"\n"
		":rtype: int\n"
		":return: The size of the string.\n"
		""},
	 { "helicsInputGetString", _wrap_helicsInputGetString, METH_O, "\n"
		"Get a string value from a subscription.\n"
		"\n"
		":type ipt: void\n"
		":param ipt: The input to get the data for.\n"
		"\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: A string data\n"
		""},
	 { "helicsInputGetInteger", _wrap_helicsInputGetInteger, METH_O, "\n"
		"Get an integer value from a subscription.\n"
		"\n"
		":type ipt: void\n"
		":param ipt: The input to get the data for.\n"
		"\n"
		"\n"
		":rtype: int\n"
		":return: An int64_t value with the current value of the input.\n"
		""},
	 { "helicsInputGetBoolean", _wrap_helicsInputGetBoolean, METH_O, "\n"
		"Get a boolean value from a subscription.\n"
		"\n"
		":type ipt: void\n"
		":param ipt: The input to get the data for.\n"
		"\n"
		"\n"
		":rtype: int\n"
		":return: A boolean value of current input value.\n"
		""},
	 { "helicsInputGetDouble", _wrap_helicsInputGetDouble, METH_O, "\n"
		"Get a double value from a subscription.\n"
		"\n"
		":type ipt: void\n"
		":param ipt: The input to get the data for.\n"
		"\n"
		"\n"
		":rtype: float\n"
		":return: The double value of the input.\n"
		""},
	 { "helicsInputGetTime", _wrap_helicsInputGetTime, METH_O, "\n"
		"Get a time value from a subscription.\n"
		"\n"
		":type ipt: void\n"
		":param ipt: The input to get the data for.\n"
		"\n"
		"\n"
		":rtype: float\n"
		":return: The resulting time value.\n"
		""},
	 { "helicsInputGetChar", _wrap_helicsInputGetChar, METH_O, "\n"
		"Get a single character value from an input.\n"
		"\n"
		":type ipt: void\n"
		":param ipt: The input to get the data for.\n"
		"\n"
		"\n"
		":rtype: char\n"
		":return: The resulting character value.\n"
		""},
	 { "helicsInputGetComplexObject", _wrap_helicsInputGetComplexObject, METH_O, "\n"
		"Get a complex object from an input object.\n"
		"\n"
		":type ipt: void\n"
		":param ipt: The input to get the data for.\n"
		"\n"
		"\n"
		":rtype: :py:class:`helics_complex`\n"
		":return: A helics_complex structure with the value.\n"
		""},
	 { "helicsInputGetComplex", _wrap_helicsInputGetComplex, METH_O, "\n"
		"Get a pair of double forming a complex number from a subscriptions.\n"
		"\n"
		":type ipt: void\n"
		":param ipt: The input to get the data for.\n"
		"\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: a pair of floating point values that represent the real and imag values\n"
		""},
	 { "helicsInputGetVectorSize", _wrap_helicsInputGetVectorSize, METH_O, "\n"
		"Get the size of a value for subscription assuming return as an array of doubles.\n"
		"\n"
		":rtype: int\n"
		":return: The number of doubles in a returned vector.\n"
		""},
	 { "helicsInputGetVector", _wrap_helicsInputGetVector, METH_O, "\n"
		"Get a vector from a subscription.\n"
		"\n"
		":type ipt: void\n"
		":param ipt: The input to get the result for.\n"
		"\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: a list of floating point values\n"
		""},
	 { "helicsInputGetNamedPoint", _wrap_helicsInputGetNamedPoint, METH_O, "\n"
		"Get a named point from a subscription.\n"
		"\n"
		":type ipt: void\n"
		":param ipt: The input to get the result for.\n"
		"\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: a string and a double value for the named point\n"
		""},
	 { "helicsInputSetDefaultRaw", _wrap_helicsInputSetDefaultRaw, METH_VARARGS, "\n"
		"Set the default as a raw data array.\n"
		"\n"
		":type ipt: void\n"
		":param ipt: The input to set the default for.\n"
		":type data: void\n"
		":param data: A pointer to the raw data to use for the default.\n"
		""},
	 { "helicsInputSetDefaultString", _wrap_helicsInputSetDefaultString, METH_VARARGS, "\n"
		"Set the default as a string.\n"
		"\n"
		":type ipt: void\n"
		":param ipt: The input to set the default for.\n"
		":type str: string\n"
		":param str: A pointer to the default string.\n"
		""},
	 { "helicsInputSetDefaultInteger", _wrap_helicsInputSetDefaultInteger, METH_VARARGS, "\n"
		"Set the default as an integer.\n"
		"\n"
		":type ipt: void\n"
		":param ipt: The input to set the default for.\n"
		":type val: int\n"
		":param val: The default integer.\n"
		""},
	 { "helicsInputSetDefaultBoolean", _wrap_helicsInputSetDefaultBoolean, METH_VARARGS, "\n"
		"Set the default as a boolean.\n"
		"\n"
		":type ipt: void\n"
		":param ipt: The input to set the default for.\n"
		":type val: int\n"
		":param val: The default boolean value.\n"
		""},
	 { "helicsInputSetDefaultTime", _wrap_helicsInputSetDefaultTime, METH_VARARGS, "\n"
		"Set the default as a time.\n"
		"\n"
		":type ipt: void\n"
		":param ipt: The input to set the default for.\n"
		":type val: float\n"
		":param val: The default time value.\n"
		""},
	 { "helicsInputSetDefaultChar", _wrap_helicsInputSetDefaultChar, METH_VARARGS, "\n"
		"Set the default as a char.\n"
		"\n"
		":type ipt: void\n"
		":param ipt: The input to set the default for.\n"
		":type val: char\n"
		":param val: The default char value.\n"
		""},
	 { "helicsInputSetDefaultDouble", _wrap_helicsInputSetDefaultDouble, METH_VARARGS, "\n"
		"Set the default as a double.\n"
		"\n"
		":type ipt: void\n"
		":param ipt: The input to set the default for.\n"
		":type val: float\n"
		":param val: The default double value.\n"
		""},
	 { "helicsInputSetDefaultComplex", _wrap_helicsInputSetDefaultComplex, METH_VARARGS, "\n"
		"Set the default as a complex number.\n"
		"\n"
		":type ipt: void\n"
		":param ipt: The input to set the default for.\n"
		":type real: float\n"
		":param real: The default real value.\n"
		":type imag: float\n"
		":param imag: The default imaginary value.\n"
		""},
	 { "helicsInputSetDefaultVector", _wrap_helicsInputSetDefaultVector, METH_VARARGS, "\n"
		"Set the default as a vector of doubles.\n"
		"\n"
		":type ipt: void\n"
		":param ipt: The input to set the default for.\n"
		":type vectorInput: float\n"
		":param vectorInput: A pointer to an array of double data.\n"
		":type vectorLength: int\n"
		":param vectorLength: The number of points to publish.\n"
		""},
	 { "helicsInputSetDefaultNamedPoint", _wrap_helicsInputSetDefaultNamedPoint, METH_VARARGS, "\n"
		"Set the default as a NamedPoint.\n"
		"\n"
		":type ipt: void\n"
		":param ipt: The input to set the default for.\n"
		":type str: string\n"
		":param str: A pointer to a string representing the name.\n"
		":type val: float\n"
		":param val: A double value for the value of the named point.\n"
		""},
	 { "helicsInputGetType", _wrap_helicsInputGetType, METH_O, "\n"
		"Get the type of an input.\n"
		"\n"
		":type ipt: void\n"
		":param ipt: The input to query.\n"
		"\n"
		":rtype: string\n"
		":return: A void enumeration, helics_ok if everything worked.\n"
		""},
	 { "helicsInputGetPublicationType", _wrap_helicsInputGetPublicationType, METH_O, "\n"
		"Get the type the publisher to an input is sending.\n"
		"\n"
		":type ipt: void\n"
		":param ipt: The input to query.\n"
		"\n"
		":rtype: string\n"
		":return: A const char * with the type name.\n"
		""},
	 { "helicsPublicationGetType", _wrap_helicsPublicationGetType, METH_O, "\n"
		"Get the type of a publication.\n"
		"\n"
		":type pub: void\n"
		":param pub: The publication to query.\n"
		"\n"
		":rtype: string\n"
		":return: A void enumeration, helics_ok if everything worked.\n"
		""},
	 { "helicsInputGetKey", _wrap_helicsInputGetKey, METH_O, "\n"
		"Get the key of an input.\n"
		"\n"
		":type ipt: void\n"
		":param ipt: The input to query.\n"
		"\n"
		":rtype: string\n"
		":return: A void enumeration, helics_ok if everything worked.\n"
		""},
	 { "helicsSubscriptionGetKey", _wrap_helicsSubscriptionGetKey, METH_O, "\n"
		"Get the key of a subscription.\n"
		"\n"
		":rtype: string\n"
		":return: A const char with the subscription key.\n"
		""},
	 { "helicsPublicationGetKey", _wrap_helicsPublicationGetKey, METH_O, "\n"
		"Get the key of a publication.\n"
		"\n"
		"This will be the global key used to identify the publication to the federation.\n"
		"\n"
		":type pub: void\n"
		":param pub: The publication to query.\n"
		"\n"
		":rtype: string\n"
		":return: A void enumeration, helics_ok if everything worked.\n"
		""},
	 { "helicsInputGetUnits", _wrap_helicsInputGetUnits, METH_O, "\n"
		"Get the units of an input.\n"
		"\n"
		":type ipt: void\n"
		":param ipt: The input to query.\n"
		"\n"
		":rtype: string\n"
		":return: A void enumeration, helics_ok if everything worked.\n"
		""},
	 { "helicsInputGetInjectionUnits", _wrap_helicsInputGetInjectionUnits, METH_O, "\n"
		"Get the units of the publication that an input is linked to.\n"
		"\n"
		":type ipt: void\n"
		":param ipt: The input to query.\n"
		"\n"
		":rtype: string\n"
		":return: A void enumeration, helics_ok if everything worked.\n"
		""},
	 { "helicsInputGetExtractionUnits", _wrap_helicsInputGetExtractionUnits, METH_O, "\n"
		"Get the units of an input.\n"
		"\n"
		"The same as helicsInputGetUnits.\n"
		"\n"
		":type ipt: void\n"
		":param ipt: The input to query.\n"
		"\n"
		":rtype: string\n"
		":return: A void enumeration, helics_ok if everything worked.\n"
		""},
	 { "helicsPublicationGetUnits", _wrap_helicsPublicationGetUnits, METH_O, "\n"
		"Get the units of a publication.\n"
		"\n"
		":type pub: void\n"
		":param pub: The publication to query.\n"
		"\n"
		":rtype: string\n"
		":return: A void enumeration, helics_ok if everything worked.\n"
		""},
	 { "helicsInputGetInfo", _wrap_helicsInputGetInfo, METH_O, "\n"
		"Get the data in the info field of an input.\n"
		"\n"
		":type inp: void\n"
		":param inp: The input to query.\n"
		"\n"
		":rtype: string\n"
		":return: A string with the info field string.\n"
		""},
	 { "helicsInputSetInfo", _wrap_helicsInputSetInfo, METH_VARARGS, "\n"
		"Set the data in the info field for an input.\n"
		"\n"
		":type inp: void\n"
		":param inp: The input to query.\n"
		":type info: string\n"
		":param info: The string to set.\n"
		""},
	 { "helicsPublicationGetInfo", _wrap_helicsPublicationGetInfo, METH_O, "\n"
		"Get the data in the info field of an publication.\n"
		"\n"
		":type pub: void\n"
		":param pub: The publication to query.\n"
		"\n"
		":rtype: string\n"
		":return: A string with the info field string.\n"
		""},
	 { "helicsPublicationSetInfo", _wrap_helicsPublicationSetInfo, METH_VARARGS, "\n"
		"Set the data in the info field for a publication.\n"
		"\n"
		":type pub: void\n"
		":param pub: The publication to set the info field for.\n"
		":type info: string\n"
		":param info: The string to set.\n"
		""},
	 { "helicsInputGetOption", _wrap_helicsInputGetOption, METH_VARARGS, "\n"
		"Get the current value of an input handle option\n"
		"\n"
		":type inp: void\n"
		":param inp: The input to query.\n"
		":type option: int\n"
		":param option: Integer representation of the option in question see /ref helics_handle_options.\n"
		"\n"
		":rtype: int\n"
		":return: An integer value with the current value of the given option.\n"
		""},
	 { "helicsInputSetOption", _wrap_helicsInputSetOption, METH_VARARGS, "\n"
		"Set an option on an input\n"
		"\n"
		":type inp: void\n"
		":param inp: The input to query.\n"
		":type option: int\n"
		":param option: The option to set for the input /ref helics_handle_options.\n"
		":type value: int\n"
		":param value: The value to set the option to.\n"
		""},
	 { "helicsPublicationGetOption", _wrap_helicsPublicationGetOption, METH_VARARGS, "\n"
		"Get the value of an option for a publication\n"
		"\n"
		":type pub: void\n"
		":param pub: The publication to query.\n"
		":type option: int\n"
		":param option: The value to query see /ref helics_handle_options.\n"
		"\n"
		":rtype: int\n"
		":return: A string with the info field string.\n"
		""},
	 { "helicsPublicationSetOption", _wrap_helicsPublicationSetOption, METH_VARARGS, "\n"
		"Set the value of an option for a publication\n"
		"\n"
		":type pub: void\n"
		":param pub: The publication to query.\n"
		":type option: int\n"
		":param option: Integer code for the option to set /ref helics_handle_options.\n"
		":type val: int\n"
		":param val: The value to set the option to.\n"
		""},
	 { "helicsPublicationSetMinimumChange", _wrap_helicsPublicationSetMinimumChange, METH_VARARGS, "\n"
		"Set the minimum change detection tolerance.\n"
		"\n"
		":type pub: void\n"
		":param pub: The publication to modify.\n"
		":type tolerance: float\n"
		":param tolerance: The tolerance level for publication, values changing less than this value will not be published.\n"
		""},
	 { "helicsInputSetMinimumChange", _wrap_helicsInputSetMinimumChange, METH_VARARGS, "\n"
		"Set the minimum change detection tolerance.\n"
		"\n"
		":type inp: void\n"
		":param inp: The input to modify.\n"
		":type tolerance: float\n"
		":param tolerance: The tolerance level for registering an update, values changing less than this value will not show as being updated.\n"
		""},
	 { "helicsInputIsUpdated", _wrap_helicsInputIsUpdated, METH_O, "\n"
		"Check if a particular subscription was updated.\n"
		"\n"
		":rtype: int\n"
		":return: helics_true if it has been updated since the last value retrieval.\n"
		""},
	 { "helicsInputLastUpdateTime", _wrap_helicsInputLastUpdateTime, METH_O, "Get the last time a subscription was updated."},
	 { "helicsInputClearUpdate", _wrap_helicsInputClearUpdate, METH_O, "Clear the updated flag from an input."},
	 { "helicsFederateGetPublicationCount", _wrap_helicsFederateGetPublicationCount, METH_O, "\n"
		"Get the number of publications in a federate.\n"
		"\n"
		":rtype: int\n"
		":return: (-1) if fed was not a valid federate otherwise returns the number of publications.\n"
		""},
	 { "helicsFederateGetInputCount", _wrap_helicsFederateGetInputCount, METH_O, "\n"
		"Get the number of subscriptions in a federate.\n"
		"\n"
		":rtype: int\n"
		":return: (-1) if fed was not a valid federate otherwise returns the number of subscriptions.\n"
		""},
	 { "helicsFederateRegisterEndpoint", _wrap_helicsFederateRegisterEndpoint, METH_VARARGS, "\n"
		"Create an endpoint.\n"
		"\n"
		"The endpoint becomes part of the federate and is destroyed when the federate is freed\n"
		"         so there are no separate free functions for endpoints.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate object in which to create an endpoint must have been created\n"
		"              with helicsCreateMessageFederate or helicsCreateCombinationFederate.\n"
		":type name: string\n"
		":param name: The identifier for the endpoint. This will be prepended with the federate name for the global identifier.\n"
		":type type: string\n"
		":param type: A string describing the expected type of the publication (may be NULL).\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: An object containing the endpoint.\n"
		""},
	 { "helicsFederateRegisterGlobalEndpoint", _wrap_helicsFederateRegisterGlobalEndpoint, METH_VARARGS, "\n"
		"Create an endpoint.\n"
		"\n"
		"The endpoint becomes part of the federate and is destroyed when the federate is freed\n"
		"         so there are no separate free functions for endpoints.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate object in which to create an endpoint must have been created\n"
		"                  with helicsCreateMessageFederate or helicsCreateCombinationFederate.\n"
		":type name: string\n"
		":param name: The identifier for the endpoint, the given name is the global identifier.\n"
		":type type: string\n"
		":param type: A string describing the expected type of the publication (may be NULL).\n"
		"\n"
		":rtype: void\n"
		":return: An object containing the endpoint.\n"
		""},
	 { "helicsFederateGetEndpoint", _wrap_helicsFederateGetEndpoint, METH_VARARGS, "\n"
		"Get an endpoint object from a name.\n"
		"\n"
		":type fed: void\n"
		":param fed: The message federate object to use to get the endpoint.\n"
		":type name: string\n"
		":param name: The name of the endpoint.\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: A helics_endpoint object.\n"
		""},
	 { "helicsFederateGetEndpointByIndex", _wrap_helicsFederateGetEndpointByIndex, METH_VARARGS, "\n"
		"Get an endpoint by its index, typically already created via registerInterfaces file or something of that nature.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate object in which to create a publication.\n"
		":type index: int\n"
		":param index: The index of the publication to get.\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: A helics_endpoint.\n"
		""},
	 { "helicsEndpointIsValid", _wrap_helicsEndpointIsValid, METH_O, "\n"
		"Check if an endpoint is valid.\n"
		"\n"
		":type endpoint: void\n"
		":param endpoint: The endpoint object to check.\n"
		"\n"
		":rtype: int\n"
		":return: helics_true if the Endpoint object represents a valid endpoint.\n"
		""},
	 { "helicsEndpointSetDefaultDestination", _wrap_helicsEndpointSetDefaultDestination, METH_VARARGS, "\n"
		"Set the default destination for an endpoint if no other endpoint is given.\n"
		"\n"
		":type endpoint: void\n"
		":param endpoint: The endpoint to set the destination for.\n"
		":type dest: string\n"
		":param dest: A string naming the desired default endpoint.\n"
		""},
	 { "helicsEndpointGetDefaultDestination", _wrap_helicsEndpointGetDefaultDestination, METH_O, "\n"
		"Get the default destination for an endpoint.\n"
		"\n"
		":type endpoint: void\n"
		":param endpoint: The endpoint to set the destination for.\n"
		"\n"
		":rtype: string\n"
		":return: A string with the default destination.\n"
		""},
	 { "helicsEndpointSendMessageRaw", _wrap_helicsEndpointSendMessageRaw, METH_VARARGS, "\n"
		"Send a message to the specified destination.\n"
		"\n"
		":type endpoint: void\n"
		":param endpoint: The endpoint to send the data from.\n"
		":type dest: string\n"
		":param dest: The target destination.\n"
		"\n"
		"\n"
		"            \"\" to use the default destination.\n"
		"\n"
		":type data: void\n"
		":param data: The data to send.\n"
		""},
	 { "helicsEndpointSendEventRaw", _wrap_helicsEndpointSendEventRaw, METH_VARARGS, "\n"
		"Send a message at a specific time to the specified destination.\n"
		"\n"
		":type endpoint: void\n"
		":param endpoint: The endpoint to send the data from.\n"
		":type dest: string\n"
		":param dest: The target destination.\n"
		"\n"
		"\n"
		"            \"\" to use the default destination.\n"
		"\n"
		":type data: void\n"
		":param data: The data to send.\n"
		"\n"
		":type time: float\n"
		":param time: The time the message should be sent.\n"
		""},
	 { "helicsEndpointSendMessage", _wrap_helicsEndpointSendMessage, METH_VARARGS, "\n"
		"Send a message object from a specific endpoint.\n"
		"Deprecated: Use helicsEndpointSendMessageObject instead.\n"
		":type endpoint: void\n"
		":param endpoint: The endpoint to send the data from.\n"
		":type message: :py:class:`helics_message`\n"
		":param message: The actual message to send.\n"
		""},
	 { "helicsEndpointSendMessageObject", _wrap_helicsEndpointSendMessageObject, METH_VARARGS, "\n"
		"Send a message object from a specific endpoint.\n"
		"\n"
		":type endpoint: void\n"
		":param endpoint: The endpoint to send the data from.\n"
		":type message: void\n"
		":param message: The actual message to send which will be copied.\n"
		""},
	 { "helicsEndpointSendMessageObjectZeroCopy", _wrap_helicsEndpointSendMessageObjectZeroCopy, METH_VARARGS, "\n"
		"Send a message object from a specific endpoint, the message will not be copied and the message object will no longer be valid\n"
		"after the call.\n"
		"\n"
		":type endpoint: void\n"
		":param endpoint: The endpoint to send the data from.\n"
		":type message: void\n"
		":param message: The actual message to send which will be copied.\n"
		""},
	 { "helicsEndpointSubscribe", _wrap_helicsEndpointSubscribe, METH_VARARGS, "\n"
		"Subscribe an endpoint to a publication.\n"
		"\n"
		":type endpoint: void\n"
		":param endpoint: The endpoint to use.\n"
		":type key: string\n"
		":param key: The name of the publication.\n"
		""},
	 { "helicsFederateHasMessage", _wrap_helicsFederateHasMessage, METH_O, "\n"
		"Check if the federate has any outstanding messages.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to check.\n"
		"\n"
		":rtype: int\n"
		":return: helics_true if the federate has a message waiting, helics_false otherwise.\n"
		""},
	 { "helicsEndpointHasMessage", _wrap_helicsEndpointHasMessage, METH_O, "\n"
		"Check if a given endpoint has any unread messages.\n"
		"\n"
		":type endpoint: void\n"
		":param endpoint: The endpoint to check.\n"
		"\n"
		":rtype: int\n"
		":return: helics_true if the endpoint has a message, helics_false otherwise.\n"
		""},
	 { "helicsFederatePendingMessages", _wrap_helicsFederatePendingMessages, METH_O, "\n"
		"Returns the number of pending receives for the specified destination endpoint.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to get the number of waiting messages from.\n"
		""},
	 { "helicsEndpointPendingMessages", _wrap_helicsEndpointPendingMessages, METH_O, "\n"
		"Returns the number of pending receives for all endpoints of a particular federate.\n"
		"\n"
		":type endpoint: void\n"
		":param endpoint: The endpoint to query.\n"
		""},
	 { "helicsEndpointGetMessage", _wrap_helicsEndpointGetMessage, METH_O, "\n"
		"Receive a packet from a particular endpoint.\n"
		"\n"
		"Deprecated: This function is deprecated and will be removed in Helics 3.0.\n"
		"            Use helicsEndpointGetMessageObject instead.\n"
		"\n"
		"endpoint The identifier for the endpoint.\n"
		"\n"
		":rtype: :py:class:`helics_message`\n"
		":return: A message object.\n"
		""},
	 { "helicsEndpointGetMessageObject", _wrap_helicsEndpointGetMessageObject, METH_O, "\n"
		"Receive a packet from a particular endpoint.\n"
		"\n"
		"endpoint The identifier for the endpoint.\n"
		"\n"
		":rtype: void\n"
		":return: A message object.\n"
		""},
	 { "helicsEndpointCreateMessageObject", _wrap_helicsEndpointCreateMessageObject, METH_O, "\n"
		"Create a new empty message object.\n"
		"\n"
		"The message is empty and isValid will return false since there is no data associated with the message yet.\n"
		"\n"
		":type endpoint: void\n"
		":param endpoint: The endpoint object to associate the message with.\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: A new helics_message_object.\n"
		""},
	 { "helicsFederateGetMessage", _wrap_helicsFederateGetMessage, METH_O, "\n"
		"Receive a communication message for any endpoint in the federate.\n"
		"\n"
		"Deprecated: This function is deprecated and will be removed in Helics 3.0.\n"
		"            Use helicsFederateGetMessageObject instead.\n"
		"\n"
		"The return order will be in order of endpoint creation.\n"
		"         So all messages that are available for the first endpoint, then all for the second, and so on.\n"
		"         Within a single endpoint, the messages are ordered by time, then source_id, then order of arrival.\n"
		"\n"
		":rtype: :py:class:`helics_message`\n"
		":return: A unique_ptr to a Message object containing the message data.\n"
		""},
	 { "helicsFederateGetMessageObject", _wrap_helicsFederateGetMessageObject, METH_O, "\n"
		"Receive a communication message for any endpoint in the federate.\n"
		"\n"
		"The return order will be in order of endpoint creation.\n"
		"         So all messages that are available for the first endpoint, then all for the second, and so on.\n"
		"         Within a single endpoint, the messages are ordered by time, then source_id, then order of arrival.\n"
		"\n"
		":rtype: void\n"
		":return: A helics_message_object which references the data in the message.\n"
		""},
	 { "helicsFederateCreateMessageObject", _wrap_helicsFederateCreateMessageObject, METH_O, "\n"
		"Create a new empty message object.\n"
		"\n"
		"The message is empty and isValid will return false since there is no data associated with the message yet.\n"
		"\n"
		":type fed: void\n"
		":param fed: the federate object to associate the message with\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: A helics_message_object containing the message data.\n"
		""},
	 { "helicsFederateClearMessages", _wrap_helicsFederateClearMessages, METH_O, "\n"
		"Clear all stored messages from a federate.\n"
		"\n"
		"This clears messages retrieved through helicsFederateGetMessage or helicsFederateGetMessageObject\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to clear the message for.\n"
		""},
	 { "helicsEndpointClearMessages", _wrap_helicsEndpointClearMessages, METH_O, "\n"
		"Clear all message from an endpoint.\n"
		"\n"
		"Deprecated: This function does nothing and will be removed.\n"
		"            Use helicsFederateClearMessages to free all messages,\n"
		"            or helicsMessageFree to clear an individual message.\n"
		"\n"
		":type endpoint: void\n"
		":param endpoint: The endpoint object to operate on.\n"
		""},
	 { "helicsEndpointGetType", _wrap_helicsEndpointGetType, METH_O, "\n"
		"Get the type specified for an endpoint.\n"
		"\n"
		":type endpoint: void\n"
		":param endpoint: The endpoint object in question.\n"
		"\n"
		":rtype: string\n"
		":return: The defined type of the endpoint.\n"
		""},
	 { "helicsEndpointGetName", _wrap_helicsEndpointGetName, METH_O, "\n"
		"Get the name of an endpoint.\n"
		"\n"
		":type endpoint: void\n"
		":param endpoint: The endpoint object in question.\n"
		"\n"
		":rtype: string\n"
		":return: The name of the endpoint.\n"
		""},
	 { "helicsFederateGetEndpointCount", _wrap_helicsFederateGetEndpointCount, METH_O, "\n"
		"Get the number of endpoints in a federate.\n"
		"\n"
		":type fed: void\n"
		":param fed: The message federate to query.\n"
		"\n"
		":rtype: int\n"
		":return: (-1) if fed was not a valid federate, otherwise returns the number of endpoints.\n"
		""},
	 { "helicsEndpointGetInfo", _wrap_helicsEndpointGetInfo, METH_O, "\n"
		"Get the data in the info field of a filter.\n"
		"\n"
		":type end: void\n"
		":param end: The filter to query.\n"
		"\n"
		":rtype: string\n"
		":return: A string with the info field string.\n"
		""},
	 { "helicsEndpointSetInfo", _wrap_helicsEndpointSetInfo, METH_VARARGS, "\n"
		"Set the data in the info field for a filter.\n"
		"\n"
		":type end: void\n"
		":param end: The endpoint to query.\n"
		":type info: string\n"
		":param info: The string to set.\n"
		""},
	 { "helicsEndpointSetOption", _wrap_helicsEndpointSetOption, METH_VARARGS, "\n"
		"Set a handle option on an endpoint.\n"
		"\n"
		":type end: void\n"
		":param end: The endpoint to modify.\n"
		":type option: int\n"
		":param option: Integer code for the option to set /ref helics_handle_options.\n"
		":type value: int\n"
		":param value: The value to set the option to.\n"
		""},
	 { "helicsEndpointGetOption", _wrap_helicsEndpointGetOption, METH_VARARGS, "\n"
		"Set a handle option on an endpoint.\n"
		"\n"
		":type end: void\n"
		":param end: The endpoint to modify.\n"
		":type option: int\n"
		":param option: Integer code for the option to set /ref helics_handle_options.\n"
		":rtype: int\n"
		":return: the value of the option, for boolean options will be 0 or 1\n"
		""},
	 { "helicsMessageGetSource", _wrap_helicsMessageGetSource, METH_O, "\n"
		"Get the source endpoint of a message.\n"
		"\n"
		":type message: void\n"
		":param message: The message object in question.\n"
		"\n"
		":rtype: string\n"
		":return: A string with the source endpoint.\n"
		""},
	 { "helicsMessageGetDestination", _wrap_helicsMessageGetDestination, METH_O, "\n"
		"Get the destination endpoint of a message.\n"
		"\n"
		":type message: void\n"
		":param message: The message object in question.\n"
		"\n"
		":rtype: string\n"
		":return: A string with the destination endpoint.\n"
		""},
	 { "helicsMessageGetOriginalSource", _wrap_helicsMessageGetOriginalSource, METH_O, "\n"
		"Get the original source endpoint of a message, the source may have been modified by filters or other actions.\n"
		"\n"
		":type message: void\n"
		":param message: The message object in question.\n"
		"\n"
		":rtype: string\n"
		":return: A string with the source of a message.\n"
		""},
	 { "helicsMessageGetOriginalDestination", _wrap_helicsMessageGetOriginalDestination, METH_O, "\n"
		"Get the original destination endpoint of a message, the destination may have been modified by filters or other actions.\n"
		"\n"
		":type message: void\n"
		":param message: The message object in question.\n"
		"\n"
		":rtype: string\n"
		":return: A string with the original destination of a message.\n"
		""},
	 { "helicsMessageGetTime", _wrap_helicsMessageGetTime, METH_O, "\n"
		"Get the helics time associated with a message.\n"
		"\n"
		":type message: void\n"
		":param message: The message object in question.\n"
		"\n"
		":rtype: float\n"
		":return: The time associated with a message.\n"
		""},
	 { "helicsMessageGetString", _wrap_helicsMessageGetString, METH_O, "\n"
		"Get the payload of a message as a string.\n"
		"\n"
		":type message: void\n"
		":param message: The message object in question.\n"
		"\n"
		":rtype: string\n"
		":return: A string representing the payload of a message.\n"
		""},
	 { "helicsMessageGetMessageID", _wrap_helicsMessageGetMessageID, METH_O, "\n"
		"Get the messageID of a message.\n"
		"\n"
		":type message: void\n"
		":param message: The message object in question.\n"
		"\n"
		":rtype: int\n"
		":return: The messageID.\n"
		""},
	 { "helicsMessageCheckFlag", _wrap_helicsMessageCheckFlag, METH_VARARGS, "\n"
		"Check if a flag is set on a message.\n"
		"\n"
		":type message: void\n"
		":param message: The message object in question.\n"
		":type flag: int\n"
		":param flag: The flag to check should be between [0,15].\n"
		"\n"
		":rtype: int\n"
		":return: The flags associated with a message.\n"
		""},
	 { "helicsMessageGetRawDataSize", _wrap_helicsMessageGetRawDataSize, METH_O, "\n"
		"Get the size of the data payload in bytes.\n"
		"\n"
		":type message: void\n"
		":param message: The message object in question.\n"
		"\n"
		":rtype: int\n"
		":return: The size of the data payload.\n"
		""},
	 { "helicsMessageGetRawData", _wrap_helicsMessageGetRawData, METH_O, "\n"
		"Get the raw data for a message object.\n"
		"\n"
		":type message: void\n"
		":param message: A message object to get the data for.\n"
		"\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: Raw string data.\n"
		""},
	 { "helicsMessageIsValid", _wrap_helicsMessageIsValid, METH_O, "\n"
		"A check if the message contains a valid payload.\n"
		"\n"
		":type message: void\n"
		":param message: The message object in question.\n"
		"\n"
		":rtype: int\n"
		":return: helics_true if the message contains a payload.\n"
		""},
	 { "helicsMessageSetSource", _wrap_helicsMessageSetSource, METH_VARARGS, "\n"
		"Set the source of a message.\n"
		"\n"
		":type message: void\n"
		":param message: The message object in question.\n"
		":type src: string\n"
		":param src: A string containing the source.\n"
		""},
	 { "helicsMessageSetDestination", _wrap_helicsMessageSetDestination, METH_VARARGS, "\n"
		"Set the destination of a message.\n"
		"\n"
		":type message: void\n"
		":param message: The message object in question.\n"
		":type dest: string\n"
		":param dest: A string containing the new destination.\n"
		""},
	 { "helicsMessageSetOriginalSource", _wrap_helicsMessageSetOriginalSource, METH_VARARGS, "\n"
		"Set the original source of a message.\n"
		"\n"
		":type message: void\n"
		":param message: The message object in question.\n"
		":type src: string\n"
		":param src: A string containing the new original source.\n"
		""},
	 { "helicsMessageSetOriginalDestination", _wrap_helicsMessageSetOriginalDestination, METH_VARARGS, "\n"
		"Set the original destination of a message.\n"
		"\n"
		":type message: void\n"
		":param message: The message object in question.\n"
		":type dest: string\n"
		":param dest: A string containing the new original source.\n"
		""},
	 { "helicsMessageSetTime", _wrap_helicsMessageSetTime, METH_VARARGS, "\n"
		"Set the delivery time for a message.\n"
		"\n"
		":type message: void\n"
		":param message: The message object in question.\n"
		":type time: float\n"
		":param time: The time the message should be delivered.\n"
		""},
	 { "helicsMessageReserve", _wrap_helicsMessageReserve, METH_VARARGS, "\n"
		"Reserve space in a buffer but don't actually resize.\n"
		"\n"
		"The message data buffer will be reserved but not resized.\n"
		"\n"
		":type message: void\n"
		":param message: The message object in question.\n"
		":type reserveSize: int\n"
		":param reserveSize: The number of bytes to reserve in the message object.\n"
		""},
	 { "helicsMessageSetMessageID", _wrap_helicsMessageSetMessageID, METH_VARARGS, "\n"
		"Set the message ID for the message.\n"
		"\n"
		"Normally this is not needed and the core of HELICS will adjust as needed.\n"
		"\n"
		":type message: void\n"
		":param message: The message object in question.\n"
		":type messageID: int32_t\n"
		":param messageID: A new message ID.\n"
		""},
	 { "helicsMessageClearFlags", _wrap_helicsMessageClearFlags, METH_O, "\n"
		"Clear the flags of a message.\n"
		"\n"
		":type message: void\n"
		":param message: The message object in question\n"
		""},
	 { "helicsMessageSetFlagOption", _wrap_helicsMessageSetFlagOption, METH_VARARGS, "\n"
		"Set a flag on a message.\n"
		"\n"
		":type message: void\n"
		":param message: The message object in question.\n"
		":type flag: int\n"
		":param flag: An index of a flag to set on the message.\n"
		":type flagValue: int\n"
		":param flagValue: The desired value of the flag.\n"
		""},
	 { "helicsMessageSetString", _wrap_helicsMessageSetString, METH_VARARGS, "\n"
		"Set the data payload of a message as a string.\n"
		"\n"
		":type message: void\n"
		":param message: The message object in question.\n"
		":type str: string\n"
		":param str: A string containing the message data.\n"
		""},
	 { "helicsMessageSetData", _wrap_helicsMessageSetData, METH_VARARGS, "\n"
		"Set the data payload of a message as raw data.\n"
		"\n"
		":type message: void\n"
		":param message: The message object in question.\n"
		":type data: void\n"
		":param data: A string containing the message data.\n"
		":type inputDataLength: int\n"
		":param inputDataLength: The length of the data to input.\n"
		""},
	 { "helicsMessageAppendData", _wrap_helicsMessageAppendData, METH_VARARGS, "\n"
		"Append data to the payload.\n"
		"\n"
		":type message: void\n"
		":param message: The message object in question.\n"
		":type data: void\n"
		":param data: A string containing the message data to append.\n"
		":type inputDataLength: int\n"
		":param inputDataLength: The length of the data to input.\n"
		""},
	 { "helicsMessageCopy", _wrap_helicsMessageCopy, METH_VARARGS, "\n"
		"Copy a message object.\n"
		"\n"
		":type source_message: void\n"
		":param source_message: The message object to copy from.\n"
		":type dest_message: void\n"
		":param dest_message: The message object to copy to.\n"
		""},
	 { "helicsMessageClone", _wrap_helicsMessageClone, METH_O, "\n"
		"Clone a message object.\n"
		"\n"
		":type message: void\n"
		":param message: The message object to copy from.\n"
		""},
	 { "helicsMessageFree", _wrap_helicsMessageFree, METH_O, "\n"
		"Free a message object from memory\n"
		"memory for message is managed so not using this function does not create memory leaks, this is an indication\n"
		"to the system that the memory for this message is done being used and can be reused for a new message.\n"
		"helicsFederateClearMessages() can also be used to clear up all stored messages at once\n"
		""},
	 { "helicsFederateRegisterFilter", _wrap_helicsFederateRegisterFilter, METH_VARARGS, "\n"
		"Create a source Filter on the specified federate.\n"
		"\n"
		"Filters can be created through a federate or a core, linking through a federate allows\n"
		"         a few extra features of name matching to function on the federate interface but otherwise equivalent behavior\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to register through.\n"
		":type type: int\n"
		":param type: The type of filter to create /ref helics_filter_type.\n"
		":type name: string\n"
		":param name: The name of the filter (can be NULL).\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: A helics_filter object.\n"
		""},
	 { "helicsFederateRegisterGlobalFilter", _wrap_helicsFederateRegisterGlobalFilter, METH_VARARGS, "\n"
		"Create a global source filter through a federate.\n"
		"\n"
		"Filters can be created through a federate or a core, linking through a federate allows\n"
		"         a few extra features of name matching to function on the federate interface but otherwise equivalent behavior.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to register through.\n"
		":type type: int\n"
		":param type: The type of filter to create /ref helics_filter_type.\n"
		":type name: string\n"
		":param name: The name of the filter (can be NULL).\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: A helics_filter object.\n"
		""},
	 { "helicsFederateRegisterCloningFilter", _wrap_helicsFederateRegisterCloningFilter, METH_VARARGS, "\n"
		"Create a cloning Filter on the specified federate.\n"
		"\n"
		"Cloning filters copy a message and send it to multiple locations, source and destination can be added\n"
		"         through other functions.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to register through.\n"
		":type name: string\n"
		":param name: The name of the filter (can be NULL).\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: A helics_filter object.\n"
		""},
	 { "helicsFederateRegisterGlobalCloningFilter", _wrap_helicsFederateRegisterGlobalCloningFilter, METH_VARARGS, "\n"
		"Create a global cloning Filter on the specified federate.\n"
		"\n"
		"Cloning filters copy a message and send it to multiple locations, source and destination can be added\n"
		"         through other functions.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate to register through.\n"
		":type name: string\n"
		":param name: The name of the filter (can be NULL).\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: A helics_filter object.\n"
		""},
	 { "helicsCoreRegisterFilter", _wrap_helicsCoreRegisterFilter, METH_VARARGS, "\n"
		"Create a source Filter on the specified core.\n"
		"\n"
		"Filters can be created through a federate or a core, linking through a federate allows\n"
		"         a few extra features of name matching to function on the federate interface but otherwise equivalent behavior.\n"
		"\n"
		":type core: void\n"
		":param core: The core to register through.\n"
		":type type: int\n"
		":param type: The type of filter to create /ref helics_filter_type.\n"
		":type name: string\n"
		":param name: The name of the filter (can be NULL).\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: A helics_filter object.\n"
		""},
	 { "helicsCoreRegisterCloningFilter", _wrap_helicsCoreRegisterCloningFilter, METH_VARARGS, "\n"
		"Create a cloning Filter on the specified core.\n"
		"\n"
		"Cloning filters copy a message and send it to multiple locations, source and destination can be added\n"
		"         through other functions.\n"
		"\n"
		":type core: void\n"
		":param core: The core to register through.\n"
		":type name: string\n"
		":param name: The name of the filter (can be NULL).\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: A helics_filter object.\n"
		""},
	 { "helicsFederateGetFilterCount", _wrap_helicsFederateGetFilterCount, METH_O, "\n"
		"Get the number of filters registered through a federate.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate object to use to get the filter.\n"
		"\n"
		":rtype: int\n"
		":return: A count of the number of filters registered through a federate.\n"
		""},
	 { "helicsFederateGetFilter", _wrap_helicsFederateGetFilter, METH_VARARGS, "\n"
		"Get a filter by its name, typically already created via registerInterfaces file or something of that nature.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate object to use to get the filter.\n"
		":type name: string\n"
		":param name: The name of the filter.\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: A helics_filter object, the object will not be valid and err will contain an error code if no filter with the specified name\n"
		"    exists.\n"
		""},
	 { "helicsFederateGetFilterByIndex", _wrap_helicsFederateGetFilterByIndex, METH_VARARGS, "\n"
		"Get a filter by its index, typically already created via registerInterfaces file or something of that nature.\n"
		"\n"
		":type fed: void\n"
		":param fed: The federate object in which to create a publication.\n"
		":type index: int\n"
		":param index: The index of the publication to get.\n"
		"\n"
		"\n"
		":rtype: void\n"
		":return: A helics_filter, which will be NULL if an invalid index is given.\n"
		""},
	 { "helicsFilterIsValid", _wrap_helicsFilterIsValid, METH_O, "\n"
		"Check if a filter is valid.\n"
		"\n"
		":type filt: void\n"
		":param filt: The filter object to check.\n"
		"\n"
		":rtype: int\n"
		":return: helics_true if the Filter object represents a valid filter.\n"
		""},
	 { "helicsFilterGetName", _wrap_helicsFilterGetName, METH_O, "\n"
		"Get the name of the filter and store in the given string.\n"
		"\n"
		":type filt: void\n"
		":param filt: The given filter.\n"
		"\n"
		":rtype: string\n"
		":return: A string with the name of the filter.\n"
		""},
	 { "helicsFilterSet", _wrap_helicsFilterSet, METH_VARARGS, "\n"
		"Set a property on a filter.\n"
		"\n"
		":type filt: void\n"
		":param filt: The filter to modify.\n"
		":type prop: string\n"
		":param prop: A string containing the property to set.\n"
		":type val: float\n"
		":param val: A numerical value for the property.\n"
		""},
	 { "helicsFilterSetString", _wrap_helicsFilterSetString, METH_VARARGS, "\n"
		"Set a string property on a filter.\n"
		"\n"
		":type filt: void\n"
		":param filt: The filter to modify.\n"
		":type prop: string\n"
		":param prop: A string containing the property to set.\n"
		":type val: string\n"
		":param val: A string containing the new value.\n"
		""},
	 { "helicsFilterAddDestinationTarget", _wrap_helicsFilterAddDestinationTarget, METH_VARARGS, "\n"
		"Add a destination target to a filter.\n"
		"\n"
		"All messages going to a destination are copied to the delivery address(es).\n"
		":type filt: void\n"
		":param filt: The given filter to add a destination target to.\n"
		":type dest: string\n"
		":param dest: The name of the endpoint to add as a destination target.\n"
		""},
	 { "helicsFilterAddSourceTarget", _wrap_helicsFilterAddSourceTarget, METH_VARARGS, "\n"
		"Add a source target to a filter.\n"
		"\n"
		"All messages coming from a source are copied to the delivery address(es).\n"
		"\n"
		":type filt: void\n"
		":param filt: The given filter.\n"
		":type source: string\n"
		":param source: The name of the endpoint to add as a source target.\n"
		""},
	 { "helicsFilterAddDeliveryEndpoint", _wrap_helicsFilterAddDeliveryEndpoint, METH_VARARGS, "\n"
		"Add a delivery endpoint to a cloning filter.\n"
		"\n"
		"All cloned messages are sent to the delivery address(es).\n"
		"\n"
		":type filt: void\n"
		":param filt: The given filter.\n"
		":type deliveryEndpoint: string\n"
		":param deliveryEndpoint: The name of the endpoint to deliver messages to.\n"
		""},
	 { "helicsFilterRemoveTarget", _wrap_helicsFilterRemoveTarget, METH_VARARGS, "\n"
		"Remove a destination target from a filter.\n"
		"\n"
		":type filt: void\n"
		":param filt: The given filter.\n"
		":type target: string\n"
		":param target: The named endpoint to remove as a target.\n"
		""},
	 { "helicsFilterRemoveDeliveryEndpoint", _wrap_helicsFilterRemoveDeliveryEndpoint, METH_VARARGS, "\n"
		"Remove a delivery destination from a cloning filter.\n"
		"\n"
		":type filt: void\n"
		":param filt: The given filter (must be a cloning filter).\n"
		":type deliveryEndpoint: string\n"
		":param deliveryEndpoint: A string with the delivery endpoint to remove.\n"
		""},
	 { "helicsFilterGetInfo", _wrap_helicsFilterGetInfo, METH_O, "\n"
		"Get the data in the info field of a filter.\n"
		"\n"
		":type filt: void\n"
		":param filt: The given filter.\n"
		"\n"
		":rtype: string\n"
		":return: A string with the info field string.\n"
		""},
	 { "helicsFilterSetInfo", _wrap_helicsFilterSetInfo, METH_VARARGS, "\n"
		"Set the data in the info field for a filter.\n"
		"\n"
		":type filt: void\n"
		":param filt: The given filter.\n"
		":type info: string\n"
		":param info: The string to set.\n"
		""},
	 { "helicsFilterSetOption", _wrap_helicsFilterSetOption, METH_VARARGS, "\n"
		"Set the data in the info field for a filter.\n"
		"\n"
		":type filt: void\n"
		":param filt: The given filter.\n"
		":type option: int\n"
		":param option: The option to set /ref helics_handle_options.\n"
		":type value: int\n"
		":param value: The value of the option commonly 0 for false 1 for true.\n"
		""},
	 { "helicsFilterGetOption", _wrap_helicsFilterGetOption, METH_VARARGS, "\n"
		"Get a handle option for the filter.\n"
		"\n"
		":type filt: void\n"
		":param filt: The given filter to query.\n"
		":type option: int\n"
		":param option: The option to query /ref helics_handle_options.\n"
		""},
	 { NULL, NULL, 0, NULL }
};

static PyMethodDef SwigMethods_proxydocs[] = {
	 { NULL, NULL, 0, NULL }
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (BEGIN) -------- */

static swig_type_info _swigt__p_char = {"_p_char", "char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_double = {"_p_double", "helics_time *|double *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_helics_complex = {"_p_helics_complex", "struct helics_complex *|helics_complex *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_helics_core_type = {"_p_helics_core_type", "enum helics_core_type *|helics_core_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_helics_data_type = {"_p_helics_data_type", "enum helics_data_type *|helics_data_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_helics_error = {"_p_helics_error", "struct helics_error *|helics_error *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_helics_error_types = {"_p_helics_error_types", "enum helics_error_types *|helics_error_types *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_helics_federate_flags = {"_p_helics_federate_flags", "enum helics_federate_flags *|helics_federate_flags *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_helics_federate_state = {"_p_helics_federate_state", "enum helics_federate_state *|helics_federate_state *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_helics_filter_type = {"_p_helics_filter_type", "enum helics_filter_type *|helics_filter_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_helics_handle_options = {"_p_helics_handle_options", "enum helics_handle_options *|helics_handle_options *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_helics_iteration_request = {"_p_helics_iteration_request", "enum helics_iteration_request *|helics_iteration_request *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_helics_iteration_result = {"_p_helics_iteration_result", "enum helics_iteration_result *|helics_iteration_result *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_helics_log_levels = {"_p_helics_log_levels", "enum helics_log_levels *|helics_log_levels *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_helics_message = {"_p_helics_message", "struct helics_message *|helics_message *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_helics_multi_input_mode = {"_p_helics_multi_input_mode", "enum helics_multi_input_mode *|helics_multi_input_mode *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_helics_properties = {"_p_helics_properties", "enum helics_properties *|helics_properties *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_int = {"_p_int", "helics_bool *|int *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_int16_t = {"_p_int16_t", "int16_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_int32_t = {"_p_int32_t", "int32_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_char = {"_p_p_char", "char **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_void = {"_p_void", "helics_publication|helics_core|helics_broker|helics_query|helics_filter|helics_federate_info|helics_message_object|helics_federate|helics_endpoint|void *|helics_input", 0, 0, (void*)0, 0};

static swig_type_info *swig_type_initial[] = {
  &_swigt__p_char,
  &_swigt__p_double,
  &_swigt__p_helics_complex,
  &_swigt__p_helics_core_type,
  &_swigt__p_helics_data_type,
  &_swigt__p_helics_error,
  &_swigt__p_helics_error_types,
  &_swigt__p_helics_federate_flags,
  &_swigt__p_helics_federate_state,
  &_swigt__p_helics_filter_type,
  &_swigt__p_helics_handle_options,
  &_swigt__p_helics_iteration_request,
  &_swigt__p_helics_iteration_result,
  &_swigt__p_helics_log_levels,
  &_swigt__p_helics_message,
  &_swigt__p_helics_multi_input_mode,
  &_swigt__p_helics_properties,
  &_swigt__p_int,
  &_swigt__p_int16_t,
  &_swigt__p_int32_t,
  &_swigt__p_p_char,
  &_swigt__p_void,
};

static swig_cast_info _swigc__p_char[] = {  {&_swigt__p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_double[] = {  {&_swigt__p_double, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_helics_complex[] = {  {&_swigt__p_helics_complex, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_helics_core_type[] = {  {&_swigt__p_helics_core_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_helics_data_type[] = {  {&_swigt__p_helics_data_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_helics_error[] = {  {&_swigt__p_helics_error, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_helics_error_types[] = {  {&_swigt__p_helics_error_types, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_helics_federate_flags[] = {  {&_swigt__p_helics_federate_flags, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_helics_federate_state[] = {  {&_swigt__p_helics_federate_state, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_helics_filter_type[] = {  {&_swigt__p_helics_filter_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_helics_handle_options[] = {  {&_swigt__p_helics_handle_options, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_helics_iteration_request[] = {  {&_swigt__p_helics_iteration_request, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_helics_iteration_result[] = {  {&_swigt__p_helics_iteration_result, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_helics_log_levels[] = {  {&_swigt__p_helics_log_levels, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_helics_message[] = {  {&_swigt__p_helics_message, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_helics_multi_input_mode[] = {  {&_swigt__p_helics_multi_input_mode, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_helics_properties[] = {  {&_swigt__p_helics_properties, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_int[] = {  {&_swigt__p_int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_int16_t[] = {  {&_swigt__p_int16_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_int32_t[] = {  {&_swigt__p_int32_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_char[] = {  {&_swigt__p_p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_void[] = {  {&_swigt__p_void, 0, 0, 0},{0, 0, 0, 0}};

static swig_cast_info *swig_cast_initial[] = {
  _swigc__p_char,
  _swigc__p_double,
  _swigc__p_helics_complex,
  _swigc__p_helics_core_type,
  _swigc__p_helics_data_type,
  _swigc__p_helics_error,
  _swigc__p_helics_error_types,
  _swigc__p_helics_federate_flags,
  _swigc__p_helics_federate_state,
  _swigc__p_helics_filter_type,
  _swigc__p_helics_handle_options,
  _swigc__p_helics_iteration_request,
  _swigc__p_helics_iteration_result,
  _swigc__p_helics_log_levels,
  _swigc__p_helics_message,
  _swigc__p_helics_multi_input_mode,
  _swigc__p_helics_properties,
  _swigc__p_int,
  _swigc__p_int16_t,
  _swigc__p_int32_t,
  _swigc__p_p_char,
  _swigc__p_void,
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (END) -------- */

static swig_const_info swig_const_table[] = {
{0, 0, 0, 0.0, 0, 0}};

#ifdef __cplusplus
}
#endif
/* -----------------------------------------------------------------------------
 * Type initialization:
 * This problem is tough by the requirement that no dynamic
 * memory is used. Also, since swig_type_info structures store pointers to
 * swig_cast_info structures and swig_cast_info structures store pointers back
 * to swig_type_info structures, we need some lookup code at initialization.
 * The idea is that swig generates all the structures that are needed.
 * The runtime then collects these partially filled structures.
 * The SWIG_InitializeModule function takes these initial arrays out of
 * swig_module, and does all the lookup, filling in the swig_module.types
 * array with the correct data and linking the correct swig_cast_info
 * structures together.
 *
 * The generated swig_type_info structures are assigned statically to an initial
 * array. We just loop through that array, and handle each type individually.
 * First we lookup if this type has been already loaded, and if so, use the
 * loaded structure instead of the generated one. Then we have to fill in the
 * cast linked list. The cast data is initially stored in something like a
 * two-dimensional array. Each row corresponds to a type (there are the same
 * number of rows as there are in the swig_type_initial array). Each entry in
 * a column is one of the swig_cast_info structures for that type.
 * The cast_initial array is actually an array of arrays, because each row has
 * a variable number of columns. So to actually build the cast linked list,
 * we find the array of casts associated with the type, and loop through it
 * adding the casts to the list. The one last trick we need to do is making
 * sure the type pointer in the swig_cast_info struct is correct.
 *
 * First off, we lookup the cast->type name to see if it is already loaded.
 * There are three cases to handle:
 *  1) If the cast->type has already been loaded AND the type we are adding
 *     casting info to has not been loaded (it is in this module), THEN we
 *     replace the cast->type pointer with the type pointer that has already
 *     been loaded.
 *  2) If BOTH types (the one we are adding casting info to, and the
 *     cast->type) are loaded, THEN the cast info has already been loaded by
 *     the previous module so we just ignore it.
 *  3) Finally, if cast->type has not already been loaded, then we add that
 *     swig_cast_info to the linked list (because the cast->type) pointer will
 *     be correct.
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#if 0
} /* c-mode */
#endif
#endif

#if 0
#define SWIGRUNTIME_DEBUG
#endif


SWIGRUNTIME void
SWIG_InitializeModule(void *clientdata) {
  size_t i;
  swig_module_info *module_head, *iter;
  int init;
  
  /* check to see if the circular list has been setup, if not, set it up */
  if (swig_module.next==0) {
    /* Initialize the swig_module */
    swig_module.type_initial = swig_type_initial;
    swig_module.cast_initial = swig_cast_initial;
    swig_module.next = &swig_module;
    init = 1;
  } else {
    init = 0;
  }
  
  /* Try and load any already created modules */
  module_head = SWIG_GetModule(clientdata);
  if (!module_head) {
    /* This is the first module loaded for this interpreter */
    /* so set the swig module into the interpreter */
    SWIG_SetModule(clientdata, &swig_module);
  } else {
    /* the interpreter has loaded a SWIG module, but has it loaded this one? */
    iter=module_head;
    do {
      if (iter==&swig_module) {
        /* Our module is already in the list, so there's nothing more to do. */
        return;
      }
      iter=iter->next;
    } while (iter!= module_head);
    
    /* otherwise we must add our module into the list */
    swig_module.next = module_head->next;
    module_head->next = &swig_module;
  }
  
  /* When multiple interpreters are used, a module could have already been initialized in
       a different interpreter, but not yet have a pointer in this interpreter.
       In this case, we do not want to continue adding types... everything should be
       set up already */
  if (init == 0) return;
  
  /* Now work on filling in swig_module.types */
#ifdef SWIGRUNTIME_DEBUG
  printf("SWIG_InitializeModule: size %lu\n", (unsigned long)swig_module.size);
#endif
  for (i = 0; i < swig_module.size; ++i) {
    swig_type_info *type = 0;
    swig_type_info *ret;
    swig_cast_info *cast;
    
#ifdef SWIGRUNTIME_DEBUG
    printf("SWIG_InitializeModule: type %lu %s\n", (unsigned long)i, swig_module.type_initial[i]->name);
#endif
    
    /* if there is another module already loaded */
    if (swig_module.next != &swig_module) {
      type = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, swig_module.type_initial[i]->name);
    }
    if (type) {
      /* Overwrite clientdata field */
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: found type %s\n", type->name);
#endif
      if (swig_module.type_initial[i]->clientdata) {
        type->clientdata = swig_module.type_initial[i]->clientdata;
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: found and overwrite type %s \n", type->name);
#endif
      }
    } else {
      type = swig_module.type_initial[i];
    }
    
    /* Insert casting types */
    cast = swig_module.cast_initial[i];
    while (cast->type) {
      /* Don't need to add information already in the list */
      ret = 0;
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: look cast %s\n", cast->type->name);
#endif
      if (swig_module.next != &swig_module) {
        ret = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, cast->type->name);
#ifdef SWIGRUNTIME_DEBUG
        if (ret) printf("SWIG_InitializeModule: found cast %s\n", ret->name);
#endif
      }
      if (ret) {
        if (type == swig_module.type_initial[i]) {
#ifdef SWIGRUNTIME_DEBUG
          printf("SWIG_InitializeModule: skip old type %s\n", ret->name);
#endif
          cast->type = ret;
          ret = 0;
        } else {
          /* Check for casting already in the list */
          swig_cast_info *ocast = SWIG_TypeCheck(ret->name, type);
#ifdef SWIGRUNTIME_DEBUG
          if (ocast) printf("SWIG_InitializeModule: skip old cast %s\n", ret->name);
#endif
          if (!ocast) ret = 0;
        }
      }
      
      if (!ret) {
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: adding cast %s\n", cast->type->name);
#endif
        if (type->cast) {
          type->cast->prev = cast;
          cast->next = type->cast;
        }
        type->cast = cast;
      }
      cast++;
    }
    /* Set entry in modules->types array equal to the type */
    swig_module.types[i] = type;
  }
  swig_module.types[i] = 0;
  
#ifdef SWIGRUNTIME_DEBUG
  printf("**** SWIG_InitializeModule: Cast List ******\n");
  for (i = 0; i < swig_module.size; ++i) {
    int j = 0;
    swig_cast_info *cast = swig_module.cast_initial[i];
    printf("SWIG_InitializeModule: type %lu %s\n", (unsigned long)i, swig_module.type_initial[i]->name);
    while (cast->type) {
      printf("SWIG_InitializeModule: cast type %s\n", cast->type->name);
      cast++;
      ++j;
    }
    printf("---- Total casts: %d\n",j);
  }
  printf("**** SWIG_InitializeModule: Cast List ******\n");
#endif
}

/* This function will propagate the clientdata field of type to
* any new swig_type_info structures that have been added into the list
* of equivalent types.  It is like calling
* SWIG_TypeClientData(type, clientdata) a second time.
*/
SWIGRUNTIME void
SWIG_PropagateClientData(void) {
  size_t i;
  swig_cast_info *equiv;
  static int init_run = 0;
  
  if (init_run) return;
  init_run = 1;
  
  for (i = 0; i < swig_module.size; i++) {
    if (swig_module.types[i]->clientdata) {
      equiv = swig_module.types[i]->cast;
      while (equiv) {
        if (!equiv->converter) {
          if (equiv->type && !equiv->type->clientdata)
          SWIG_TypeClientData(equiv->type, swig_module.types[i]->clientdata);
        }
        equiv = equiv->next;
      }
    }
  }
}

#ifdef __cplusplus
#if 0
{
  /* c-mode */
#endif
}
#endif



#ifdef __cplusplus
extern "C" {
#endif
  
  /* Python-specific SWIG API */
#define SWIG_newvarlink()                             SWIG_Python_newvarlink()
#define SWIG_addvarlink(p, name, get_attr, set_attr)  SWIG_Python_addvarlink(p, name, get_attr, set_attr)
#define SWIG_InstallConstants(d, constants)           SWIG_Python_InstallConstants(d, constants)
  
  /* -----------------------------------------------------------------------------
   * global variable support code.
   * ----------------------------------------------------------------------------- */
  
  typedef struct swig_globalvar {
    char       *name;                  /* Name of global variable */
    PyObject *(*get_attr)(void);       /* Return the current value */
    int       (*set_attr)(PyObject *); /* Set the value */
    struct swig_globalvar *next;
  } swig_globalvar;
  
  typedef struct swig_varlinkobject {
    PyObject_HEAD
    swig_globalvar *vars;
  } swig_varlinkobject;
  
  SWIGINTERN PyObject *
  swig_varlink_repr(swig_varlinkobject *SWIGUNUSEDPARM(v)) {
#if PY_VERSION_HEX >= 0x03000000
    return PyUnicode_InternFromString("<Swig global variables>");
#else
    return PyString_FromString("<Swig global variables>");
#endif
  }
  
  SWIGINTERN PyObject *
  swig_varlink_str(swig_varlinkobject *v) {
#if PY_VERSION_HEX >= 0x03000000
    PyObject *str = PyUnicode_InternFromString("(");
    PyObject *tail;
    PyObject *joined;
    swig_globalvar *var;
    for (var = v->vars; var; var=var->next) {
      tail = PyUnicode_FromString(var->name);
      joined = PyUnicode_Concat(str, tail);
      Py_DecRef(str);
      Py_DecRef(tail);
      str = joined;
      if (var->next) {
        tail = PyUnicode_InternFromString(", ");
        joined = PyUnicode_Concat(str, tail);
        Py_DecRef(str);
        Py_DecRef(tail);
        str = joined;
      }
    }
    tail = PyUnicode_InternFromString(")");
    joined = PyUnicode_Concat(str, tail);
    Py_DecRef(str);
    Py_DecRef(tail);
    str = joined;
#else
    PyObject *str = PyString_FromString("(");
    swig_globalvar *var;
    for (var = v->vars; var; var=var->next) {
      PyString_ConcatAndDel(&str,PyString_FromString(var->name));
      if (var->next) PyString_ConcatAndDel(&str,PyString_FromString(", "));
    }
    PyString_ConcatAndDel(&str,PyString_FromString(")"));
#endif
    return str;
  }
  
  SWIGINTERN void
  swig_varlink_dealloc(swig_varlinkobject *v) {
    swig_globalvar *var = v->vars;
    while (var) {
      swig_globalvar *n = var->next;
      free(var->name);
      free(var);
      var = n;
    }
  }
  
  SWIGINTERN PyObject *
  swig_varlink_getattr(swig_varlinkobject *v, char *n) {
    PyObject *res = NULL;
    swig_globalvar *var = v->vars;
    while (var) {
      if (strcmp(var->name,n) == 0) {
        res = (*var->get_attr)();
        break;
      }
      var = var->next;
    }
    if (res == NULL && !PyErr_Occurred()) {
      PyErr_Format(PyExc_AttributeError, "Unknown C global variable '%s'", n);
    }
    return res;
  }
  
  SWIGINTERN int
  swig_varlink_setattr(swig_varlinkobject *v, char *n, PyObject *p) {
    int res = 1;
    swig_globalvar *var = v->vars;
    while (var) {
      if (strcmp(var->name,n) == 0) {
        res = (*var->set_attr)(p);
        break;
      }
      var = var->next;
    }
    if (res == 1 && !PyErr_Occurred()) {
      PyErr_Format(PyExc_AttributeError, "Unknown C global variable '%s'", n);
    }
    return res;
  }
  
  SWIGINTERN PyTypeObject*
  swig_varlink_type(void) {
    static char varlink__doc__[] = "Swig var link object";
    static PyTypeObject varlink_type;
    static int type_init = 0;
    if (!type_init) {
      const PyTypeObject tmp = {
#if PY_VERSION_HEX >= 0x03000000
        PyVarObject_HEAD_INIT(NULL, 0)
#else
        PyObject_HEAD_INIT(NULL)
        0,                                  /* ob_size */
#endif
        "swigvarlink",                      /* tp_name */
        sizeof(swig_varlinkobject),         /* tp_basicsize */
        0,                                  /* tp_itemsize */
        (destructor) swig_varlink_dealloc,  /* tp_dealloc */
        0,                                  /* tp_print */
        (getattrfunc) swig_varlink_getattr, /* tp_getattr */
        (setattrfunc) swig_varlink_setattr, /* tp_setattr */
        0,                                  /* tp_compare */
        (reprfunc) swig_varlink_repr,       /* tp_repr */
        0,                                  /* tp_as_number */
        0,                                  /* tp_as_sequence */
        0,                                  /* tp_as_mapping */
        0,                                  /* tp_hash */
        0,                                  /* tp_call */
        (reprfunc) swig_varlink_str,        /* tp_str */
        0,                                  /* tp_getattro */
        0,                                  /* tp_setattro */
        0,                                  /* tp_as_buffer */
        0,                                  /* tp_flags */
        varlink__doc__,                     /* tp_doc */
        0,                                  /* tp_traverse */
        0,                                  /* tp_clear */
        0,                                  /* tp_richcompare */
        0,                                  /* tp_weaklistoffset */
        0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* tp_iter -> tp_weaklist */
        0,                                  /* tp_del */
        0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
        0,                                  /* tp_finalize */
#endif
#ifdef COUNT_ALLOCS
        0,                                  /* tp_allocs */
        0,                                  /* tp_frees */
        0,                                  /* tp_maxalloc */
        0,                                  /* tp_prev */
        0                                   /* tp_next */
#endif
      };
      varlink_type = tmp;
      type_init = 1;
      if (PyType_Ready(&varlink_type) < 0)
      return NULL;
    }
    return &varlink_type;
  }
  
  /* Create a variable linking object for use later */
  SWIGINTERN PyObject *
  SWIG_Python_newvarlink(void) {
    swig_varlinkobject *result = PyObject_NEW(swig_varlinkobject, swig_varlink_type());
    if (result) {
      result->vars = 0;
    }
    return ((PyObject*) result);
  }
  
  SWIGINTERN void 
  SWIG_Python_addvarlink(PyObject *p, const char *name, PyObject *(*get_attr)(void), int (*set_attr)(PyObject *p)) {
    swig_varlinkobject *v = (swig_varlinkobject *) p;
    swig_globalvar *gv = (swig_globalvar *) malloc(sizeof(swig_globalvar));
    if (gv) {
      size_t size = strlen(name)+1;
      gv->name = (char *)malloc(size);
      if (gv->name) {
        memcpy(gv->name, name, size);
        gv->get_attr = get_attr;
        gv->set_attr = set_attr;
        gv->next = v->vars;
      }
    }
    v->vars = gv;
  }
  
  SWIGINTERN PyObject *
  SWIG_globals(void) {
    static PyObject *globals = 0;
    if (!globals) {
      globals = SWIG_newvarlink();
    }
    return globals;
  }
  
  /* -----------------------------------------------------------------------------
   * constants/methods manipulation
   * ----------------------------------------------------------------------------- */
  
  /* Install Constants */
  SWIGINTERN void
  SWIG_Python_InstallConstants(PyObject *d, swig_const_info constants[]) {
    PyObject *obj = 0;
    size_t i;
    for (i = 0; constants[i].type; ++i) {
      switch(constants[i].type) {
      case SWIG_PY_POINTER:
        obj = SWIG_InternalNewPointerObj(constants[i].pvalue, *(constants[i]).ptype,0);
        break;
      case SWIG_PY_BINARY:
        obj = SWIG_NewPackedObj(constants[i].pvalue, constants[i].lvalue, *(constants[i].ptype));
        break;
      default:
        obj = 0;
        break;
      }
      if (obj) {
        PyDict_SetItemString(d, constants[i].name, obj);
        Py_DECREF(obj);
      }
    }
  }
  
  /* -----------------------------------------------------------------------------*/
  /* Fix SwigMethods to carry the callback ptrs when needed */
  /* -----------------------------------------------------------------------------*/
  
  SWIGINTERN void
  SWIG_Python_FixMethods(PyMethodDef *methods,
    swig_const_info *const_table,
    swig_type_info **types,
    swig_type_info **types_initial) {
    size_t i;
    for (i = 0; methods[i].ml_name; ++i) {
      const char *c = methods[i].ml_doc;
      if (!c) continue;
      c = strstr(c, "swig_ptr: ");
      if (c) {
        int j;
        swig_const_info *ci = 0;
        const char *name = c + 10;
        for (j = 0; const_table[j].type; ++j) {
          if (strncmp(const_table[j].name, name, 
              strlen(const_table[j].name)) == 0) {
            ci = &(const_table[j]);
            break;
          }
        }
        if (ci) {
          void *ptr = (ci->type == SWIG_PY_POINTER) ? ci->pvalue : 0;
          if (ptr) {
            size_t shift = (ci->ptype) - types;
            swig_type_info *ty = types_initial[shift];
            size_t ldoc = (c - methods[i].ml_doc);
            size_t lptr = strlen(ty->name)+2*sizeof(void*)+2;
            char *ndoc = (char*)malloc(ldoc + lptr + 10);
            if (ndoc) {
              char *buff = ndoc;
              memcpy(buff, methods[i].ml_doc, ldoc);
              buff += ldoc;
              memcpy(buff, "swig_ptr: ", 10);
              buff += 10;
              SWIG_PackVoidPtr(buff, ptr, ty->name, lptr);
              methods[i].ml_doc = ndoc;
            }
          }
        }
      }
    }
  } 
  
  /* -----------------------------------------------------------------------------
   * Method creation and docstring support functions
   * ----------------------------------------------------------------------------- */
  
  /* -----------------------------------------------------------------------------
   * Function to find the method definition with the correct docstring for the
   * proxy module as opposed to the low-level API
   * ----------------------------------------------------------------------------- */
  
  SWIGINTERN PyMethodDef *SWIG_PythonGetProxyDoc(const char *name) {
    /* Find the function in the modified method table */
    size_t offset = 0;
    int found = 0;
    while (SwigMethods_proxydocs[offset].ml_meth != NULL) {
      if (strcmp(SwigMethods_proxydocs[offset].ml_name, name) == 0) {
        found = 1;
        break;
      }
      offset++;
    }
    /* Use the copy with the modified docstring if available */
    return found ? &SwigMethods_proxydocs[offset] : NULL;
  }
  
  /* -----------------------------------------------------------------------------
   * Wrapper of PyInstanceMethod_New() used in Python 3
   * It is exported to the generated module, used for -fastproxy
   * ----------------------------------------------------------------------------- */
  
  SWIGINTERN PyObject *SWIG_PyInstanceMethod_New(PyObject *SWIGUNUSEDPARM(self), PyObject *func) {
    if (PyCFunction_Check(func)) {
      PyCFunctionObject *funcobj = (PyCFunctionObject *)func;
      PyMethodDef *ml = SWIG_PythonGetProxyDoc(funcobj->m_ml->ml_name);
      if (ml)
      func = PyCFunction_NewEx(ml, funcobj->m_self, funcobj->m_module);
    }
#if PY_VERSION_HEX >= 0x03000000
    return PyInstanceMethod_New(func);
#else
    return PyMethod_New(func, NULL, NULL);
#endif
  }
  
  /* -----------------------------------------------------------------------------
   * Wrapper of PyStaticMethod_New()
   * It is exported to the generated module, used for -fastproxy
   * ----------------------------------------------------------------------------- */
  
  SWIGINTERN PyObject *SWIG_PyStaticMethod_New(PyObject *SWIGUNUSEDPARM(self), PyObject *func) {
    if (PyCFunction_Check(func)) {
      PyCFunctionObject *funcobj = (PyCFunctionObject *)func;
      PyMethodDef *ml = SWIG_PythonGetProxyDoc(funcobj->m_ml->ml_name);
      if (ml)
      func = PyCFunction_NewEx(ml, funcobj->m_self, funcobj->m_module);
    }
    return PyStaticMethod_New(func);
  }
  
#ifdef __cplusplus
}
#endif

/* -----------------------------------------------------------------------------*
 *  Partial Init method
 * -----------------------------------------------------------------------------*/

#ifdef __cplusplus
extern "C"
#endif

SWIGEXPORT 
#if PY_VERSION_HEX >= 0x03000000
PyObject*
#else
void
#endif
SWIG_init(void) {
  PyObject *m, *d, *md, *globals;
  
#if PY_VERSION_HEX >= 0x03000000
  static struct PyModuleDef SWIG_module = {
    PyModuleDef_HEAD_INIT,
    SWIG_name,
    NULL,
    -1,
    SwigMethods,
    NULL,
    NULL,
    NULL,
    NULL
  };
#endif
  
#if defined(SWIGPYTHON_BUILTIN)
  static SwigPyClientData SwigPyObject_clientdata = {
    0, 0, 0, 0, 0, 0, 0
  };
  static PyGetSetDef this_getset_def = {
    (char *)"this", &SwigPyBuiltin_ThisClosure, NULL, NULL, NULL
  };
  static SwigPyGetSet thisown_getset_closure = {
    SwigPyObject_own,
    SwigPyObject_own
  };
  static PyGetSetDef thisown_getset_def = {
    (char *)"thisown", SwigPyBuiltin_GetterClosure, SwigPyBuiltin_SetterClosure, NULL, &thisown_getset_closure
  };
  PyTypeObject *builtin_pytype;
  int builtin_base_count;
  swig_type_info *builtin_basetype;
  PyObject *tuple;
  PyGetSetDescrObject *static_getset;
  PyTypeObject *metatype;
  PyTypeObject *swigpyobject;
  SwigPyClientData *cd;
  PyObject *public_interface, *public_symbol;
  PyObject *this_descr;
  PyObject *thisown_descr;
  PyObject *self = 0;
  int i;
  
  (void)builtin_pytype;
  (void)builtin_base_count;
  (void)builtin_basetype;
  (void)tuple;
  (void)static_getset;
  (void)self;
  
  /* Metaclass is used to implement static member variables */
  metatype = SwigPyObjectType();
  assert(metatype);
#endif
  
  (void)globals;
  
  /* Create singletons now to avoid potential deadlocks with multi-threaded usage after module initialization */
  SWIG_This();
  SWIG_Python_TypeCache();
  SwigPyPacked_type();
#ifndef SWIGPYTHON_BUILTIN
  SwigPyObject_type();
#endif
  
  /* Fix SwigMethods to carry the callback ptrs when needed */
  SWIG_Python_FixMethods(SwigMethods, swig_const_table, swig_types, swig_type_initial);
  
#if PY_VERSION_HEX >= 0x03000000
  m = PyModule_Create(&SWIG_module);
#else
  m = Py_InitModule(SWIG_name, SwigMethods);
#endif
  
  md = d = PyModule_GetDict(m);
  (void)md;
  
  SWIG_InitializeModule(0);
  
#ifdef SWIGPYTHON_BUILTIN
  swigpyobject = SwigPyObject_TypeOnce();
  
  SwigPyObject_stype = SWIG_MangledTypeQuery("_p_SwigPyObject");
  assert(SwigPyObject_stype);
  cd = (SwigPyClientData*) SwigPyObject_stype->clientdata;
  if (!cd) {
    SwigPyObject_stype->clientdata = &SwigPyObject_clientdata;
    SwigPyObject_clientdata.pytype = swigpyobject;
  } else if (swigpyobject->tp_basicsize != cd->pytype->tp_basicsize) {
    PyErr_SetString(PyExc_RuntimeError, "Import error: attempted to load two incompatible swig-generated modules.");
# if PY_VERSION_HEX >= 0x03000000
    return NULL;
# else
    return;
# endif
  }
  
  /* All objects have a 'this' attribute */
  this_descr = PyDescr_NewGetSet(SwigPyObject_type(), &this_getset_def);
  (void)this_descr;
  
  /* All objects have a 'thisown' attribute */
  thisown_descr = PyDescr_NewGetSet(SwigPyObject_type(), &thisown_getset_def);
  (void)thisown_descr;
  
  public_interface = PyList_New(0);
  public_symbol = 0;
  (void)public_symbol;
  
  PyDict_SetItemString(md, "__all__", public_interface);
  Py_DECREF(public_interface);
  for (i = 0; SwigMethods[i].ml_name != NULL; ++i)
  SwigPyBuiltin_AddPublicSymbol(public_interface, SwigMethods[i].ml_name);
  for (i = 0; swig_const_table[i].name != 0; ++i)
  SwigPyBuiltin_AddPublicSymbol(public_interface, swig_const_table[i].name);
#endif
  
  SWIG_InstallConstants(d,swig_const_table);
  
  
  pHelicsException = PyErr_NewException("_helics.HelicsException", NULL, NULL);
  Py_INCREF(pHelicsException);
  PyModule_AddObject(m, "HelicsException", pHelicsException);
  
  SWIG_Python_SetConstant(d, "helics_core_type_default",SWIG_From_int((int)(helics_core_type_default)));
  SWIG_Python_SetConstant(d, "helics_core_type_zmq",SWIG_From_int((int)(helics_core_type_zmq)));
  SWIG_Python_SetConstant(d, "helics_core_type_mpi",SWIG_From_int((int)(helics_core_type_mpi)));
  SWIG_Python_SetConstant(d, "helics_core_type_test",SWIG_From_int((int)(helics_core_type_test)));
  SWIG_Python_SetConstant(d, "helics_core_type_interprocess",SWIG_From_int((int)(helics_core_type_interprocess)));
  SWIG_Python_SetConstant(d, "helics_core_type_ipc",SWIG_From_int((int)(helics_core_type_ipc)));
  SWIG_Python_SetConstant(d, "helics_core_type_tcp",SWIG_From_int((int)(helics_core_type_tcp)));
  SWIG_Python_SetConstant(d, "helics_core_type_udp",SWIG_From_int((int)(helics_core_type_udp)));
  SWIG_Python_SetConstant(d, "helics_core_type_zmq_test",SWIG_From_int((int)(helics_core_type_zmq_test)));
  SWIG_Python_SetConstant(d, "helics_core_type_nng",SWIG_From_int((int)(helics_core_type_nng)));
  SWIG_Python_SetConstant(d, "helics_core_type_tcp_ss",SWIG_From_int((int)(helics_core_type_tcp_ss)));
  SWIG_Python_SetConstant(d, "helics_core_type_http",SWIG_From_int((int)(helics_core_type_http)));
  SWIG_Python_SetConstant(d, "helics_core_type_websocket",SWIG_From_int((int)(helics_core_type_websocket)));
  SWIG_Python_SetConstant(d, "helics_core_type_inproc",SWIG_From_int((int)(helics_core_type_inproc)));
  SWIG_Python_SetConstant(d, "helics_core_type_null",SWIG_From_int((int)(helics_core_type_null)));
  SWIG_Python_SetConstant(d, "helics_data_type_string",SWIG_From_int((int)(helics_data_type_string)));
  SWIG_Python_SetConstant(d, "helics_data_type_double",SWIG_From_int((int)(helics_data_type_double)));
  SWIG_Python_SetConstant(d, "helics_data_type_int",SWIG_From_int((int)(helics_data_type_int)));
  SWIG_Python_SetConstant(d, "helics_data_type_complex",SWIG_From_int((int)(helics_data_type_complex)));
  SWIG_Python_SetConstant(d, "helics_data_type_vector",SWIG_From_int((int)(helics_data_type_vector)));
  SWIG_Python_SetConstant(d, "helics_data_type_complex_vector",SWIG_From_int((int)(helics_data_type_complex_vector)));
  SWIG_Python_SetConstant(d, "helics_data_type_named_point",SWIG_From_int((int)(helics_data_type_named_point)));
  SWIG_Python_SetConstant(d, "helics_data_type_boolean",SWIG_From_int((int)(helics_data_type_boolean)));
  SWIG_Python_SetConstant(d, "helics_data_type_time",SWIG_From_int((int)(helics_data_type_time)));
  SWIG_Python_SetConstant(d, "helics_data_type_raw",SWIG_From_int((int)(helics_data_type_raw)));
  SWIG_Python_SetConstant(d, "helics_data_type_multi",SWIG_From_int((int)(helics_data_type_multi)));
  SWIG_Python_SetConstant(d, "helics_data_type_any",SWIG_From_int((int)(helics_data_type_any)));
  SWIG_Python_SetConstant(d, "helics_flag_observer",SWIG_From_int((int)(helics_flag_observer)));
  SWIG_Python_SetConstant(d, "helics_flag_uninterruptible",SWIG_From_int((int)(helics_flag_uninterruptible)));
  SWIG_Python_SetConstant(d, "helics_flag_interruptible",SWIG_From_int((int)(helics_flag_interruptible)));
  SWIG_Python_SetConstant(d, "helics_flag_source_only",SWIG_From_int((int)(helics_flag_source_only)));
  SWIG_Python_SetConstant(d, "helics_flag_only_transmit_on_change",SWIG_From_int((int)(helics_flag_only_transmit_on_change)));
  SWIG_Python_SetConstant(d, "helics_flag_only_update_on_change",SWIG_From_int((int)(helics_flag_only_update_on_change)));
  SWIG_Python_SetConstant(d, "helics_flag_wait_for_current_time_update",SWIG_From_int((int)(helics_flag_wait_for_current_time_update)));
  SWIG_Python_SetConstant(d, "helics_flag_restrictive_time_policy",SWIG_From_int((int)(helics_flag_restrictive_time_policy)));
  SWIG_Python_SetConstant(d, "helics_flag_rollback",SWIG_From_int((int)(helics_flag_rollback)));
  SWIG_Python_SetConstant(d, "helics_flag_forward_compute",SWIG_From_int((int)(helics_flag_forward_compute)));
  SWIG_Python_SetConstant(d, "helics_flag_realtime",SWIG_From_int((int)(helics_flag_realtime)));
  SWIG_Python_SetConstant(d, "helics_flag_single_thread_federate",SWIG_From_int((int)(helics_flag_single_thread_federate)));
  SWIG_Python_SetConstant(d, "helics_flag_slow_responding",SWIG_From_int((int)(helics_flag_slow_responding)));
  SWIG_Python_SetConstant(d, "helics_flag_delay_init_entry",SWIG_From_int((int)(helics_flag_delay_init_entry)));
  SWIG_Python_SetConstant(d, "helics_flag_enable_init_entry",SWIG_From_int((int)(helics_flag_enable_init_entry)));
  SWIG_Python_SetConstant(d, "helics_flag_ignore_time_mismatch_warnings",SWIG_From_int((int)(helics_flag_ignore_time_mismatch_warnings)));
  SWIG_Python_SetConstant(d, "helics_flag_terminate_on_error",SWIG_From_int((int)(helics_flag_terminate_on_error)));
  SWIG_Python_SetConstant(d, "helics_log_level_no_print",SWIG_From_int((int)(helics_log_level_no_print)));
  SWIG_Python_SetConstant(d, "helics_log_level_error",SWIG_From_int((int)(helics_log_level_error)));
  SWIG_Python_SetConstant(d, "helics_log_level_warning",SWIG_From_int((int)(helics_log_level_warning)));
  SWIG_Python_SetConstant(d, "helics_log_level_summary",SWIG_From_int((int)(helics_log_level_summary)));
  SWIG_Python_SetConstant(d, "helics_log_level_connections",SWIG_From_int((int)(helics_log_level_connections)));
  SWIG_Python_SetConstant(d, "helics_log_level_interfaces",SWIG_From_int((int)(helics_log_level_interfaces)));
  SWIG_Python_SetConstant(d, "helics_log_level_timing",SWIG_From_int((int)(helics_log_level_timing)));
  SWIG_Python_SetConstant(d, "helics_log_level_data",SWIG_From_int((int)(helics_log_level_data)));
  SWIG_Python_SetConstant(d, "helics_log_level_trace",SWIG_From_int((int)(helics_log_level_trace)));
  SWIG_Python_SetConstant(d, "helics_error_fatal",SWIG_From_int((int)(helics_error_fatal)));
  SWIG_Python_SetConstant(d, "helics_error_external_type",SWIG_From_int((int)(helics_error_external_type)));
  SWIG_Python_SetConstant(d, "helics_error_other",SWIG_From_int((int)(helics_error_other)));
  SWIG_Python_SetConstant(d, "helics_error_insufficient_space",SWIG_From_int((int)(helics_error_insufficient_space)));
  SWIG_Python_SetConstant(d, "helics_error_execution_failure",SWIG_From_int((int)(helics_error_execution_failure)));
  SWIG_Python_SetConstant(d, "helics_error_invalid_function_call",SWIG_From_int((int)(helics_error_invalid_function_call)));
  SWIG_Python_SetConstant(d, "helics_error_invalid_state_transition",SWIG_From_int((int)(helics_error_invalid_state_transition)));
  SWIG_Python_SetConstant(d, "helics_warning",SWIG_From_int((int)(helics_warning)));
  SWIG_Python_SetConstant(d, "helics_error_system_failure",SWIG_From_int((int)(helics_error_system_failure)));
  SWIG_Python_SetConstant(d, "helics_error_discard",SWIG_From_int((int)(helics_error_discard)));
  SWIG_Python_SetConstant(d, "helics_error_invalid_argument",SWIG_From_int((int)(helics_error_invalid_argument)));
  SWIG_Python_SetConstant(d, "helics_error_invalid_object",SWIG_From_int((int)(helics_error_invalid_object)));
  SWIG_Python_SetConstant(d, "helics_error_connection_failure",SWIG_From_int((int)(helics_error_connection_failure)));
  SWIG_Python_SetConstant(d, "helics_error_registration_failure",SWIG_From_int((int)(helics_error_registration_failure)));
  SWIG_Python_SetConstant(d, "helics_ok",SWIG_From_int((int)(helics_ok)));
  SWIG_Python_SetConstant(d, "helics_property_time_delta",SWIG_From_int((int)(helics_property_time_delta)));
  SWIG_Python_SetConstant(d, "helics_property_time_period",SWIG_From_int((int)(helics_property_time_period)));
  SWIG_Python_SetConstant(d, "helics_property_time_offset",SWIG_From_int((int)(helics_property_time_offset)));
  SWIG_Python_SetConstant(d, "helics_property_time_rt_lag",SWIG_From_int((int)(helics_property_time_rt_lag)));
  SWIG_Python_SetConstant(d, "helics_property_time_rt_lead",SWIG_From_int((int)(helics_property_time_rt_lead)));
  SWIG_Python_SetConstant(d, "helics_property_time_rt_tolerance",SWIG_From_int((int)(helics_property_time_rt_tolerance)));
  SWIG_Python_SetConstant(d, "helics_property_time_input_delay",SWIG_From_int((int)(helics_property_time_input_delay)));
  SWIG_Python_SetConstant(d, "helics_property_time_output_delay",SWIG_From_int((int)(helics_property_time_output_delay)));
  SWIG_Python_SetConstant(d, "helics_property_int_max_iterations",SWIG_From_int((int)(helics_property_int_max_iterations)));
  SWIG_Python_SetConstant(d, "helics_property_int_log_level",SWIG_From_int((int)(helics_property_int_log_level)));
  SWIG_Python_SetConstant(d, "helics_property_int_file_log_level",SWIG_From_int((int)(helics_property_int_file_log_level)));
  SWIG_Python_SetConstant(d, "helics_property_int_console_log_level",SWIG_From_int((int)(helics_property_int_console_log_level)));
  SWIG_Python_SetConstant(d, "helics_multi_input_no_op",SWIG_From_int((int)(helics_multi_input_no_op)));
  SWIG_Python_SetConstant(d, "helics_multi_input_vectorize_operation",SWIG_From_int((int)(helics_multi_input_vectorize_operation)));
  SWIG_Python_SetConstant(d, "helics_multi_input_and_operation",SWIG_From_int((int)(helics_multi_input_and_operation)));
  SWIG_Python_SetConstant(d, "helics_multi_input_or_operation",SWIG_From_int((int)(helics_multi_input_or_operation)));
  SWIG_Python_SetConstant(d, "helics_multi_input_sum_operation",SWIG_From_int((int)(helics_multi_input_sum_operation)));
  SWIG_Python_SetConstant(d, "helics_multi_input_diff_operation",SWIG_From_int((int)(helics_multi_input_diff_operation)));
  SWIG_Python_SetConstant(d, "helics_multi_input_max_operation",SWIG_From_int((int)(helics_multi_input_max_operation)));
  SWIG_Python_SetConstant(d, "helics_multi_input_min_operation",SWIG_From_int((int)(helics_multi_input_min_operation)));
  SWIG_Python_SetConstant(d, "helics_multi_input_average_operation",SWIG_From_int((int)(helics_multi_input_average_operation)));
  SWIG_Python_SetConstant(d, "helics_handle_option_connection_required",SWIG_From_int((int)(helics_handle_option_connection_required)));
  SWIG_Python_SetConstant(d, "helics_handle_option_connection_optional",SWIG_From_int((int)(helics_handle_option_connection_optional)));
  SWIG_Python_SetConstant(d, "helics_handle_option_single_connection_only",SWIG_From_int((int)(helics_handle_option_single_connection_only)));
  SWIG_Python_SetConstant(d, "helics_handle_option_multiple_connections_allowed",SWIG_From_int((int)(helics_handle_option_multiple_connections_allowed)));
  SWIG_Python_SetConstant(d, "helics_handle_option_buffer_data",SWIG_From_int((int)(helics_handle_option_buffer_data)));
  SWIG_Python_SetConstant(d, "helics_handle_option_strict_type_checking",SWIG_From_int((int)(helics_handle_option_strict_type_checking)));
  SWIG_Python_SetConstant(d, "helics_handle_option_ignore_unit_mismatch",SWIG_From_int((int)(helics_handle_option_ignore_unit_mismatch)));
  SWIG_Python_SetConstant(d, "helics_handle_option_only_transmit_on_change",SWIG_From_int((int)(helics_handle_option_only_transmit_on_change)));
  SWIG_Python_SetConstant(d, "helics_handle_option_only_update_on_change",SWIG_From_int((int)(helics_handle_option_only_update_on_change)));
  SWIG_Python_SetConstant(d, "helics_handle_option_ignore_interrupts",SWIG_From_int((int)(helics_handle_option_ignore_interrupts)));
  SWIG_Python_SetConstant(d, "helics_handle_option_multi_input_handling_method",SWIG_From_int((int)(helics_handle_option_multi_input_handling_method)));
  SWIG_Python_SetConstant(d, "helics_handle_option_input_priority_location",SWIG_From_int((int)(helics_handle_option_input_priority_location)));
  SWIG_Python_SetConstant(d, "helics_handle_option_clear_priority_list",SWIG_From_int((int)(helics_handle_option_clear_priority_list)));
  SWIG_Python_SetConstant(d, "helics_handle_option_connections",SWIG_From_int((int)(helics_handle_option_connections)));
  SWIG_Python_SetConstant(d, "helics_filter_type_custom",SWIG_From_int((int)(helics_filter_type_custom)));
  SWIG_Python_SetConstant(d, "helics_filter_type_delay",SWIG_From_int((int)(helics_filter_type_delay)));
  SWIG_Python_SetConstant(d, "helics_filter_type_random_delay",SWIG_From_int((int)(helics_filter_type_random_delay)));
  SWIG_Python_SetConstant(d, "helics_filter_type_random_drop",SWIG_From_int((int)(helics_filter_type_random_drop)));
  SWIG_Python_SetConstant(d, "helics_filter_type_reroute",SWIG_From_int((int)(helics_filter_type_reroute)));
  SWIG_Python_SetConstant(d, "helics_filter_type_clone",SWIG_From_int((int)(helics_filter_type_clone)));
  SWIG_Python_SetConstant(d, "helics_filter_type_firewall",SWIG_From_int((int)(helics_filter_type_firewall)));
  globals = SWIG_globals();
  if (!globals) {
    PyErr_SetString(PyExc_TypeError, "Failure to create SWIG globals.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  PyDict_SetItemString(md, "cvar", globals);
  Py_DECREF(globals);
  SWIG_addvarlink(globals, "helics_time_zero", Swig_var_helics_time_zero_get, Swig_var_helics_time_zero_set);
  SWIG_addvarlink(globals, "helics_time_epsilon", Swig_var_helics_time_epsilon_get, Swig_var_helics_time_epsilon_set);
  SWIG_addvarlink(globals, "helics_time_invalid", Swig_var_helics_time_invalid_get, Swig_var_helics_time_invalid_set);
  SWIG_addvarlink(globals, "helics_time_maxtime", Swig_var_helics_time_maxtime_get, Swig_var_helics_time_maxtime_set);
  SWIG_addvarlink(globals, "helics_true", Swig_var_helics_true_get, Swig_var_helics_true_set);
  SWIG_addvarlink(globals, "helics_false", Swig_var_helics_false_get, Swig_var_helics_false_set);
  SWIG_Python_SetConstant(d, "helics_iteration_request_no_iteration",SWIG_From_int((int)(helics_iteration_request_no_iteration)));
  SWIG_Python_SetConstant(d, "helics_iteration_request_force_iteration",SWIG_From_int((int)(helics_iteration_request_force_iteration)));
  SWIG_Python_SetConstant(d, "helics_iteration_request_iterate_if_needed",SWIG_From_int((int)(helics_iteration_request_iterate_if_needed)));
  SWIG_Python_SetConstant(d, "helics_iteration_result_next_step",SWIG_From_int((int)(helics_iteration_result_next_step)));
  SWIG_Python_SetConstant(d, "helics_iteration_result_error",SWIG_From_int((int)(helics_iteration_result_error)));
  SWIG_Python_SetConstant(d, "helics_iteration_result_halted",SWIG_From_int((int)(helics_iteration_result_halted)));
  SWIG_Python_SetConstant(d, "helics_iteration_result_iterating",SWIG_From_int((int)(helics_iteration_result_iterating)));
  SWIG_Python_SetConstant(d, "helics_state_startup",SWIG_From_int((int)(helics_state_startup)));
  SWIG_Python_SetConstant(d, "helics_state_initialization",SWIG_From_int((int)(helics_state_initialization)));
  SWIG_Python_SetConstant(d, "helics_state_execution",SWIG_From_int((int)(helics_state_execution)));
  SWIG_Python_SetConstant(d, "helics_state_finalize",SWIG_From_int((int)(helics_state_finalize)));
  SWIG_Python_SetConstant(d, "helics_state_error",SWIG_From_int((int)(helics_state_error)));
  SWIG_Python_SetConstant(d, "helics_state_pending_init",SWIG_From_int((int)(helics_state_pending_init)));
  SWIG_Python_SetConstant(d, "helics_state_pending_exec",SWIG_From_int((int)(helics_state_pending_exec)));
  SWIG_Python_SetConstant(d, "helics_state_pending_time",SWIG_From_int((int)(helics_state_pending_time)));
  SWIG_Python_SetConstant(d, "helics_state_pending_iterative_time",SWIG_From_int((int)(helics_state_pending_iterative_time)));
  SWIG_Python_SetConstant(d, "helics_state_pending_finalize",SWIG_From_int((int)(helics_state_pending_finalize)));
#if PY_VERSION_HEX >= 0x03000000
  return m;
#else
  return;
#endif
}

